
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// args.cpp
//*********************************************************************
//
// Notes: The Macintosh side of the following code has several
// limitations that should be understood. First, let me explain
// how the Mac CArgs works.
//
// A program that uses the CArgs class can be invoked from one of
// the following three mechanisms:
//
//    1. From an icon on the desktop,
//    2. From a document in a folder,
//    3. From another program.
//
// In the first case, the program does not receive parameters so
// CArgs::Number() is zero. This version of CArgs does not support
// the second case, though Greg R did write a class for receiving
// a document list and storing it in a CList container for use.
//
// This version does support the third case, however, if the
// program that launched this program uses the CLaunch::LaunchChild()        
// method. This method writes a small file called "$$$PARMS.FAR"
// to the Temporary Items folder before launching the child. The file
// consists of one long and a C-string. The long is the signature of
// the launching program (the parent), and the C-string is the space-
// delimited parameters in the fashion of Unix, Windows, and DOS.
// The CArgs constructor reads the $$$PARMS.FAR file and parses the
// string in the same fashion as it parses the Windows parameter
// string. It also deletes the $$$PARMS.FAR file to avoid confusion in
// other FARLIB programs.
//
// The caveats of this simple approach are: one, that if two parents
// attempt to simultaneously launch children, the single name for the
// parameter file will be a source of conflict. And, two, if a
// $$$PARMS.FAR file lingers after its use is complete, it may confuse
// a FARLIB application that is launched by methods one or two. This
// latter problem can be solved by looking for the OpenApp and OpenDoc
// highlevel events, which this CArgs chooses to ignore. This code
// should be added at some time in the not-too-distant future. The
// current use for CArgs, however, don't require it, since, except
// for the "Resident Exec," the programs are all child programs in
// the sense of method three. Since the Resident Executive doesn't
// expect parameters, it will ignore any lingering $$$PARMS.FAR file,
// and will overwrite the file when it launches a new child.
//
// Since you can pass parameters to the child program in the Windows
// environment, all this crazyness is not an issue there.
//
//*********************************************************************

#include "prefixb.h"
#include "args.h"
//#include "file.h"
#include "misc.h"
#include "record.h"
#include "strlist.h"
#include "file.h"

long        CArgs::parentSignature = 0;
short       CArgs::myArgC          = 0;
long        CArgs::myFlags         = 0;
long        CArgs::myValue         = 0;
Pchar       CArgs::myArgV[maxNrParms];
Pchar       CArgs::parms           = 0;
PCStrList   CArgs::docList;

CArgs::aNamedParm CArgs::names[maxNrParms];

PCArgs CArgs::TheArgs = 0;

//*********************************************************************
// aNamedParm class
//*********************************************************************

//---------------------------------------------------------------------
//
CArgs::aNamedParm::aNamedParm(void):
   name (0),
   value(0) {
   } // aNamedParm

//---------------------------------------------------------------------
//
CArgs::aNamedParm::~aNamedParm() {
   delete[] name;
   name = 0;
   delete[] value;
   value = 0;
   } // ~aNamedParm

//---------------------------------------------------------------------
//
void CArgs::aNamedParm::SetName(Pchar _name) {
   Assert(!StrEmpty(_name));

   STRINIT(name, _name)
   if (!StrEmpty(name))
      CapStr(name);
   } // SetName

//---------------------------------------------------------------------
//
Pchar CArgs::aNamedParm::GetName(void) {
   return name;
   } // GetName

//---------------------------------------------------------------------
//
void CArgs::aNamedParm::SetValue(Pchar _value) {
   STRINIT(value, _value)
   } // SetValue

//---------------------------------------------------------------------
//
Pchar CArgs::aNamedParm::GetValue(void) {
   return value;
   } // GetValue

//---------------------------------------------------------------------
//
bool CArgs::aNamedParm::Match(Pchar _name) {
   char temp[sizeString];
   Strcpy(temp, _name, sizeString);
   CapStr(temp);
   return (strcmp(temp, name) == 0);
   } // Match

//---------------------------------------------------------------------
//
bool CArgs::aNamedParm::AtEnd(void) {
   return (!name);
   } // AtEnd

//*********************************************************************
// CArgs class
//*********************************************************************

//---------------------------------------------------------------------
// Expose 0x7f as ' ', 0x7e as '/', 0x7d as '-'
//
void exposeChars(Pchar _p) {
   for (Pchar p = _p; *p; p++)
      if      (*p == char(0x7f))
         *p = ' ';
      else if (*p == char(0x7e))
         *p = '/';
      else if (*p == char(0x7d))
         *p = '-';
   } // exposeChars

//---------------------------------------------------------------------
// Constructor
// (Use this constructor on the Mac)
//
CArgs::CArgs(int _argc, Pchar _argv[]):
   CAtom() {

   Assert(!TheArgs);
   TheArgs = this;

   int iv;

   for (iv = 0; iv < maxNrParms; iv++)
      myArgV[iv] = 0;

#ifdef FV_MACINTOSH
   char*** cmmd = 0;
   int argc     = ccommand(cmmd);
   Pchar* argv = *cmmd;
#else
   int argc    = _argc;
   Pchar* argv = _argv;
#endif // FV_MACINTOSH

   _argc;     //  Silencing Compiler Warning
   _argv;     //  Silencing Compiler Warning

   parms    = new char[sizeString];
   parms[0] = 0;

   for (iv = 0; iv < argc; iv++) {
      char parm[sizeString];
      Strcpy(parm, argv[iv], sizeString);
      Trim(parm);

      Strcat(parms, parm, sizeString);
      ChrCat(parms, ' ', sizeString);

      if (parm[0] == '"')
         strcpy(parm, parm+1);
      if (parm[strlen(parm)-1] == '"')
         parm[strlen(parm)-1] = 0;

      // Parameters beginning with "-" or "/"
      // are flags or a number
      if (parm[0] == '-' || parm[0] == '/') {
         // Handle flag parameter
         Strcpy(parm, parm+1, sizeString);
         Trim(parm);
         CapStr(parm);
         if (IsNumber(parm))
            myValue = atol(parm);
         else for (int i = 0; i < strlen(parm); i++) {
            if (parm[i] >= 'A' && parm[i] <= 'Z') {
               int iFlag = parm[i] - 'A';
               long mask = 1L << iFlag;
               myFlags  |= mask;
               }
            }
         }
      else if (!StrEmpty(parm)) {
         // Handle ordinary parameter
         STRINIT(myArgV[myArgC], parm);
         myArgC++;
         if (myArgC > maxNrParms)
            break;
         }
      }
   } // CArgs

#ifndef FV_UNIX

//---------------------------------------------------------------------
// Constructor
//
CArgs::CArgs(Pchar _lpCmdLine, Pchar _names):
   CAtom() {

   Assert(!TheArgs);
   TheArgs = this;

   int iv;

   for (iv = 0; iv < maxNrParms; iv++)
      myArgV[iv] = 0;

   parms    = new char[sizeString];
   parms[0] = 0;

#if   defined(FV_MSWIN16)
   Strcpy(parms, Pchar("***execName*** "), sizeString);
   Strcat(parms, _lpCmdLine,               sizeString);
//?Squeeze(parms, ' ');

#elif defined(FV_MSWIN32)
   _lpCmdLine;

   Strcpy(parms, ::GetCommandLine(), sizeString);
//?Squeeze(parms, ' ');

   // Hide embedded special characters (spaces, slashes, and dashes)
   bool inq = false;
   for (Pchar p = parms; *p; p++)
      if (inq) {                                       // inside quotes
         if      (*p == '"')
            inq = false;
         else if (*p == ' ')
            *p = char(0x7f);
         else if (*p == '/')
            *p = char(0x7e);
         else if (*p == '-')
            *p = char(0x7d);
         }
      else if (*p == '"')                             // outside quotes
         inq = true;
//?Assert(!inq);

#elif defined(FV_MACINTOSH)
   _lpCmdLine;

//?   receiveInfo();
   receiveParms(0, 0, 0);

   // Check for a document list
   if (docList) {
      delete[] parms;
      parms = 0;
      return;
      }

   Trim(parms);
   PCRecord r = new CRecord(parms, Pchar(" "));

   // Set up to parse the named parameters
   PCRecord n = new CRecord();
   if (!StrEmpty(_names)) {
      n->NewRecord(_names, Pchar(","));
//?   names = new aNamedParm[n->Number()+1];
      }

   // Parse the command line into separate parameters
   for (iv = 1; iv <= r->Number(); iv++) {
      char parm[sizeString];
      r->GetField(iv, parm, sizeString);

      if (StrEmpty(parm))
         continue;

      // Remove quotes
      if (parm[0] == '"')
         strcpy(parm, parm+1);
      if (parm[strlen(parm)-1] == '"')
         parm[strlen(parm)-1] = 0;

      // Parameters listed in the names constructor parameter
      // are named parameters
      if (n->Number()) {
         exposeChars(parm);
         bool isValue = false;
         for (int in = 1; in <= n->Number(); in++) {
            char name[sizeString];
            n->GetField(in, name, sizeString);
            if      (ChrPos('=', parm) >= 0) {
               // parameter has format <name=value>
               PCRecord p = new CRecord(parm, Pchar("="));
               char work[sizeString];
               p->GetField(1, work, sizeString);
               if (stricmp(name, work) == 0) {
                  names[in-1].SetName(work);
                  p->GetField(2, work, sizeString);
                  names[in-1].SetValue(work);
                  isValue = true;
                  break;
                  }
               }
            }
         if (isValue)
            continue;
         }

      // Parameters beginning with "-" or "/"
      // are flags or a number
      if (parm[0] == '-' || parm[0] == '/') {
         // Handle flag parameter
         Strcpy(parm, parm+1, sizeString);
         CapStr(parm);
         if (IsNumber(parm))
            myValue = atol(parm);
         else for (int i = 0; i < strlen(parm); i++) {
            if (parm[i] >= 'A' && parm[i] <= 'Z') {
               int iFlag = parm[i] - 'A';
               long mask = 1L << iFlag;
               myFlags  |= mask;
               }
            }
         }

      // Handle ordinary parameter
      else {
         exposeChars(parm);
         myArgV[myArgC] = new char[strlen(parm)+1];
         strcpy(myArgV[myArgC], parm);
         myArgC++;
         if (myArgC > maxNrParms)
            break;
         }
      }
#endif // FV_MACINTOSH

   delete[] parms;
   parms = 0;
   } // CArgs

#endif // ifndef FV_UNIX

//---------------------------------------------------------------------
// Destructor
//
CArgs::~CArgs() {
//?delete docList;

   int iv;

   for (iv = 0; iv < maxNrParms; iv++) {
      delete[] myArgV[iv];
      myArgV[iv] = 0;
      }
   } // ~CArgs

//---------------------------------------------------------------------
//
RTTI_CPP(Args, Atom)

//---------------------------------------------------------------------
//
bool CArgs::NewArgs(int _argc, Pchar _argv[]) {
   parentSignature = 0;
   myArgC          = 0;
   myFlags         = 0;
   myValue         = 0;
   delete[] parms;
   parms           = 0;

   int iv;

   for (iv = 0; iv < maxNrParms; iv++)
      if (myArgV[iv]) {
         delete[] myArgV[iv];
         myArgV[iv] = 0;
         }

#ifdef FV_MACINTOSH
   char*** cmmd = 0;
   int argc     = ccommand(cmmd);
   Pchar* argv  = *cmmd;
#else
   int argc     = _argc;
   Pchar* argv  = _argv;
#endif // FV_MACINTOSH

   _argc;     //  Silencing Compiler Warning
   _argv;     //  Silencing Compiler Warning

   delete[] parms;
   parms    = new char[sizeString];
   parms[0] = 0;

   for (iv = 0; iv < argc; iv++) {
      char parm[sizeString];
      Strcpy(parm, argv[iv], sizeString);
      Trim(parm);

      Strcat(parms, parm, sizeString);
      ChrCat(parms, ' ', sizeString);

      if (parm[0] == '"')
         strcpy(parm, parm+1);
      if (parm[strlen(parm)-1] == '"')
         parm[strlen(parm)-1] = 0;


      // Parameters beginning with "-" or "/"
      // are flags or a number
      if (parm[0] == '-' || parm[0] == '/') {
         // Handle flag parameter
         Strcpy(parm, parm+1, sizeString);
         Trim(parm);
         CapStr(parm);
         if (IsNumber(parm))
            myValue = atol(parm);
         else for (int i = 0; i < strlen(parm); i++) {
            if (parm[i] >= 'A' && parm[i] <= 'Z') {
               int iFlag = parm[i] - 'A';
               long mask = 1L << iFlag;
               myFlags  |= mask;
               }
            }
         }
      else if (!StrEmpty(parm)) {
         // Handle ordinary parameter
         STRINIT(myArgV[myArgC], parm);
         myArgC++;
         if (myArgC > maxNrParms)
            break;
         }
      }

   return true;
   } // NewArgs

//---------------------------------------------------------------------
//
bool CArgs::NewArgs(Pchar _lpCmdLine, Pchar _names) {
   delete[] parms;
   parms    = new char[sizeString];
   parms[0] = 0;

   // Pick up parameter zero from the original command line
   ChrCpy(parms, '"');
   Strcat(parms, Arg(0, false), sizeString);
   Strcat(parms, Pchar("\" "),  sizeString);
   Strcat(parms, _lpCmdLine,    sizeString);

   parentSignature = 0;
   myArgC          = 0;
   myFlags         = 0;
   myValue         = 0;

   int iv = 0;

   for (iv = 0; iv < maxNrParms; iv++)
      if (myArgV[iv]) {
         delete[] myArgV[iv];
         myArgV[iv] = 0;
         }

   // Hide embedded spaces
   bool inq = false;
   for (Pchar p = parms; *p; p++)
      if (inq) {                                       // inside quotes
         if      (*p == '"')
            inq = false;
         else if (*p == ' ')
            *p = char(0x7f);
         }
      else if (*p == '"')                             // outside quotes
         inq = true;
//?Assert(!inq);

   Trim(parms);
   PCRecord r = new CRecord(parms, Pchar(" "));

   // Set up to parse the named parameters
   PCRecord n = new CRecord();
   if (_names) {
      n->NewRecord(_names, Pchar(","));
//?   names = new aNamedParm[n->Number()+1];
      }

   // Parse the command line into separate parameters
   short isValue = 0;
   for (iv = 1; iv <= r->Number(); iv++) {
      char parm[sizeString];
      r->GetField(iv, parm, sizeString);

      if (StrEmpty(parm))
         continue;

      // Remove quotes
      if (parm[0] == '"')
         strcpy(parm, parm+1);
      if (parm[strlen(parm)-1] == '"')
         parm[strlen(parm)-1] = 0;

      // Re-expose embedded spaces
      for (Pchar p = parm; *p; p++)
         if (*p == char(0x7f))
            *p = ' ';

      // If isValue not zero, then parameter is value
      // parameter following a name parameter. (see the
      // else-if code below for set up conditions.
      if (isValue) {
         names[isValue].SetValue(parm);
         isValue = 0;
         }

      // Parameters listed are named parameters
      else if (n->Number()) {
         for (int in = 1; in <= n->Number(); in++) {
            char name[sizeString];
            n->GetField(in, name, sizeString);
            if      (ChrPos('=', parm) >= 0) {
               // parameter has format <name=value>
               PCRecord p = new CRecord(parm, Pchar("="));
               char work[sizeString];
               p->GetField(1, work, sizeString);
               if (stricmp(name, work) == 0) {
                  names[in-1].SetName(work);
                  p->GetField(2, work, sizeString);
                  names[in-1].SetValue(work);
                  break;
                  }
               }
            else if (stricmp(name, parm) == 0) {
               // Next parameter is the associated value (see
               // if code above for value handling)
               names[in-1].SetName(parm);
               isValue = iv;
               break;
               }
            }
         }

      // Parameters beginning with "-" or "/"
      // are flags or a number
      else if (parm[0] == '-' || parm[0] == '/') {
         // Handle flag parameter
         Strcpy(parm, parm+1, sizeString);
         CapStr(parm);
         if (IsNumber(parm))
            myValue = atol(parm);
         else for (int i = 0; i < strlen(parm); i++) {
            if (parm[i] >= 'A' && parm[i] <= 'Z') {
               int iFlag = parm[i] - 'A';
               long mask = 1L << iFlag;
               myFlags  |= mask;
               }
            }
         }

      // Handle ordinary parameter
      else {
         STRINIT(myArgV[myArgC], parm)
         myArgC++;
         if (myArgC > maxNrParms)
            break;
         }
      }
   delete[] parms;
   parms = 0;

   return true;
   } // NewArgs

//---------------------------------------------------------------------
// Return number of parameters
//
short CArgs::Number(void) {
   if (docList)
      return docList->Number();
   else
      return myArgC-1;
   } // Number

//---------------------------------------------------------------------
// Return the specified named parameter string. Note that
// _value is a value returned by VArg, it should not be
// a pointer to a preallocated area.
//
// ***Warning: Reserved for farSlang use only***
//
Pchar CArgs::VArg(Pchar _name, Pchar _dflt) {
   Assert(_name);

   // Test for any named parameters specified
   if (!names) {
      Pchar p = new char[1];
      p[0]    = 0;
      return p;
      }

   // Search to find the specified name
   for (short in = 0; !(names[in].AtEnd()); in++) {
      if (names[in].Match(_name)) {
         if (!(names[in].GetValue()))
            names[in].SetValue(_dflt);
         Pchar p = new char[strlen(names[in].GetValue())+1];
         strcpy(p, names[in].GetValue());
         return p;
         }
      }

   // No named parameter matched the specified name
   Pchar p = new char[1];
   p[0]    = 0;
   return p;
   } // VArg

//---------------------------------------------------------------------
// Return the specified named parameter string. Note that
// _value is a value returned by VArg, it should not be
// a pointer to a preallocated area.
//
bool CArgs::VArg(Pchar _name, Pchar _value, ushort _size, Pchar _dflt) {
   Assert(!StrEmpty(_name));
   Assert(_value);

   _value[0] = 0;

   // Test for any named parameters specified
   if (!names)
      return false;

   // Search to find the specified name
   for (short in = 0; !(names[in].AtEnd()); in++) {
      if (names[in].Match(_name)) {
         if (!(names[in].GetValue()))
            names[in].SetValue(_dflt);
         Strcpy(_value, names[in].GetValue(), _size);
         return true;
         }
      }

   // No named parameter matched the specified name
   return false;
   } // VArg

//---------------------------------------------------------------------
// Return the specified parameter string
//
// _iArg    Result
//   -1     unparsed command line
//    0     path of executable
//    1     first parameter
//    2     second parameter
//   etc
//
Pchar CArgs::Arg(short _iArg, bool _alc) {
   char work[sizeString];

   if (docList) {
      docList->GotoItem(_iArg);
      return docList->StrItem(_alc);
      }
   else {
//???????      _iArg--;
      if (_iArg < 0 || _iArg >= myArgC)
         if      (_iArg == -1)
            if (_alc) {
               Pchar p = new char[strlen(parms)+1];
               strcpy(p, parms);
               return p;
               }
            else
               return parms;
         else
            if (_alc) {
               Pchar p = new char[1];
               p[0]    = 0;
               return p;
               }
            else
               return 0;
      else
         if (_alc) {
            Pchar p = new char[strlen(myArgV[_iArg])+1];
            strcpy(p, myArgV[_iArg]);
            return p;
            }
         else
            return myArgV[_iArg];
      }
   } // Arg

//---------------------------------------------------------------------
// farSlang version of Arg
//
Pchar CArgs::Arg(short _iArg) {
   return Arg(_iArg, true);
   } // Arg

//---------------------------------------------------------------------
// Return true if the specified flag parameter is set
//
bool CArgs::FArg(char _flag) {
   CapChr(_flag);
   if (_flag < 'A' || _flag > 'Z')
      return false;

   long iFlag = _flag - 'A';
   long mask = 1L << iFlag;
   if (myFlags & mask)
      return true;
   return false;
   } // FArg

//---------------------------------------------------------------------
// Return value
//
long CArgs::IArg(void) {
   return myValue;
   } // IArg

//---------------------------------------------------------------------
// Set/Clear specified argument
//
void CArgs::SetArg(short _iArg, Pchar _val) {
   if (docList)
      Assert(true);
   else {
      if (!InBnd(_iArg, short(0), short(maxNrParms-1)))
         return;

      delete myArgV[myArgC];
      if (StrEmpty(_val)) {
         myArgV[myArgC]    = new char[1];
         myArgV[myArgC][0] = '\0';
         }
      else {
         myArgV[myArgC]    = new char[strlen(_val)];
         strcpy(myArgV[myArgC], _val);
         }
      }
   } // SetArg

//---------------------------------------------------------------------
// Set specified flag
//
void CArgs::SetFArg(char _flag) {
   CapChr(_flag);
   if (!InBnd(_flag, 'A', 'Z'))
      return;

   int iFlag = _flag - 'A';
   long mask = 1L << iFlag;
   myFlags  |= mask;
   } // SetFArg

//---------------------------------------------------------------------
// Clear specified flag
//
void CArgs::ClrFArg(char _flag) {
   CapChr(_flag);
   if (!InBnd(_flag, 'A', 'Z'))
      return;

   int iFlag = _flag - 'A';
   long mask = ~(1L << iFlag);
   myFlags  &= mask;
   } // ClrFArg

//---------------------------------------------------------------------
// Set integer flag value
//
void CArgs::SetIArg(long _val) {
   myValue = _val;
   } // SetIArg

#if   defined(FV_MACINTOSH)

/*?
//---------------------------------------------------------------------
// Get starting information from the program that launched me.
// That could either be a document list or a parameter string.
//
void CArgs::receiveInfo(void)
{
    // Setup the event loop to trap selected high-level events

    // Setup for documents event
    AEEventHandlerUPP doOpenDocUPP = NewAEEventHandlerProc(receiveDocs);
   ::AEInstallEventHandler(kCoreEventClass,
                            kAEOpenDocuments,
                            doOpenDocUPP,
                          0,
                          false);

    // Setup for normal event
    AEEventHandlerUPP doOpenAppUPP = NewAEEventHandlerProc(receiveParms);
    ::AEInstallEventHandler(kCoreEventClass,  // ??? what is this to be???
                            kAEOpenApplication,
                            doOpenAppUPP,
                            0,
                            false);

    // event loop to receive and process the selected high-level events
    long timeOut = 1000;                          // wait a decent time
   EventRecord event;
    while (true)
    {
        ::WaitNextEvent(highLevelEventMask, &event, 0, NULL);
        if (timeOut-- <= 0)
            break;
        if (event.what == nullEvent)
           continue;
        if (event.what == kHighLevelEvent)
        {
           ::AEProcessAppleEvent(&event);
           break;
        }
        else
        {   Beep();
            Assert(0);
        }
    }
} // receiveInfo

//---------------------------------------------------------------------
// Get passed parameters
//
pascal OSErr CArgs::receiveDocs(const AppleEvent* _ae,
                                      AppleEvent* ,
                                      long        )
{
   OSErr      myErr;
   AEDescList itemList;

   // get the direct parameter - a descriptor
    // list - and put it into documentList
   myErr = ::AEGetParamDesc(_ae, keyDirectObject, typeChar, &itemList);
   if (myErr != noErr)
      return myErr;

    long itemsInList;

   // count the number of descriptor records in the list
   myErr = ::AECountItems(&itemList, &itemsInList);
   if (myErr != noErr)
        return myErr;

    docList = new CStrList();

   // now get each descriptor record from the list,
   // coerce the returned data to an FSSpec record,
   // turn it into a fullpath, and insert it
   for (long index = 1; index <= itemsInList; ++index)
   {
      DescType  retType;
      FSSpec    docToOpen;
      AEKeyword keyWd;
      Size      size;

      myErr = ::AEGetNthPtr(&itemList,
                              index,
                              typeFSS,
                              &keyWd,
                         &retType,
                              &docToOpen,
                              sizeof(docToOpen),
                              &size);
      if (myErr != noErr)
         return myErr;

        char path[sizeString];
      FSSpecToFullPath(docToOpen, path);
      docList->StrInsert(path);
   }
   return noErr;
} // receiveDocs
?*/

//---------------------------------------------------------------------
// Get parameters from the caller, if any. Check the
// TemporaryItems folder for a file called "$$$PARMS.FAR"
//
pascal OSErr CArgs::receiveParms(const AppleEvent*, AppleEvent*, long) {
   // Get the name for the parameter file
   char name[sizeString];
   memset(name, 0, sizeString);
   if (!getParmsName(name))
      return false;

   // Prepare the file for input
   PCFile file = new CFile(name);
   file->OpenInput();
   if (file->Error() != NOPROBLEM)                  // Probably no file
      return false;

   // Read the parameters
   file->ReadString(parms, sizeString);

   // Read the signature
   char work[sizeString];
   memset(work, 0, sizeString);
   file->ReadString(work, sizeString);
   parentSignature = StrToLong(work);

   // Close out and destroy the parameter file
   file->Close();
   RemoveFile(name);

   return true;
   } // receiveParms

//---------------------------------------------------------------------
// Construct the path name for the parameter file.
//
bool CArgs::getParmsName(Pchar _name, short _size) {
   // Get the salient information for the boot disk and
   // the TemporaryItems folder (make the folder if it
   // doesn't already exist)
   OSErr err;
   short bootDsk;
   long  tempDir;
   err = long(::FindFolder(kOnSystemDisk,
                           'temp',
                           kCreateFolder,
                           &bootDsk,
                           &tempDir));
   if (err != noErr)
      return false;

   // Create the parameter file name

   // Get the path string (Mac Programming FAQs, pg 149)
   Strcpy(_name, Pchar("$$$PARMS.FAR"), _size);

   unsigned char dir[sizeString];

   PCInfoPBRec myCPB = new CInfoPBRec();
   myCPB.dirInfo.ioNamePtr = (unsigned char*)dir;
   myCPB.dirInfo.ioDrParID = tempDir;

   do {
      memset(dir, 0, sizeString);
      myCPB.dirInfo.ioVRefNum   = bootDsk;
      myCPB.dirInfo.ioFDirIndex = -1;
      myCPB.dirInfo.ioDrDirID   = myCPB.dirInfo.ioDrParID;
      err = long(::PBGetCatInfoSync(&myCPB));
      if (err != noErr)
         return false;

      PtoCstr(dir);
      Strcat(Pchar(dir), Pchar(":"),   sizeString);
      Strcat(Pchar(dir), _name,        sizeString);
      Strcpy(_name,      Pchar(dir),   _size);
      } while (myCPB.dirInfo.ioDrDirID != fsRtDirID);
   return true;
   } // getParmsName

//---------------------------------------------------------------------
// Fill fullPath with a C string that is the full path of the file
// specified by fss. (First it gets the path of the directory)
//
OSErr CArgs::FSSpecToFullPath(FSSpec _fss,
                              Pchar  _pathName,
                              short  _size) {
   OSErr err;

   err = FSSpecToDirectoryPath(_fss, _pathName);
   p2cstr(_fss.name);
   Strcat(_pathName, Pchar(_fss.name), _size);
   return err;
   } // FSSpecToFullPath

//---------------------------------------------------------------------
// Fill dirPath with a C string that is the path of the directory
// that contains the file specified by fss.  The filename itself is
// ignored.
//
OSErr CArgs::FSSpecToDirectoryPath (FSSpec _fss, Pchar _dirPath) {
   DirInfo block;
   Str255  dirName;
   OSErr   err;

   _dirPath[0] = 0;       // treated like a pascal string until the end

   block.ioDrParID = _fss.parID;
   block.ioNamePtr = dirName;
   do {
      block.ioVRefNum   = _fss.vRefNum;
      block.ioFDirIndex = -1;
      block.ioDrDirID   = block.ioDrParID;
      err               = ::PBGetCatInfoSync((CInfoPBRec *)&block);
      if (err != noErr)
         return err;
      PStrCat(dirName, StringPtr(Pchar("\p:")));
      PStrInsert(StringPtr(_dirPath), dirName);
   } while (block.ioDrDirID != 2);

   p2cstr((StringPtr)_dirPath);
   return err;
   } // FSSpecToDirectoryPath

//---------------------------------------------------------------------
// Add string 'src' to end of string 'dst'
//
void CArgs::PStrCat(StringPtr _dst, StringPtr _src) {
   // copy string in
   BlockMove(_src + 1, _dst + *_dst + 1, *_src);
   // adjust length byte
   *_dst += *_src;
   } // PStrCat

//---------------------------------------------------------------------
// Insert string 'src' at beginning of string 'dst'
//
void CArgs::PStrInsert(StringPtr _dst, StringPtr _src) {
   // make room for new string
   BlockMove(_dst + 1, _dst + *_src + 1, *_dst);
   // copy new string in
   BlockMove(_src + 1, _dst + 1, *_src);
   // adjust length byte
   *_dst += *_src;
   } // PStrInsert

#endif // FV_MACINTOSH

#ifdef FV_UNIX

//---------------------------------------------------------------------
// Get the full path name of the specified executable file name
// Original code by Dave Fullerton 2.22.2001
//
bool CArgs::GetExecPath(Pchar _execName, Pchar _fullPath, ushort _size) {
   ulong err;
   // Check if full path name in the executable name parameter
   if (_execName[0] == '/') {
      Strcpy(_fullPath, _execName, _size);

      // Check for existence
      if (FileExists(_fullPath, err))
         return true;
      else {
         Assert(false);
         return false;
         }
      }

   // Assume it is relative

   // Next, check the current directory
   if (::getcwd(_fullPath, sizeLine) == 0) {
      Assert(false);
      return false;
      }

   // Construct the relative path name
   ChrCat(_fullPath, '/',       _size);
   Strcat(_fullPath, _execName, _size);

   // Test for existence
   if (FileExists(_fullPath, err))
      return true;

   // Ok, let's search the paths in the PATH variable

   // Get the path environment variable
   PCRecord paths = new CRecord(::getenv("PATH"), Pchar(":"));
   for (ushort i = 1; i <= paths->Number(); i++) {
      paths->GetField(i, _fullPath, _size);
      if (StrEmpty(_fullPath))
         continue;
      if (_fullPath[strlen(_fullPath)-1] != '/')
         ChrCat(_fullPath, '/', _size);
      Strcat(_fullPath, _execName, _size);

      // Check for existence
      if (FileExists(_fullPath, err))
         return true;
      }

   // Couldn't find it, even though it's around here somewhere
   Assert(false);
   return false;
   } // GetExecPath

#endif // FV_UNIX
