
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// note: memcpy can only copy up to 65k bytes.

//*********************************************************************
// block.cpp
//*********************************************************************

#include "prefixb.h"
#include "common.h"
#include "block.h"
#include "file.h"

#define MEM_BUFFER_SIZE 10

//---------------------------------------------------------------------
// Conversion constructor
//
CBlock::CBlock(void):
   CItem (),
   h     (0),
   p     (0),
   ct    (0),
   sz    (0),
   pBuff (0),
   iBuff (0),
   saveAs(saveAsText) {
   } // CBlock

//---------------------------------------------------------------------
// Copy Constructor
//
CBlock::CBlock(const PCBlock _block):
   CItem (),
   h     (0),
   p     (0),
   ct    (0),
   sz    (_block->sz),
   pBuff (0),
   iBuff (0),
   saveAs(saveAsText) {

   if (sz) {

#if   defined(FV_MSWINDOWS)
      // Allocate the block and copy the source
      Alloc(sz);
      Lock();
      Assert(p);
      _block->Lock();
      memcpy(p, _block->p, size_t(sz));
      _block->Unlock();
      Unlock();

#elif defined(FV_MACINTOSH)
      h = _block->h;
      ::HandToHand (&h);

#elif defined(FV_UNIX)
      Alloc(sz);
      Lock();
      Assert(p);
      _block->Lock();
      memcpy(p, _block->p, size_t(sz));
      _block->Unlock();
      Unlock();
#endif // FV_UNIX
      }
   } // CBlock

//---------------------------------------------------------------------
// Conversion Constructor
//
CBlock::CBlock(const Pchar _buffer, ulong _size):
   CItem (),
   h     (0),
   p     (0),
   ct    (0),
   sz    (_size),
   pBuff (0),
   iBuff (0),
   saveAs(saveAsText) {

   if (_buffer && sz == 0)
      sz = strlen(_buffer)+1;

   // Allocate the block and copy the source
   if (_buffer && sz > 0) {
      Alloc(sz);
      Lock();
      Assert(p);
      memcpy(p, _buffer, sz);
      Unlock();
      }
   } // CBlock

#ifndef FV_UNIX

//---------------------------------------------------------------------
//
CBlock::CBlock(HGLOBAL _handle):
   CItem (),
   h     (_handle),
   p     (0),
   ct    (0),
   sz    (0),
   pBuff (0),
   iBuff (0),
   saveAs(saveAsText) {

#if   defined(FV_MSWINDOWS)
   sz = ::GlobalSize(h);

#elif defined(FV_MACINTOSH)
   sz = ::GetHandleSize(h);

#endif // FV_MACINTOSH
   } // // CBlock

#endif // FV_UNIX

//---------------------------------------------------------------------
// Destructor
//
CBlock::~CBlock(void) {
   Free();
   } // ~CBlock

//---------------------------------------------------------------------
//
RTTI_CPP(Block, Item)

//---------------------------------------------------------------------
// Copy block
//
bool CBlock::NewBlock(const PCBlock _block) {
   Free();
   _block->Lock();

#if   defined(FV_MSWINDOWS)
   if (_block->p && _block->sz > 0) {
      // Allocate the block and copy the source
      Alloc(_block->sz);
      Lock();
      Assert(p);
      memcpy(p, _block->p, size_t(sz));
      Unlock();
      _block->Unlock();
      return true;
      }
   else {
      _block->Unlock();
      return false;
      }

#elif defined(FV_MACINTOSH)
//TBD
   h = _block->h;
   ::HandToHand (&h);

#elif defined(FV_UNIX)
   if (_block->p && _block->sz > 0) {
      // Allocate the block and copy the source
      Alloc(_block->sz);
      Lock();
      Assert(p);
      memcpy(p, _block->p, size_t(sz));
      Unlock();
      _block->Unlock();
      return true;
      }
   else {
      _block->Unlock();
      return false;
      }
#endif // FV_UNIX
   } // NewBlock

//---------------------------------------------------------------------
//
bool CBlock::NewBlock(const Pchar _buffer, ulong _size) {
   Free();

   // Allocate the block and copy the source
   sz = _size;
   if (_buffer && sz == 0)
      sz = strlen(_buffer)+1;

   if (_buffer && sz > 0) {
      Alloc(sz);
      Lock();
      Assert(p);
      memcpy(p, _buffer, sz);
      Unlock();
      return true;
      }
   else
      return false;
   } // NewBlock

//---------------------------------------------------------------------
// Copy this. return a pointer to a new CBlock object
//
// if _begin > 0, then offset beginning as specified.
// if _end   > 0, then offset ending as specified.
// if _begin < 0, then add sz to set beginning.
// if _end   < 0, then add sz to set ending.
// if _begin = 0 and _end = 0, then copy entire block.
//
PCBlock CBlock::SubBlock(long _begin, long _end) {
   PCBlock block = new CBlock();                         // empty block

   if (sz) {
      // Bound the block to copy
      long begin = 0;
      long end   = sz-1;
      if (_begin == 0 && _end == 0);
      else {
         if      (_begin < 0)
            begin = _begin + sz;
         else if (_begin > 0)
            begin = _begin;
         if      (_end   < 0)
            end   = _end   + sz;
         else if (_end   > 0)
            end   = _end;
         }
      if (begin <= sz-1 && end > 0) {
         begin = Bound(begin, 0L, long(sz-1));
         end   = Bound(end,   0L, long(sz-1));
         long size = end - begin + 1;

         // Allocate the block and copy it
         block->Alloc(size);
         block->Lock();
         if (p)
            memcpy(block->p, p + begin, size);
         else {
            Lock();
            memcpy(block->p, p + begin, size);
            Unlock();
            }
         block->Unlock();
         }
      }

   return block;
   } // SubBlock

//---------------------------------------------------------------------
// Copy this. return a pointer to a new CBlock object
//
// if _begin > 0, then offset beginning as specified.
// if _end   > 0, then offset ending as specified.
// if _begin < 0, then add sz to set beginning.
// if _end   < 0, then add sz to set ending.
// if _begin = 0 and _end = 0, then copy entire block.
//
PCBlock CBlock::operator()(long _begin, long _end) {
   PCBlock block = new CBlock();                         // empty block

   if (sz) {
      // Bound the block to copy
      long begin = 0;
      long end   = sz-1;
      if (_begin == 0 && _end == 0);
      else {
         if      (_begin < 0)
            begin = _begin + sz;
         else if (_begin > 0)
            begin = _begin;
         if      (_end   < 0)
            end   = _end   + sz;
         else if (_end   > 0)
            end   = _end;
         }
      if (begin <= sz-1 && end > 0) {
         begin = Bound(begin, 0L, long(sz-1));
         end   = Bound(end,   0L, long(sz-1));
         long size = end - begin + 1;

         // Allocate the block and copy it
         block->Alloc(size);
         block->Lock();
         if (p)
            memcpy(block->p, p + begin, size);
         else {
            Lock();
            memcpy(block->p, p + begin, size);
            Unlock();
            }
         block->Unlock();
         }
      }
   return block;
   } //  operator()

//---------------------------------------------------------------------
// Test the equality of the specified block
//
bool CBlock::Equals(PCBlock _block) {
   if (!h || !_block->h)
      return false;

   if (sz != _block->sz)
      return false;

   Lock();
   _block->Lock();

   if (!p && !_block->p) {
      Unlock();
      _block->Unlock();
      return true;
      }

   if (!p || !_block->p) {
      Unlock();
      _block->Unlock();
      return false;
      }

   bool equals = (memcmp(p, _block->p, sz) == 0);

   Unlock();
   _block->Unlock();

   return equals;
   } // Equals

//---------------------------------------------------------------------
// Allocate global memory
//
bool CBlock::Alloc(ulong _sz, ulong _mod) {
   Assert(_sz);
   if (h)
      Free();
   ct = 0;
   p  = 0;

   // Round size up if requested
   sz = _sz;
   if (_mod && (sz % _mod))
      sz = _mod*(1+(sz/_mod));

#if   defined(FV_MSWINDOWS)
   h       = ::GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, sz);

#elif defined(FV_MACINTOSH)
   h       = ::NewHandle(sz);

#elif defined(FV_UNIX)
   h       = (HGLOBAL)(::malloc(sz));
#endif // FV_UNIX

if (!h)
   Assert(h);
   return (h != 0);
   } // Alloc

//---------------------------------------------------------------------
// Reallocate global memory
//
bool CBlock::ReAlloc(ulong _sz, ulong _mod) {
   Assert(_sz);
   int wasLocked = ct;
   while (p)
      Unlock();

   // Round size up if requested
   sz = _sz;
   if (_mod && (sz % _mod))
      sz = _mod*(1+(sz/_mod));

#if   defined(FV_MSWINDOWS)
   h = ::GlobalReAlloc(h, sz, GMEM_MOVEABLE | GMEM_ZEROINIT);

#elif defined(FV_MACINTOSH)
tbd h = ::NewHandle(sz);

#elif defined(FV_UNIX)
   h = (HGLOBAL)(::malloc(sz));

#endif // FV_UNIX

   Assert(h);

   while (wasLocked--)
      Lock();

   return (h != 0);
   } // ReAlloc

//---------------------------------------------------------------------
// Return allocated block handle
//
HGLOBAL CBlock::Handle(void) {
   return h;
   } // Handle

//---------------------------------------------------------------------
// Return true if the block is locked
//
bool CBlock::Locked(void) {
   return (h != 0 && p != 0);
   } // Locked

//---------------------------------------------------------------------
// Lock the block and return the address of the buffer
//
Pchar CBlock::Lock(void) {
   if (!h)
      return 0;

   if (p == 0) {
#if   defined(FV_MSWINDOWS)
      p         = Pchar(::GlobalLock(h));

#elif defined(FV_MACINTOSH)
      int state = ::HGetState(h);
      OSErr err = ::MemError();
      ::HLock(h);
      p = *h;

#elif defined(FV_UNIX)
      p = Pchar(h);
#endif // FV_UNIX
      }
      
   Assert(p);
   ct++;
   return p;
   } // Lock

//---------------------------------------------------------------------
// Return the address of the locked block
//
Pchar CBlock::Pointer(void) {
   return p;
   } // Pointer

//---------------------------------------------------------------------
// Return true if the block is empty
//
bool CBlock::Empty(void) {
   return h == 0;
   } // Empty

//---------------------------------------------------------------------
// Return the size of the block
//
ulong CBlock::GetSize(void) {
   return sz;
   } // GetSize

//---------------------------------------------------------------------
// Return the size of the used portion of the block. CBlock doesn't
// know this, but CVar does.
//
ulong CBlock::GetContentSize(void) {
   return sz;
   } // GetContentSize

//---------------------------------------------------------------------
// Clear the block of memory
//
void CBlock::Clear(char _ch) {
   Assert(p);

   memset(p, _ch, sz);
   } // Clear

//---------------------------------------------------------------------
// Replace the contents of this with the contents of the specified
// buffer.
//
void CBlock::Copy(Pchar _buffer, ulong _size) {
   NewBlock(_buffer, _size);
   } // Copy

//---------------------------------------------------------------------
// Replace the contents of this with the contents of the specified
// string.
//
void CBlock::Copy(Pchar _string) {
   Free();
   Create();
   AddString(_string);
   } // Copy

//---------------------------------------------------------------------
// Replace the contents of this with the contents of the specified
// variable.
//
void CBlock::Copy(PCBlock _blk) {
   NewBlock(_blk);
   iBuff = 0;
   } // Copy

//---------------------------------------------------------------------
// Allocate a memory buffer of MEM_BUFFER_SIZE. This buffer can then
// grow as needed.
//
void CBlock::Create(ulong _size) {
if (sz)
   Assert(!sz);

   Alloc(Max(_size, MEM_BUFFER_SIZE));
   pBuff = 0;
   Lock();
   Assert(p);
   *p = 0;
   Unlock();
   } // Create

//---------------------------------------------------------------------
// Increase by half again the size of the buffer.
//
void CBlock::Grow(ulong _size) {
   if (!sz)
      Create(_size);
   else {
      size_t size = ((sz+_size) == 1) ? MEM_BUFFER_SIZE
                                      : ((sz+_size) * 3) >> 1;
      ReAlloc(ulong(size));
      }
   } // Grow

//---------------------------------------------------------------------
// Trim the trailing excess from the buffer.
//
void CBlock::Trim(void) {
   if (!sz)
      return;
   else {
      size_t size = sz - 1;
      Lock();
      Pchar pp = p + sz - 1;
      while (!*pp && pp > p) {
         pp--;
         size--;
         }
      Unlock();
      ReAlloc(ulong(size+1));
      }
   } // Grow

//---------------------------------------------------------------------
// Add a single byte to the buffer, grow if needed.
//
void CBlock::AddByte(uchar _byte) {
   if (size_t(pBuff) + 1 >= sz)
      Grow();

   Lock();
   Assert(p);
   p[pBuff++] = _byte;
   Unlock();
   } // AddByte

//---------------------------------------------------------------------
// Add a range of bytes to the buffer, grow if needed. At end, pBuffer
// points to the next available byte.
//
void CBlock::AddBuffer(Pchar _buffer, ulong _size) {
   if (_size == 0)
      return;
      
   while (size_t(pBuff) + _size >= sz)
       Grow(_size);

   Lock();
   Assert(p);
   memcpy(p+ pBuff, _buffer, _size);
   Unlock();
   pBuff += long(_size);
   } // AddBuffer

//---------------------------------------------------------------------
// Add the contents of the specified variable, grow if needed. At end,
// pBuffer points to the next available byte.
//
void CBlock::AddBuffer(PCBlock _blk, ulong _size) {
   _blk->Lock();
   Assert(_blk->p);

   long size = _size < _blk->sz ? _size : _blk->sz;

   while (size_t(pBuff) + size >= sz)
      Grow(size);

   Lock();
   Assert(p);

   memcpy(p+pBuff, _blk->p, size);
   pBuff += long(size);
   Unlock();

   _blk->Unlock();
   } // AddBuffer

//---------------------------------------------------------------------
// Add the specified character to the buffer, grow if needed.
//
void CBlock::AddChar(char _ch) {
   if (size_t(pBuff) + 2 >= sz)
      Grow();

   Lock();
   Assert(p);
   
   ChrCat(p, _ch, ushort(sz));
   pBuff++;
   Unlock();
   } // AddChar

//---------------------------------------------------------------------
// Add the specified string to the buffer, grow if needed.
//
void CBlock::AddString(Pchar _string) {
   size_t szB = strlen(_string);
   while (size_t(pBuff) + szB + 1 >= sz)
       Grow(size_t(pBuff)+szB+1);

   Lock();
   Assert(p);
   Strcat(p, _string, sz);
   pBuff += long(szB);

   Unlock();
   } // AddString

//---------------------------------------------------------------------
// Add the specified string to the buffer, grow if needed.
//
void CBlock::AddString(PCBlock _blk) {
   _blk->Lock();
   Assert(_blk->p);


   size_t szB = strlen(_blk->p);
   while (size_t(pBuff) + szB + 1 >= sz)
       Grow(size_t(pBuff)+szB+1);

   Lock();
   Assert(p);
   Strcat(p, _blk->p, sz);
   pBuff += long(strlen(_blk->p));
   Unlock();

   _blk->Unlock();
   } // AddString

//---------------------------------------------------------------------
//
void CBlock::Reset(ulong _iBuff) {
   iBuff = _iBuff;
   } // Reset

//---------------------------------------------------------------------
//
void CBlock::BackUp(ulong _i) {
   if (iBuff - _i >= 0)
      iBuff -= _i;
   } // BackUp

//---------------------------------------------------------------------
// Return a character from the current location
//
char CBlock::GetChar(void) {
   Assert(p);
   Assert(iBuff >= 0);
   Assert(iBuff < sz);

   return p[iBuff++];
   } // GetChar

//---------------------------------------------------------------------
// Return a character from the specified location
//
char CBlock::GetChar(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index < sz);

   return p[_index];
   } // GetChar

//---------------------------------------------------------------------
// Set the specified character
//
void CBlock::PutChar(ulong _index, char _value) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index < sz);

   p[_index] = _value;
   } // PutChar

//---------------------------------------------------------------------
//
bool CBlock::GetFirstLine(Pchar _line, ushort _size) {
   if (!p)
      return false;

   tp = p;
   if (*tp == 0)
      return false;

   Pchar tr = tp;
   for (;;)
      if      (*tr == 0) {
         Strcpy(_line, tp, _size);
         tp = tr+1;
         return true;
         }
      else if (*tr == '\n') {
         *tr = 0;
         Strcpy(_line, tp, _size);
         *tr++ = '\n';
         if (*tr == '\r')
            tr++;
         tp = tr;
         return true;
         }
      else if (*tr == '\r') {
         *tr = 0;
         Strcpy(_line, tp, _size);
         *tr++ = '\r';
         if (*tr == '\n')
            tr++;
         tp = tr;
         return true;
         }
      else
         tr++;
   } // GetFirstLine

//---------------------------------------------------------------------
//<method name="GetNextLine" type="?">
//<param name="_line" type="$"/>
//<param name="_size" type="#"/>
//<info>
// Sets _line to the next string obtained from the buffer starting at
// the end of the last gotten string and ending either at a new-line or
// a zero (the end of the buffer). Uses _size to avoid a buffer overrun.
// Leaves the block in a state to obtain the next available string in
// the buffer on a subsequent call to this method. Returns true if a
// string was found, false if at the end of the buffer, indicating that
// there are no more strings.
//</info>
//</method>
//
bool CBlock::GetNextLine(Pchar _line, ushort _size) {
   if (!p)
      return false;

   if (!tp)
      return false;

   if (*tp == 0)
      return false;

   Pchar tr = tp;
   for (;;)
      if      (*tr == 0) {
         Strcpy(_line, tp, _size);
         tp = tr+1;
         return true;
         }
      else if (*tr == '\n') {
         *tr = 0;
         Strcpy(_line, tp, _size);
         *tr++ = '\n';
         if (*tr == '\r')
            tr++;
         tp = tr;
         return true;
         }
      else if (*tr == '\r') {
         *tr = 0;
         Strcpy(_line, tp, _size);
         *tr++ = '\r';
         if (*tr == '\n')
            tr++;
         tp = tr;
         return true;
         }
      else
         tr++;
   } // GetNextLine

//---------------------------------------------------------------------
// Return a short from the specified location
//
short CBlock::GetShort(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*2+1 < sz);

   Pshort ps = Pshort(p+_index*2);
   return *ps;
   } // GetShort

//---------------------------------------------------------------------
// Set the specified short
//
void CBlock::PutShort(ulong _index, short _value) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*2+1 < sz);

   Pshort ps = Pshort(p+_index*2);
   *ps       = _value;
   } // PutShort

//---------------------------------------------------------------------
// Return a long from the specified location
//
long CBlock::GetLong(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*4+3 < sz);

   Plong pl = Plong(p+_index*4);
   return *pl;
   } // GetLong

//---------------------------------------------------------------------
// Set the specified long
//
void CBlock::PutLong(ulong _index, long _value) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*4+3 < sz);

   Plong pl = Plong(p + _index*4);
   *pl      = _value;
   } // PutLong

#ifndef FV_NOPOINTS
//---------------------------------------------------------------------
// A point is two longs
//
PCPoint CBlock::GetPoint(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   return new CPoint(pl[0], pl[1]);
   } // GetPoint

//---------------------------------------------------------------------
//
long CBlock::GetX(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   return pl[0];
   } // GetX

//---------------------------------------------------------------------
//
long CBlock::GetY(ulong _index) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   return pl[1];
   } // GetY

//---------------------------------------------------------------------
//
void CBlock::PutPoint(ulong _index, PCPoint _point) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   pl[0] = _point->x;
   pl[1] = _point->y;
   } // PutPoint

//---------------------------------------------------------------------
//
void CBlock::PutPoint(ulong _index, long _x, long _y) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   pl[0] = _x;
   pl[1] = _y;
   } // PutPoint

//---------------------------------------------------------------------
//
void CBlock::PutX(ulong _index, long _x) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   pl[0] = _x;
   } // PutX

//---------------------------------------------------------------------
//
void CBlock::PutY(ulong _index, long _y) {
   Assert(p);
   Assert(_index >= 0);
   Assert(_index*8+7 < sz);

   Plong pl = Plong(p + _index*8);
   pl[1] = _y;
   } // PutY
#endif // FV_NOPOINTS

//---------------------------------------------------------------------
// Copy a block
//
void CBlock::PutBlock(Pvoid _block, ulong _size) {
   Assert(_size);

   Alloc(_size);
   Lock();
   Assert(p);

   if (_block)
      memcpy(p, _block, sz);
   else
      memset(p, 0, sz);

   Unlock();
   } // PutBlock

//---------------------------------------------------------------------
// Unlock the block of memory
//
void CBlock::Unlock(void) {
   if (p == 0)
      return;

   if (ct == 1) {

#if   defined(FV_MSWINDOWS)
      int i = ::GlobalFlags(h) & GMEM_LOCKCOUNT;
      while (i) {
         ::GlobalUnlock(h);
         i = ::GlobalFlags(h) & GMEM_LOCKCOUNT;
         }

#elif defined(FV_MACINTOSH)
      ::HUnlock (h);
      OSErr err = MemError();

#elif defined(FV_UNIX)
   
      // no action for unix
#endif // FV_UNIX

      p = 0;
      }

   ct--;
   } // Unlock

//---------------------------------------------------------------------
// Release the allocated memory back to Windows
//
void CBlock::Free(void) {
#if defined(FV_MSWINDOWS)
   if (h) {
      if (p) {
         int i = ::GlobalFlags(h) & GMEM_LOCKCOUNT;
         while (i) {
            ::GlobalUnlock(h);
            i = ::GlobalFlags(h) & GMEM_LOCKCOUNT;
            }
         }

      ::GlobalFree(h);
      }

#elif defined(FV_MACINTOSH)
   if (h)
      ::DisposeHandle(h);

#elif defined(FV_UNIX)

   if (h)
      ::free(Pchar(h));
#endif // FV_UNIX

   h  = 0;
   p  = 0;
   ct = 0;
   sz = 0;
   } // Free

//---------------------------------------------------------------------
// Lose the allocated memory.
// (use with caution - you must have deleted the memory through other
// means.)
//
void CBlock::Lose(void) {
   h  = 0;
   p  = 0;
   ct = 0;
   sz = 0;
   } // Lose

//---------------------------------------------------------------------
//
bool CBlock::BlockTo(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   // Check for actual data
   Pchar p = block->Lock();
   Assert(p);

   // Copy the block
   NewBlock(p, block->GetSize());

   block->Unlock();

   return true;
   } // BlockTo

//---------------------------------------------------------------------
//
bool CBlock::ToBlock(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   // Check for actual data
   Pchar p = block->Lock();
   Assert(p);

   // Copy the block
   block->NewBlock(p, block->GetSize());

   return true;
   } // ToBlock

//---------------------------------------------------------------------
// Return the size of the space that the block occupies in the farBook
//
ulong CBlock::BookSize(void) {
   return sz + sizeof(sz);
   } // BookSize

//---------------------------------------------------------------------
// Read a binary block from the specified file
//
bool CBlock::ReadFile(PCAtom _input, bool _isBook) {
   _isBook;

   Free();

   err = FVE_OK;

   // Read the block size from the file
   if (!(TYPECAST(_input, File)->ReadBlock(&sz, sizeof(sz)))) {
      err = _input->Error();
      return false;
      }

   Alloc(sz);
   Lock();
   Assert(p);

   // Read the block from the file
   ulong szRead = 0;
   if (!(TYPECAST(_input, File)->ReadBlock(p, sz, &szRead))) {
      char msg[sizeString];
      sprintf(msg, "Expected %ld but read %ld", sz, szRead);
      Assert(strlen(msg) < sizeString);
      Free();
      err = _input->Error();
      return false;
      }

   Unlock();

   return true;
   } // ReadFile

//---------------------------------------------------------------------
// Write the block to the specified file
//
bool CBlock::WriteFile(PCAtom _output, bool _isBook) {
   _isBook;
   
   Assert(h);

   err = FVE_OK;

   Lock();
   Assert(p);

   // Write the block size to the file
   if (!(TYPECAST(_output, File)->WriteBlock(&sz, sizeof(sz)))) {
      Unlock();
      err = _output->Error();
      return false;
      }

   // Write the block to the file
   if (!(TYPECAST(_output, File)->WriteBlock(p, sz))) {
      Unlock();
      err = _output->Error();
      return false;
      }

   Unlock();

   return true;
   } // WriteFile

//---------------------------------------------------------------------
// Read a binary file into the block
// (Block is left unlocked)
//
bool CBlock::Load(Pchar _fileName) {
   err = FVE_OK;

   Free();

   // Prepare the file for input
   CFile input(_fileName);
   if (!input.OpenInput()) {
      err = input.Error();
      return false;
      }

   Alloc(input.Size()+1);
   Lock();
   Assert(p);
   memset(p, 0, sz);

   // Read the block from the file
   if (!input.ReadBlock(p, sz-1)) {
      Free();
      err = input.Error();
      input.Close();
      return false;
      }

   Unlock();
   input.Close();

   return true;
   } // Load

//---------------------------------------------------------------------
// Write a binary file from the block
// (Block is left unlocked)
//
bool CBlock::Save(Pchar _fileName) {
   if (!h)
      return false;

   err = FVE_OK;

   // Prepare the file for output
   CFile output(_fileName);
   if (!output.OpenOutput()) {
      err = output.Error();
      return false;
      }

   Pchar hold = p;
   Lock();
   Assert(p);

   if (saveAs == saveAsText) {
      // Write the block to the file
      if (!output.WriteString(p)) {
         if (!hold)
            Unlock();
         err = output.Error();
         output.Close();
         return false;
         }
      }
   else
      // Write the block to the file
      if (!output.WriteBlock(p, sz)) {
         if (!hold)
            Unlock();
         err = output.Error();
         output.Close();
         return false;
         }

   Unlock();
   output.Close();

   return true;
   } // Save

