
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// book.cpp
//*********************************************************************
//
// Notes:
//
// Do not replace farview version without resolving differences.
//
// Keys:
// * retain marks to identify name-type and duplicate-type keys.
// * retain extended form of reference instead of simple ref.
// (what is an "extended form"? PJM 2000/2/6.It includes the offset
// field. PJM 2002/03/29)
//
// Refs table optimization:
//
// I can avoid wasting unneeded refs subtables if I don't allocate them
// for single alias keys. I would need to identify the key as pointing
// directly to the node descriptor. When a second key is created, I
// would have to update the hash key entry and create an arefs
// subtable. The offset field is 32-bits long. This is really wasted
// space. So, using a bit here would have virtually no impact on the
// address space of the system. This can all be managed in CBook. No
// changes to CHash or CRefs are necessary. This means additional time
// overhead when building a new farBook in batch mode, since it means
// having to update the CHash entry for the key when a second alias is
// received. The approach gains more than it costs, however.
// (PJM 2000/2/1)
//
// Parts space reuse:
//
// I can avoid wasting space in the farBook by reusing old space by
// allocating some extra for a node part according to user preference
// when I write a part out initially. When it is updated, check if
// there is still enough space for the new version. If so, use the old
// space, if not, write the part to a new location at the end of the
// farBook, again allocating additional space according to the
// algorithm.
//
// Buttons and targets:
//
// In HTML, an <href...> has a mark and a URL. A farVIEW button has
// a mark and a targetset name. Targets also have a mark and a
// targetset name. This allows more than one button to link to more
// than one target. When a target is declared, its targetset name is
// added to the dictionary as a form of alias. When a button is
// declared, no action is taken. When you click the button, the system
// looks up the targetset name in the dictionary. That points to a Refs
// bucket chain that contains pointers for the target(s).
//
// I expect I will support XML markup of native farVIEW documents. A
// button would be like
//
// <Btn color="..." font="..."targetset= "mytarget">Go there</Button>
//
// and a target would be like
//
// <Tgt color="..." targetset="mytarget">Come here</Tgt>
//
// or, if no mark is needed,
//
// <Tgt targetset="mytarget"/>
//
// (To be clicked, a button needs a mark, though it would not be
// required.)
//
// BTW, a targetset name can also be a URL, in which case, no target is
// required, of course, and the <Btn> acts like an <href>.
//
// Well, thinking about it, if I support markup in the text, how am I
// going to know when it has changed? I guess it would have to be an
// operation in the editor so that the change is made on the spot. It
// might be good to provide a command to verify marks when a document
// is imported from another editor.
//
// Locking: add a field to the farBook header that identifies the
// farBook as sharable. When an farBook is sharable, it writes a single
// byte preceeding each node descriptor. The single byte is the "lock
// byte." It is used only for locking and unlocking the node, but it is
// never read from nor written to again. This will help to make the
// farBook compatible across platforms. Once the header flag is
// implemented, always set it on when creating a new farBook. This is
// to protect legacy farBooks; viz., mine.
//
//*********************************************************************

#include "prefixb.h"
#include "book.h"
#include "bookinfo.h"
#include "misc.h"

bool  checkVER           = true;

const ushort nrMatchKeys =   42;
const ushort blowOut     =    1;
const ushort infoSize    =  512;
const char   lockByte    =    0;

static char  version[]          = "3.00";
static char  cpyrite[]          = "Copyright 1987-2006, "
                                  "Paul J Medlock, All Rights Reserved";

//=====================================================================
//                       aDesc implementation
//=====================================================================

short aDesc::nrParts = 2;

//---------------------------------------------------------------------
// Node Descriptor Methods
// Note: this class limits the largest embeddable document size to 
// 16,777,215 bytes, since it uses the high 8-bits to store the farVIEW 
// type of the object.
//
aDesc::aDesc(void):
   parts(0) {
   parts = new ALNdx[nrParts];

   for (short i = 0; i < nrParts; ++i) {
      parts[i].where = 0;
      parts[i].size  = 0;
      }
   } // aDesc

//---------------------------------------------------------------------
//
aDesc::aDesc(const RaDesc _desc) {
   for (short i = 0; i < nrParts; ++i) {
      parts[i].where = _desc.parts[i].where;
      parts[i].size  = _desc.parts[i].size;
      }
   } // aDesc

//---------------------------------------------------------------------
//
aDesc::~aDesc() {
   delete[] parts;
   parts = 0;
   } // ~aDesc

//---------------------------------------------------------------------
//
void  aDesc::NrParts(short _nrParts) {
   nrParts = _nrParts;
   } // NrParts

//---------------------------------------------------------------------
//
short aDesc::NrParts(void) {
   return nrParts;
   } // NrParts

//---------------------------------------------------------------------
//
void aDesc::Clear(void) {
   for (short i = 0; i < nrParts; ++i) {
      parts[i].where = 0;
      parts[i].size  = 0;
      }
   } // Clear

//---------------------------------------------------------------------
//
short aDesc::Type(short _iPart) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);

   return short((parts[_iPart].size >> 24) & 0xff);
   } // Type

//---------------------------------------------------------------------
//
void aDesc::Type(short _iPart, short _type) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);

   ulong type = ulong(_type) << 24;
   ulong size = parts[_iPart].size & 0x00ffffff;
   parts[_iPart].size = type | size;
   } // Type

//---------------------------------------------------------------------
//
ulong aDesc::Where(short _iPart) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);

   return parts[_iPart].where;
   } // Where

//---------------------------------------------------------------------
//
void aDesc::Where(short _iPart, ulong _where) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);

   parts[_iPart].where = _where;
   } // Where

//---------------------------------------------------------------------
//
ulong aDesc::Size(short _iPart) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);

   return parts[_iPart].size & 0x00ffffff;
   } // Size

//---------------------------------------------------------------------
//
void aDesc::Size(short _iPart, ulong _size) {
   Assert(_iPart >= 0);
   Assert(_iPart < nrParts);
   Assert(_size <= 0x00ffffff);

   ulong type = parts[_iPart].size & 0xff000000;
   parts[_iPart].size = type | _size;
   } // Size

//---------------------------------------------------------------------
//
bool aDesc::ReadDesc(RCFile _file) {
   delete[] parts;
   parts = new ALNdx[nrParts];
   return _file.ReadBlock(parts, nrParts * sizeof(ALNdx));
   } // ReadDesc

//---------------------------------------------------------------------
//
bool aDesc::WriteDesc(RCFile _file) {
   return _file.WriteBlock(parts, nrParts * sizeof(ALNdx));
   } // WriteDesc

//=====================================================================
//                       CBook implementation
//=====================================================================

//---------------------------------------------------------------------
// farBook object constructor
//
CBook::CBook(Pchar _bookName):
   CAtom   (),
   bookName(0),
   bookInfo(0),
   action  (noAction),
   iDesc   (),
   lDesc   (),
   error   (0) {

   STRINIT(bookName, _bookName)

   // Initialize the farBook information record
   bookInfo = new CBookInfo(infoSize);
   bookInfo->SetTemplate(Pchar("|bookName|version|copyright|bookNode|homeNode"
                               "|nrParts|nrRefs|nrHash|szKey|nrKeys|shared"));
   bookInfo->NewRecord  (Pchar("|||||||||||"));
   bookInfo->PutField   (Pchar("bookName"),  _bookName);
   bookInfo->PutField   (Pchar("version"),   version);
   bookInfo->PutField   (Pchar("copyright"), cpyrite);
   bookInfo->PutField   (Pchar("bookNode"),  Pchar("<Book>"));
   bookInfo->PutField   (Pchar("homeNode"),  Pchar("<Home>"));
   bookInfo->PutField   (Pchar("nrParts"),   long(aDesc::NrParts()));
   bookInfo->PutField   (Pchar("nrRefs"),    Pchar("0"));
   bookInfo->PutField   (Pchar("nrHash"),    Pchar("0"));
   bookInfo->PutField   (Pchar("szKey"),     Pchar("0"));
   bookInfo->PutField   (Pchar("nrKeys"),    Pchar("0"));
   bookInfo->PutField   (Pchar("shared"),    long(1));

   sizes = new CRecord();
   } // CBook

//---------------------------------------------------------------------
// farBook object destructor
//
CBook::~CBook() {
   delete[] bookName;
   bookName = 0;
   } // ~CBook

//---------------------------------------------------------------------
//
RTTI_CPP(Book, Atom)

//---------------------------------------------------------------------
//
Pchar CBook::Name(bool _alc) {
   if (_alc) {
      Pchar n = new char[strlen(bookName)+1];
      strcpy(n, bookName);
      return n;
      }
   else
      return bookName;
   } // Name

//---------------------------------------------------------------------
//
bool CBook::Slow(void) {
   return false;
   } // Slow

//---------------------------------------------------------------------
//
Pchar empty = Pchar("");
Pchar CBook::GetFVL(void) {
   return empty;
   } // GetFVL

//---------------------------------------------------------------------
//
short CBook::NrParts(void) {
   return aDesc::NrParts();
   } // NrParts

//---------------------------------------------------------------------
//
void CBook::NrParts(short _nrParts) {
   aDesc::NrParts(_nrParts);
   bookInfo->PutField(Pchar("nrParts"), long(aDesc::NrParts()));
   } // NrParts

//---------------------------------------------------------------------
//
bool CBook::PutInfoField(Pchar _name, Pchar _value) {
   if (bookInfo->GetFieldNr(_name) == 0)
      return bookInfo->AppendField(_name, _value);
   return bookInfo->PutField(_name, _value);
   } // PutInfoField

//---------------------------------------------------------------------
// ***Warning: Reserved for farSlang use only***
//
Pchar CBook::GetInfoField(Pchar _name) {
   return bookInfo->GetField(_name);
   } // GetInfoField

//---------------------------------------------------------------------
//
bool CBook::GetInfoField(Pchar _name, Pchar _value, short _size) {
   return bookInfo->GetField(_name, _value, _size);
   } // GetInfoField

//---------------------------------------------------------------------
//
bool CBook::UpdateInfo(void) {
   return false;
   } // UpdateInfo

//---------------------------------------------------------------------
// Create a new farBook, deleting the old one, if present. Note that
// SetSizes must already have been called.
//
bool CBook::Create(long _nrParts,
                   long _nrRefSlots,
                   long _nrHshSlots,
                   long _sizeKey,
                   long _nrKeys) {
   _nrParts, _nrRefSlots, _nrHshSlots, _sizeKey, _nrKeys;

   return false;
   } // Create

//---------------------------------------------------------------------
// Open the farBook files to prepare for access.
//
bool CBook::Open(void) {
   return false;
   } // Open

//---------------------------------------------------------------------
// Close the farBook files
//
bool CBook::Close(void) {
   return false;
   } // Close

//---------------------------------------------------------------------
// Lock the iNode specified by _id.nodeID.
//
bool CBook::Lock(PCRef _ref, AWaitKind _wait) {
   _ref, _wait;
   
   return false;
   } // Lock

//---------------------------------------------------------------------
// Unlock the iNode specified by _id.nodeID.
//
bool CBook::Unlock(PCRef _ref) {
   _ref;
   
   return false;
   } // Unlock

//---------------------------------------------------------------------
//
PCBookInfo CBook::GetBookInfo(void) {
   return bookInfo;
   } // GetBookInfo

//---------------------------------------------------------------------
//
PCRefs CBook::GetFile(void) {
   return 0;
   } // GetFile

//---------------------------------------------------------------------
//
bool CBook::GetDesc(ulong _loc, PCRecord _desc) {
   _desc->Clear();
   return false;
   } // GetDesc

//---------------------------------------------------------------------
//                            link i/o
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// Remove extra spaces from the key and capitalize it
//
void CBook::prepKey(AMark _mark, Pchar _key, short _size, short _szKey) {
   Assert(!StrEmpty(_key));

   // Check if the key is already prepared
   if (_mark == isPrepped)
      return;

   // Trim key to regulation size
   Pack(_key);
   CapStr(_key);
   if (_key[0] == anAlias) {
      if (strlen(_key) > _szKey - 2)
         _key[_szKey - 2] = '\0';
      }
   else
      if (strlen(_key) > _szKey)
         _key[_szKey] = '\0';
   Trim(_key, backEnd);

   // Prefix the key with the mark
   char key[sizeString];
   memset(key, 0, sizeString);
   key[0] = _mark;
   strncat(key, _key, sizeString-2);

   // Remove spaces and capitalize
   Strcpy(_key, key, _size);
   } // prepKey

//---------------------------------------------------------------------
// Add the specified key to the farBook dictionary.
//
// Marks are treated as follows:
//   * aName type keys must be unique.
//   * other types need not be unique.
//
bool CBook::MakeBookKey(AMark  _mark,
                        Pchar  _key,
                        PCRef _ref,
                        bool   _trunc) {
   _mark,
   _key,
   _ref,
   _trunc;

   return false;
   } // MakeBookKey

//---------------------------------------------------------------------
// Check the farBook dictionary to determine if the key is present.
//
bool CBook::IfBookKey(AMark _mark, Pchar _key, bool _trunc) {
   _mark,
   _key,
   _trunc;

   return false;
   } // IfBookKey

//---------------------------------------------------------------------
// Use the key to find all descriptors that are identified
// by the key, and add them to the findList. Return true if
// any descriptors are found.
//
bool CBook::FindBookKey(AMark   _mark,
                        Pchar   _key,
                        PCList _pFind,
                        bool    _trunc) {
   _mark,
   _key,
   _pFind,
   _trunc;
   
   return false;
   } // FindBookKey

//---------------------------------------------------------------------
// Delete the key from the farBook index
//
bool CBook::DeleteBookKey(AMark  _mark,
                          Pchar  _key,
                          PCRef _ref,
                          bool   _trunc) {
   _mark,
   _key,
   _ref,
   _trunc;

   return false;
   } // DeleteBookKey

//---------------------------------------------------------------------
// Rename the current iNode
//
bool CBook::RenameNode(Pchar _oldName, Pchar _newName, bool _trunc) {
   _oldName,
   _newName,
   _trunc;
   
   return false;
   } // RenameNode

//---------------------------------------------------------------------
// Search interface
//
bool CBook::Search(Pchar   _domain,
                   Pchar   _expression,
                   bool    _fulText,
                   PCList _pFind) {
   _domain;
   _expression;
   _fulText;
   _pFind;
   
   return false;
   } // Search
   
//---------------------------------------------------------------------
//                            iNode i/o
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// Lock, open and read the specified part
//
PCAtom CBook::ReadPart(PCRef _ref) {
   _ref;

   return 0;
   } // ReadPart

//---------------------------------------------------------------------
// Lock open and write the specified part
//
bool CBook::WritePart(PCRef _ref, PCAtom _part) {
   _ref,
   _part;

   return false;
   } // WritePart

//---------------------------------------------------------------------
// Lock open and create the specified node
//
bool CBook::CreateNode(PCRef _ref, PCAtom _body) {
   _ref,
   _body;

   return false;
   } // CreateNode

//---------------------------------------------------------------------
// Get content if in cache, else use ReadPartWithLock to get it
//
PCAtom CBook::GetContent(PCRef _ref) {
   _ref;

   return 0;
   } // GetContent

//---------------------------------------------------------------------
// Default action is to let the IfBookKey() method do it
//
bool CBook::HasTargets(AMark _mark, Pchar _key, bool _trunc) {
   return IfBookKey(_mark, _key, _trunc);
   } // HasTargets

