
//*********************************************************************
// sample.cpp
//*********************************************************************
//
// 1. Create a book, if needed.
// 2. Open the book.
// 3. Read this source file into a strlist.
// 4. Add the source as a topic to the book.
// 5. Reference the topic in the book by another name.
// 6. Check that the topic is known by both names.
// 7. Get the source file from the book using the other name.
// 8. Write it to a different file.
// 9. Close the book.
//
//*********************************************************************

#include "prefixb.h"
#include "args.h"
#include "block.h"
#include "misc.h"
#include "odb.h"
#include "strlist.h"
#include "syslog.h"

// Parameters to set the sizes in the DB if we need to create it. These
// should be in the configuration file to be read only when we need to
// create the DB, but they are here for now.
long nrRefSlots         = 32;
long nrHshSlots         = 32;
long szKey              = 128;
long nrKeys             = 10000;

char logName[] = "farlibsample.tlog";
char dbName[]  = "farlibsample.tdb";
char srcName[] = "farlibsample.cpp";
char savName[] = "farlibsample.tsav";

SysLog* logger;

//---------------------------------------------------------------------
//
#if    defined(FV_MSWINDOWS)

int WINAPI WinMain(HINSTANCE _hInstance,
                   HINSTANCE _hPrevInstance,
                   LPSTR     _lpszCmdShow,
                   int       _nCmdShow) {

   _hInstance, _hPrevInstance, _lpszCmdShow, _nCmdShow;

   // Parse the command-line parameters
   new CArgs(_lpszCmdShow);
   
#elif defined(FV_UNIX)

int main(int argc, char** argv) {

   // Parse the command-line parameters
   new CArgs(argc, argv);
   
#endif

   // Normally, you set command line flags on the command line
   // using the usual, e.g., -N, but since you are just starting,
   // I'll make it easy here by setting them in this code.
   CArgs::TheArgs->SetFArg('N');

   // Disable logging cache (when debugging)
   CArgs::TheArgs->SetFArg('Z');

   // Note: Windows compilers don't require the Pchar() thingie,
   // but I needed it for gcc for some reason. If you know how
   // to avoid it, let me know. Using the Pchar() doesn't bother
   // the Windows compilers though.
   
   logger = new SysLog(logName);
   SysLog::Log('N', Pchar("Sample Program Begins"));

   PCODB theDB = new CODB(dbName);

   //------------------------------------------------------------------
   // Create the DB if needed
   ulong err;
   if (!FileExists(dbName, err)) {
      if (!theDB->Create(1, nrRefSlots, nrHshSlots, szKey, nrKeys)) {
         SysLog::Log('N', Pchar("Error: Could not create the DB"));
         delete logger;
         return 0;
         }
      SysLog::Log('N', Pchar("Created %s"), dbName);
      }

   SysLog::Log('N', Pchar("Opening %s"), dbName);
   SysLog::Log('N', Pchar(""));

   if (!theDB->Open()) {
      SysLog::Log('E', Pchar("Failed to open the DB at %s"), dbName);
      delete logger;
      return 0;
      }

   // From here, I am dispensing with the error checking. 
      
   //------------------------------------------------------------------
   // Read the source file into a string list
   PCStrList src = new CStrList();
   src->Load(srcName);

   //------------------------------------------------------------------
   // Save the source as a topic to the book
   theDB->AddTopic(src, Pchar("farlibsample"));

   //------------------------------------------------------------------
   // Add an unrelated reference
   theDB->AddRef(Pchar("farlibsample"), Pchar("unrelated reference"));
   
   //------------------------------------------------------------------
   // Check that the name links to the topic in the book
   if (theDB->TopicExists(Pchar("farlibsample")))
      SysLog::Log('N', Pchar("Found the topic in the book"));
   else
      SysLog::Log('N', Pchar("Could not find the topic in the book"));

   //------------------------------------------------------------------
   // Retrieve the topic using the reference name
   theDB->Query(Pchar("unrelated reference"), ODB_REF);
   PCBlock bsav = theDB->Topic(0);
   PCStrList sav = new CStrList();
   sav->BlockTo(bsav);
    
   //------------------------------------------------------------------
   // Save the source in the string list to a different file
   sav->Save(savName);

   SysLog::Log('N', Pchar("Wrote the source to the save file"));

   theDB->Close();
   
   delete logger; // not a child of CAtom, so we should delete it

   SysLog::Log('N', Pchar("Sample Program Ends"));

   return 0;
   } // WinMain
