
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// file.cpp
//*********************************************************************

#include "prefixb.h"
#include "atom.h"
#include "file.h"
#include "misc.h"
#include "record.h"
#ifndef FV_NOREMOTE
#include "rmtfile.h"
#endif
#ifndef FV_NOLOG
#include "syslog.h"
#endif
#ifndef FV_NOTHREAD
#include "thread.h"
#endif

const char cr = '\r';
const char lf = '\n';

// default end-of-line delimiter
#if   defined(FV_MSWINDOWS)
char CFile::eolDelim[] = {0x0d, 0x0a, 0};

#elif defined(FV_MACINTOSH)
char CFile::eolDelim[] = "\r";

#elif defined(FV_UNIX)
char CFile::eolDelim[] = "\n";

#endif // FV_UNIX

//---------------------------------------------------------------------
// Static method to return a pointer
// to the platform's standare EOL
//
Pchar CFile::GetDefaultDelim(void) {
   return &eolDelim[0];
   } // GetDefaultDelim

#ifndef FV_NOREMOTE
//---------------------------------------------------------------------
// Analyse the specified URI parameter and return the C.File object to
// handle the farBook. The form of a URI is
//
//    http://domainName[:port]/filePath
// or
//    far://domainName[:port]/filePath
//
// to identify a remote file, and
//
//    http://filePath
// or
//    far:///filePath
//
// to identify a local file. The simple notation
//
//    filePath
//
// is also supported to indicate a local file.
//
PCFile CFile::NewFile(PCComm _comm, Pchar _userName, Pchar _URI) {
   char uri[sizeString];
   Strcpy(uri, _URI, sizeString);
   Trim(uri);

   char proto[sizeLine];
   proto[0] = 0;

   // Check for the far:// prefix
   long ofst = StrPos("far://", uri);
   // Check for an error
   if      (ofst >  0) {
#ifndef FV_NOLOG
      SysLog::Log('D', "Badly formed URI <%s>", _URI);
#endif
      return 0;
      }
   else if (ofst == 0)
      strcpy(proto, "far://");

   // Check for the http:// prefix
   ofst = StrPos("http://", uri);
   // Check for an error
   if      (ofst > 0) {
#ifndef FV_NOLOG
      SysLog::Log('D', "Badly formed URI <%s>", _URI);
#endif
      return 0;
      }
   else if (ofst == 0)
      strcpy(proto, "http://");

   // Check for a local filename (no protocol prefix)
   if (StrEmpty(proto))
      return new CFile(uri);

   // Check for a local filename (w/ protocol prefix)
   strcpy(uri, uri+strlen(proto));
   if (uri[0] == '/') {
      strcpy(uri, uri+1);
      Trim(uri);
      return new CFile(uri);
      }

   // Check that remote services are active
   if (!_comm) {
#ifndef FV_NOLOG
      SysLog::Log('D', "Can't open a remote file for\n"
                       "<%s>\n"
                       "farVIEW is not online",
                       _URI);
#endif
      return 0;
      }

   // Create a remote file handler object
   return new CRemoteFile(_comm, _userName, _URI);
   } // NewFile
#endif // FV_NOREMOTE

//---------------------------------------------------------------------
// Constructor common section
//
void CFile::construct(Pchar _name, Pchar _path) {
   memset(delim, 0, sizeName);
   Strcpy(delim, eolDelim, sizeName);

   Assert(!StrEmpty(_name));

#ifdef FV_MSWINDOWS
#ifndef _UTILITY
   // Temporary - Don't allow file names outside of the farVIEW subtree
//?Assert(_name[0] != '\\');
//?Assert(StrPos("..", _name) < 0);
#endif
#endif // FV_MSWINDOWS

#ifndef FV_NOTHREAD
   fileCS = new CCriticalSection();
#endif

   STRINIT(name, _name)
   ChrRep('/', pathDelim[0], name);                  // Unix -> Windows
   Squeeze(name, pathDelim[0]);

   // Check for standard in or standard out
   if      (stricmp(name, "stdin") == 0) {
      stdInOut   = stdIn;
      fileExists = true;
      return;
      }
   else if (stricmp(name, "stdout") == 0) {
      stdInOut   = stdOut;
      fileExists = false;
      return;
      }

   // process the file set parameter if requested
   if (StrEmpty(_path)) {
      // the filename includes the path,
      // else the file is in the cwd
      // Test for file existence
      ulong err;
      if (FileExists(name, err)) {
         // found the file
         fileExists = true;

         // Qualify the file name
         char curDir[sizeString];
         GetCurrentDir(curDir, sizeString);
         char work[sizeString];
         Strcpy(work, name, sizeString);
         CatName(curDir, work, 0, sizeString);
         delete[] name;
         name = new char[strlen(work)+1];
         strcpy(name, work);
         }
      }
   else {
#ifdef FV_MSWINDOWS
#ifndef _UTILITY
      // Temporary - Don't allow file names outside of the farVIEW
      // subtree
//?   Assert(_path[0] != '\\');
//?   Assert(StrPos(":\\", _path) < 0);
//?   Assert(StrPos("..",  _path) < 0);
#endif
#endif // FV_MSWINDOWS

      // prepare the path string
      PCRecord path = new CRecord(Pchar(""), Pchar(";"));
      path->PutField(1, Pchar("."));
      path->AppendField(_path);
      char pathName[sizeString];
      char work[sizeString];

      if (path->Number() == 1) {
         // path is specified
         path->GetField(1, pathName, sizeString);
         Strcpy(work, _name, sizeString);
         CatName(pathName, work, 0, sizeString);
         delete[] name;
         name = new char[strlen(work)+1];
         strcpy(name, work);
         if (name[0] == '"')
            strcpy(name, name+1);

         ulong err;
         ChrRep('/', pathDelim[0], name);
         if (FileExists(name, err))
            // found the file
            fileExists = true;
         }
      // traverse the path list (record)
      else for (short i = 1; i <= path->Number(); i++) {
         path->GetField(i, work, sizeString);
         Strcat(work, pathDelim, sizeString);// pathDelim in [":", "\"]
         Strcat(work, _name, sizeString); // form of dir is di:fileName
                                                      // or di\fileName
         ChrRep('/', pathDelim[0], work);
         ulong err;
         if (FileExists(work, err)) {
            // found the file
            fileExists = true;
            delete[] name;
            name = new char[strlen(work)+1];
            strcpy(name, work);
            break;
            }
         }
      }
   } // construct

//---------------------------------------------------------------------
// Constructor
//
CFile::CFile(void):
   CAtom       (),
   name        (0),
   stdInOut    (notStd),
   share       (false),
#ifndef FV_NOTHREAD
   fileCS      (),
#endif
   fileExists  (false),
   fileIsOpen  (false),
   fileCanWrite(false),
   file        (0),
   whereAt     (0),
   sysErr      (FVE_OK) {

#ifndef FV_NOTHREAD
   fileCS = new CCriticalSection();
#endif

   memset(delim, 0, sizeName);
   Strcpy(delim, eolDelim, sizeName);
   } // CFile

//---------------------------------------------------------------------
// Constructor
//
CFile::CFile(const Pchar _name):
   CAtom       (),
   name        (0),
   stdInOut    (notStd),
   share       (false),
#ifndef FV_NOTHREAD
   fileCS      (0),
#endif
   fileExists  (false),
   fileIsOpen  (false),
   fileCanWrite(false),
   file        (0),
   whereAt     (0),
   sysErr      (FVE_OK) {

   construct(_name, Pchar(""));
   } // CFile

//---------------------------------------------------------------------
// Constructor
//
CFile::CFile(const Pchar _name, const Pchar _path, bool _share):
   CAtom       (),
   name        (0),
   stdInOut    (notStd),
   share       (_share),
#ifndef FV_NOTHREAD
   fileCS      (0),
#endif
   fileExists  (false),
   fileIsOpen  (false),
   fileCanWrite(false),
   file        (0),
   whereAt     (0),
   sysErr      (FVE_OK) {

   construct(_name, _path);
   } // CFile

//---------------------------------------------------------------------
// Copy Constructor
//
CFile::CFile(PCFile _file):
   CAtom       (),
   name        (0),
   stdInOut    (_file->stdInOut),
   share       (_file->share),
#ifndef FV_NOTHREAD
   fileCS      (),
#endif
   fileExists  (_file->fileExists),
   fileIsOpen  (_file->fileIsOpen),
   fileCanWrite(_file->fileCanWrite),
   file        (_file->file),
   whereAt     (_file->whereAt),
   sysErr      (FVE_OK) {

#ifndef FV_NOTHREAD
   fileCS = _file->fileCS;
#endif

   memset(delim, 0, sizeName);
   Strcpy(delim, eolDelim, sizeName);

   STRINIT(name, _file->name);
   } // CFile

//---------------------------------------------------------------------
// Destructor
//
CFile::~CFile() {
   if (fileIsOpen)
      Close();

   delete[] name;
   name = 0;

   file = 0;
   } // ~CFile

//---------------------------------------------------------------------
//
RTTI_CPP(File, Atom)

//---------------------------------------------------------------------
// Used internally by CRemoteFile. _path is always set to the cache
// directory.
//
void CFile::setupFile(Pchar _name, Pchar _path) {
   STRINIT(name, _name)

   // prepare the path string
   PCRecord path = new CRecord(_path, Pchar(";"));
   char pathName[sizeString];
   char work[sizeString];

   if (path->Number() == 1) {
      // path is specified
      path->GetField(1, pathName, sizeString);
      Strcpy(work, _name, sizeString);
      CatName(pathName, work, 0, sizeString);
      STRINIT(name, work)
      if (name[0] == '"')
         strcpy(name, name+1);

      ulong err;
      if (FileExists(name, err))
         // found the file
         fileExists = true;
      }
   // traverse the path list (record)
   else for (short i = 1; i <= path->Number(); i++) {
      path->GetField(i, work, sizeString);
      Strcat(work, pathDelim, sizeString);// pathDelim in [":", "\"]
      Strcat(work, _name, sizeString); // form of dir is di:fileName
                                                   // or di\fileName
      ulong err;
      if (FileExists(work, err)) {
         // found the file
         fileExists = true;
         STRINIT(name, work)
         break;
         }
      }
   } // setupFile

//---------------------------------------------------------------------
// Return true if at end of file
//
bool CFile::atEOF(void) {
   return err == FVE_ENDOFFILE || Size() - Where() <= 0;
   } // atEOF

//---------------------------------------------------------------------
// Set the error state of the file
//
void CFile::SetError(ulong _err) {
   err = _err;
   if (err != FVE_OK &&
       err != FVE_ENDOFFILE)
#ifndef FV_NOLOG
      SysLog::Log('B',
                  Pchar("CFile.Error: %s (%ld)"),
                  name,
                  err);
      #else
      ;
#endif
   } // SetError

//---------------------------------------------------------------------
//
ulong CFile::Error(void) {
   ulong error = err;
   err = FVE_OK;
   
   return error;
   } // Error

//---------------------------------------------------------------------
//
void CFile::Error(ulong _err) {
   SetError(_err);
   } // Error

//---------------------------------------------------------------------
//
ulong CFile::logSysErr(void) {
#if defined(FV_MSWIN32)
   LPVOID lpMsgBuf;

   DWORD err = ::GetLastError();
   ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_FROM_SYSTEM,
                   0,
                   err,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                   (LPTSTR) &lpMsgBuf,
                   0,
                   0);
#ifndef FV_NOLOG
   SysLog::Log('B', Pchar(lpMsgBuf));
#endif
   ::LocalFree(lpMsgBuf);
   return err;
#endif
   } // logSysErr

//---------------------------------------------------------------------
// ***Warning: Allocates memory that must be deleted by caller***
//
Pchar CFile::Path(void) {
   char work[sizeFileName];
   work[0] = 0;
   GetFilePath(work, name, sizeFileName);
   Pchar path = new char[strlen(work)+1];
   strcpy(path, work);
   
   return path;
   } // Path

//---------------------------------------------------------------------
// ***Warning: Allocates memory that must be deleted by caller***
//
Pchar CFile::Name(bool _alc) {
   if (_alc) {
      Pchar work = new char[strlen(name)+1];
      strcpy(work, name);
      return work;
      }

   return name;
   } // Name

//---------------------------------------------------------------------
// farSlang version
//
Pchar CFile::Name(void) {
   return Name(true);
   } // Name

//---------------------------------------------------------------------
// Set the inter-string delimiter.
// (Note: set delim to empty to write
// C-style strings to the file.)
//
void CFile::SetEOLDelim(const Pchar _delim) {
   Assert(_delim);
   Strcpy(delim, _delim, sizeName);
   } // SetEOLDelim

//---------------------------------------------------------------------
// Return true if the file exists
//
bool CFile::Exists(void) {
   return fileExists;
   } // Exists

//---------------------------------------------------------------------
// Rename the file. Return true if the file is renamed.
// note: should process the path in _name since the
// file can be moved with this command. Should change the
// name to Move. TBD.
bool CFile::Rename(Pchar _name) {

   if (stdInOut != notStd)
      return false;

#if   defined(FV_MSWINDOWS)
   if (fileExists) {
#ifndef FV_NOTHREAD
      fileCS->Enter();
#endif
      if (!::MoveFile(name, _name)) {
         sysErr = logSysErr();
         SetError(FVE_CANTRENAMEFILE);
         }
      else {
         sysErr = FVE_OK;
         err    = FVE_OK;
         delete[] name;
         name   = new char[strlen(_name)+1];
         strcpy(name, _name);
         }
#ifndef FV_NOTHREAD
      fileCS->Leave();
#endif
      }
   else
      SetError(FVE_NOFILE);

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   Assert(false);

#endif // FV_UNIX

   return (err == FVE_OK);
   } // Rename

//---------------------------------------------------------------------
// Backup the file. TBD
//
bool CFile::Backup(void) {
   return true;
   } // Backup

//---------------------------------------------------------------------
// Delete the file. Return true if the file is deleted.
//
bool CFile::Delete(void) {
   Assert(!fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return true;

#if   defined(FV_MSWINDOWS)
   if (fileExists) {
#ifndef FV_NOTHREAD
      fileCS->Enter();
#endif
      fileExists = false;
      if (!::DeleteFile(name)) {
         sysErr = logSysErr();
         SetError(FVE_CANTREMOVEFILE);
         }
#ifndef FV_NOTHREAD
      fileCS->Leave();
#endif
      }
   else
      SetError(FVE_NOFILE);

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   Assert(false);

#endif // FV_UNIX

   return (err == FVE_OK);
   } // Delete

//---------------------------------------------------------------------
// Create a closed file
//
bool CFile::Create(void) {
   Assert(!fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   Assert(stdInOut == notStd);

   if (OpenOutput())
      Close();
   else
      SetError(FVE_FILEISOPEN);

   return (err == FVE_OK);
   } // Create

//---------------------------------------------------------------------
// Open the file as input
//
bool CFile::OpenInput(void) {
#ifndef FV_NOTHREAD
   fileCS->Enter();
#endif

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut == stdOut) {
      SetError(FVE_FILENOTINPUT);
      return false;
      }

   if (!fileExists) {
      SetError(FVE_NOFILE);
      return false;
      }
   Assert(!fileIsOpen);

   fileCanWrite = false;

   if (stdInOut == stdIn) {
     fileIsOpen = true;
     return true;
     }

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(::CreateFile(name,
                                 GENERIC_READ,
                                 FILE_SHARE_READ,
                                 &security,
                                 OPEN_EXISTING,
                                 FILE_ATTRIBUTE_ARCHIVE |
                                 FILE_FLAG_RANDOM_ACCESS,
                                 0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      sysErr = logSysErr();
      SetError(FVE_BADOPEN);
      file   = 0;
      return false;
      }

#elif defined(FV_MSWIN16)
   file = FB_HANDLE(sopen(name, O_RDONLY | O_BINARY, SH_DENYWR, S_IREAD | S_IWRITE));
   if (file <= 0) {
      switch (errno) {
         case EACCES:  sysErr = FVE_NOACCESS;      // permission denied
                       break;
         case EINVACC: sysErr = FVE_BADACCESS;   // invalid access code
                       break;
         case EMFILE:  sysErr = FVE_NOFILES;     // too many open files
                       break;
         case ENOENT:  sysErr = FVE_NOFILE;           // file not found
                       break;
         default:      sysErr = errno;          // error not recognized
         }
      SetError(FVE_BADOPEN);
      return false;
      }

#elif defined(FV_MACINTOSH)
   char pName[sizeString];
   Strcpy(pName, name, sizeString);
   CtoPstr(pName);

   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));

   ioParam.ioNamePtr = (unsigned char*)pName;
   ioParam.ioPermssn = fsRdWrShPerm;
   sysErr            = long(::PBOpenSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_BADOPEN);
      file  = 0;
      return false;
      }
   file = FB_HANDLE(ioParam.ioRefNum);

#elif defined(FV_UNIX)
   file = FB_HANDLE(::open(name, O_RDONLY));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      err    = FVE_BADOPEN;
      return false;
      }

#endif // FV_UNIX

   fileIsOpen = true;
   return true;
   } // OpenInput

//---------------------------------------------------------------------
// Open the file as output, destroying any existing file
//
bool CFile::OpenOutput(void) {
#ifndef FV_NOTHREAD
   fileCS->Enter();
#endif

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut == stdIn) {
      SetError(FVE_FILENOTOUTPUT);
      return false;
      }

   Assert(!fileIsOpen);

   fileCanWrite = true;

   if (stdInOut == stdOut) {
     fileIsOpen = true;
     return true;
     }

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(::CreateFile(name,
                                 GENERIC_WRITE,
                                 0,
                                 &security,
                                 CREATE_ALWAYS,
                                 FILE_ATTRIBUTE_ARCHIVE |
                                 FILE_FLAG_RANDOM_ACCESS,
                                 0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      sysErr = logSysErr();
      SetError(FVE_BADOPEN);
      file   = 0;
      return false;
      }
#elif defined(FV_MSWIN16)
   file = FB_HANDLE(sopen(name, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, SH_DENYWR, S_IREAD | S_IWRITE));
   if (file <= 0) {
      switch (errno) {
         case EACCES:  sysErr = FVE_NOACCESS;      // permission denied
                       break;
         case EINVACC: sysErr = FVE_BADACCESS;   // invalid access code
                       break;
         case EMFILE:  sysErr = FVE_NOFILES;     // too many open files
                       break;
         case ENOENT:  sysErr = FVE_NOFILE;           // file not found
                       break;
         default:      sysErr = errno;          // error not recognized
         }
      SetError(FVE_BADOPEN);
      return false;
      }

#elif defined(FV_MACINTOSH)
   // Create the file deleting any old file
   FILE* f = fopen(name, Pchar("wb"));
   if (f == 0) {
      sysErr = FVE_BADOPEN;
      SetError(FVE_BADOPEN);
      return false;
      }
   fclose(f);

   // Open the new file for output
   char pName[sizeString];
   Strcpy(pName, name, sizeString);
   CtoPstr(pName);

   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioNamePtr = (unsigned char*)pName;
   ioParam.ioPermssn = fsWrPerm;
   sysErr            = long(::PBOpenSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_BADOPEN);
      file  = 0;
      return false;
      }
   file = FB_HANDLE(ioParam.ioRefNum);

#elif defined(FV_UNIX)
   file = FB_HANDLE(::open(name, O_CREAT+O_TRUNC+O_WRONLY, 0770));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      err    = FVE_BADOPEN;
      return false;
      }

#endif // FV_UNIX

   fileIsOpen = true;
   return true;
   } // OpenOutput

//---------------------------------------------------------------------
// Open the file with the file pointer at the end of the file
//
bool CFile::OpenAppend(void) {
#ifndef FV_NOTHREAD
   fileCS->Enter();
#endif

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut == stdIn) {
      SetError(FVE_FILENOTOUTPUT);
      return false;
      }

   Assert(!fileIsOpen);

   fileCanWrite = true;

   if (stdInOut == stdOut) {
     fileIsOpen = true;
     return true;
     }

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(::CreateFile(name,
                                 GENERIC_WRITE,
                                 0,
                                 &security,
                                 OPEN_ALWAYS,
                                 FILE_ATTRIBUTE_ARCHIVE |
                                 FILE_FLAG_RANDOM_ACCESS,
                                 0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      sysErr = logSysErr();
      SetError(FVE_BADOPEN);
      file   = 0;
      return false;
      }

#elif defined(FV_MSWIN16)
   file = FB_HANDLE(sopen(name, O_WRONLY | O_CREAT | O_BINARY, SH_DENYWR, S_IREAD | S_IWRITE));
   if (file <= 0) {
      switch (errno) {
         case EACCES:  sysErr = FVE_NOACCESS;      // permission denied
                       break;
         case EINVACC: sysErr = FVE_BADACCESS;   // invalid access code
                       break;
         case EMFILE:  sysErr = FVE_NOFILES;     // too many open files
                       break;
         case ENOENT:  sysErr = FVE_NOFILE;           // file not found
                       break;
         default:      sysErr = errno;          // error not recognized
         }
      SetError(FVE_BADOPEN);
      return false;
      }

#elif defined(FV_MACINTOSH)
   char pName[sizeString];
   Strcpy(pName, name, sizeString);
   CtoPstr(pName);

   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioNamePtr = (unsigned char*)pName;
   ioParam.ioPermssn = fsWrPerm;
   sysErr            = long(::PBOpenSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_BADOPEN);
      file  = 0;
      return false;
      }
   file = FB_HANDLE(ioParam.ioRefNum);

#elif defined(FV_UNIX)
   file = FB_HANDLE(::open(name, O_CREAT+O_WRONLY+O_APPEND, 0770));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      err    = FVE_BADOPEN;
      return false;
      }

#endif // FV_UNIX

   fileIsOpen = true;
   Seek(0L, fromTail);
   return true;
   } // OpenAppend

//---------------------------------------------------------------------
// Open the file for reading and writing, placing the file
// pointer at the beginning of the file
//
bool CFile::OpenUpdate(void) {
#ifndef FV_NOTHREAD
   fileCS->Enter();
#endif

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut == stdIn) {
      SetError(FVE_FILENOTOUTPUT);
      return false;
      }

   Assert(!fileIsOpen);

   Assert(stdInOut == notStd);

   fileCanWrite = true;

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(::CreateFile(name,
                                 GENERIC_READ            |
                                 GENERIC_WRITE,
                                 FILE_SHARE_READ         |
                                 FILE_SHARE_WRITE,
                                 &security,
                                 OPEN_ALWAYS,
                                 FILE_ATTRIBUTE_ARCHIVE  |
                                 FILE_FLAG_WRITE_THROUGH |
                                 FILE_FLAG_RANDOM_ACCESS,
                                 0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      sysErr = logSysErr();
      SetError(FVE_BADOPEN);
      file   = 0;
      return false;
      }
#elif defined(FV_MSWIN16)
   file = FB_HANDLE(sopen(name, O_RDWR | O_CREAT | O_BINARY, SH_DENYWR, S_IREAD | S_IWRITE));
   if (file <= 0) {
      switch (errno) {
         case EACCES:  sysErr = FVE_NOACCESS;      // permission denied
                       break;
         case EINVACC: sysErr = FVE_BADACCESS;   // invalid access code
                       break;
         case EMFILE:  sysErr = FVE_NOFILES;     // too many open files
                       break;
         case ENOENT:  sysErr = FVE_NOFILE;           // file not found
                       break;
         default:      sysErr = errno;          // error not recognized
         }
      SetError(FVE_BADOPEN);
      return false;
      }

#elif defined(FV_MACINTOSH)
   char pName[sizeString];
   strcpy(pName, name);
   CtoPstr(pName);

   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioNamePtr = (unsigned char*)pName;
   ioParam.ioPermssn = fsRdWrShPerm;
   sysErr            = long(::PBOpenSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_BADOPEN);
      file  = 0;
      return false;
      }
   file = FB_HANDLE(ioParam.ioRefNum);

#elif defined(FV_UNIX)
   file = FB_HANDLE(::open(name, O_CREAT+O_RDWR, 0770));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      err    = FVE_BADOPEN;
      return false;
      }

#endif // FV_UNIX

   fileIsOpen = true;
   return true;
   } // OpenUpdate

//---------------------------------------------------------------------
// Close the file
//
bool CFile::Close(void) {
   sysErr = FVE_OK;
   err    = FVE_OK;

   Assert(fileIsOpen);

   if (stdInOut != notStd)
      return true;

#if   defined(FV_MSWIN32)
   if (!CloseHandle(HANDLE(file))) {
      sysErr     = logSysErr();
      SetError(FVE_BADCLOSE);
      }
   else
      fileIsOpen = false;

#elif defined(FV_MSWIN16)
   close(int(file));
   fileIsOpen = false;

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum = short(file);
   sysErr           = long(::PBCloseSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr)
      SetError(FVE_BADCLOSE);
   else
      fileIsOpen = false;

#elif defined(FV_UNIX)
   ::close(int(file));

#endif // FV_UNIX

   file = 0;
#ifndef FV_NOTHREAD
   fileCS->Leave();
#endif

   return (err == FVE_OK);
   } // Close

//---------------------------------------------------------------------
// Return record describing current file.
// Form is:
//   |filename
//   |crdate|crtime|crsecs
//   |moddate|modtime|modsecs
//   |size  + |dir/dat| tbd
//
void CFile::GetFileData(PCRecord _data) {
   Assert(fileIsOpen);
   _data->NewRecord(SRF_NRFIELDS, Pchar("|"));

   if      (stdInOut == stdIn) {
      _data->PutField(SRF_NAME, Pchar("stdin"));
      return;
      }
   else if (stdInOut == stdOut) {
     _data->PutField(SRF_NAME, Pchar("stdout"));
      return;
      }

#if   defined(FV_MSWINDOWS)
   // get the file name
   _data->PutField(SRF_NAME, name);

   FILETIME createTime, lastWrite, localTime;

   ::GetFileTime(HANDLE(file),                   // identifies the file
                 &createTime,               // address of creation time
                 0,	                    // address of last access time
                 &lastWrite);             // address of last write time

   WORD date = 0;
   WORD time = 0;

   // get the create date and time
   ::FileTimeToLocalFileTime(&createTime, &localTime);
   ::FileTimeToDosDateTime(&localTime, &date, &time);

   char swork[32];

   // convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_CRDATE, swork);

   // convert time to hh:mm:ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                               // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_CRTIME, swork);

   LARGE_INTEGER secs;
   memcpy(&secs, &localTime, sizeof(secs));
   _data->PutField(SRF_CRSECS, secs.QuadPart);

   // get the modify date and time
   ::FileTimeToLocalFileTime(&lastWrite, &localTime);
   ::FileTimeToDosDateTime(&localTime, &date, &time);

   // convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_MODDATE, swork);

   // convert time to hh/mm/ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                                // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_MODTIME, swork);

   memcpy(&secs, &localTime, sizeof(secs));
   _data->PutField(SRF_MODSECS, secs.QuadPart);

   // note that size is assumed to be less than 2 ^ 32
   _data->PutField(SRF_SIZE, long(GetFileSize(HANDLE(file), 0)));

#elif defined(FV_MACINTOSH)
   HFileInfo *fpb = (HFileInfo*) &cipbr;
   Assert(fpb->ioFDirIndex);

   // Get the file name
   _data->PutField(SRF_NAME, Pchar(fileName));

   DateTimeRec dt;

   // Get the create date and time
   ::SecondsToDate(fpb->ioFlCrDat, &dt);

   char swork[32];

   // Convert date to mm/dd/yyyy format
   DateStr(dt.year + 1904,
           dt.month,
           dt.day, swork);
   _data->PutField(SRF_CRDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr(dt.hour,
           dt.minute,
           dt.second, swork);
   _data->PutField(SRF_CRTIME, swork);

   _data->PutField(SRF_CRSECS, fpb->ioFlCrDat);

   // Get the modify date and time
   ::SecondsToDate(fpb->ioFlMdDat, &dt);

   // Convert date to mm/dd/yyyy format
   DateStr(dt.year + 1904,
           dt.month,
           dt.day, swork);
   _data->PutField(SRF_MODDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr(dt.hour,
           dt.minute,
           dt.second, swork);
   _data->PutField(SRF_MODTIME, swork);

   _data->PutField(SRF_MODSECS, fpb->ioFlMdDat);

   // Get the file size
   _data->PutField(SRF_SIZE, fpb->ioFlLgLen);

#elif defined(FV_UNIX)
   #warning FV_UNIX TBD (CFile::GetFileData)
/*
   // Get the file name
   _data->PutField(SRF_NAME, entry.d_name);

   PCRecord dt;
   char work[sizeLine];

   // Get the create date and time
   dt = SecsToDateAndTime(buff.st_ctime);
   dt->GetField   (1, work, sizeLine);
   _data->PutField(SRF_CRDATE, work);
   dt->GetField   (2, work, sizeLine);
   _data->PutField(SRF_CRTIME, work);
   _data->PutField(SRF_CRSECS, buff.st_ctime);

   // Get the modify date and time
   dt = SecsToDateAndTime(buff.st_mtime);
   dt->GetField   (short(1), work, sizeLine);
   _data->PutField(SRF_MODDATE, work);
   dt->GetField   (short(2), work, sizeLine);
   _data->PutField(SRF_MODTIME, work);
   _data->PutField(SRF_MODSECS, buff.st_mtime);

   // Get the file size
   _data->PutField(SRF_SIZE, buff.st_size);
*/
#endif // FV_UNIX

   } // GetFileData

//---------------------------------------------------------------------
// Lock the specified file region
// (handles files up to 2^32-1 bytes)
//
bool CFile::Lock(ulong _offset, ulong _length) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return false;

   bool locked = true;

#if   defined(FV_MSWIN32)
   if (!::LockFile(HANDLE(file), _offset, 0L, _length, 0L)) {
      sysErr = logSysErr();
//?   SetError(FVE_LOCKFAILED);
      locked = false;
      }

#elif defined(FV_MSWIN16)
   #ifdef __BORLANDC__
      if (::lock(int(file), _offset, _length)) {
      sysErr = errno;
//?   SetError(FVE_LOCKFAILED);
      locked = false;
      }
   #else
      long where = Where();
      if (Seek(_offset)) {
         if (_locking(int(file), _LK_LOCK, _length)) {
            sysErr = errno;
//?         SetError(FVE_LOCKFAILED);
            locked = false;
            }
         }
      else {
         sysErr = errno;
//?      SetError(FVE_LOCKFAILED);
         locked = false;
         }
      Seek(where);
   #endif

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum    = short(file);
   ioParam.ioReqCount  = _length;
   ioParam.ioPosMode   = fsFromStart;
   ioParam.ioPosOffset = _offset;
   sysErr              = long(::PBLockRangeSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_LOCKFAILED);
      locked = false;
      }

#elif defined(FV_UNIX)
   flock flk;
   flk.l_type   = F_WRLCK;
   flk.l_start  = _offset;
   flk.l_whence = SEEK_SET;
   flk.l_len    = _length;
   flk.l_pid    = 0;

   errno = 0;
   if (::fcntl(int(file), F_SETLK, &flk)) {
      if (errno)
         sysErr = errno;
//?   SetError(FVE_LOCKFAILED);
      locked = false;
      }

#endif // FV_UNIX

   return locked;
   } // Lock

//---------------------------------------------------------------------
// Unlock the specified file region
// (handles files up to 2^32 bytes)
//
bool CFile::Unlock(ulong _offset, ulong _length) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return false;

#if   defined(FV_MSWIN32)
   if (!::UnlockFile(HANDLE(file), _offset, 0L, _length, 0L)) {
      sysErr = logSysErr();
      SetError(FVE_UNLOCKFAILED);
      }

#elif defined(FV_MSWIN16)
   #ifdef __BORLANDC__
      if (::unlock(int(file), _offset, _length)) {
      sysErr = errno;
      SetError(FVE_UNLOCKFAILED);
      }
   #else
      long where = Where();
      if (Seek(_offset)) {
         if (_locking(int(file), _LK_UNLCK, _length)) {
            sysErr = errno;
            SetError(FVE_LOCKFAILED);
            }
         }
      else {
         sysErr = errno;
         SetError(FVE_LOCKFAILED);
         }
      Seek(where);
   #endif

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum    = short(file);
   ioParam.ioReqCount  = _length;
   ioParam.ioPosMode   = fsFromStart;
   ioParam.ioPosOffset = _offset;
   sysErr              = long(::PBUnlockRangeSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr)
      SetError(FVE_UNLOCKFAILED);

#elif defined(FV_UNIX)
   flock flk;
   flk.l_type   = F_UNLCK;
   flk.l_start  = _offset;
   flk.l_whence = SEEK_SET;
   flk.l_len    = _length;
   flk.l_pid    = 0;

   errno = 0;
   if (::fcntl(int(file), F_SETLK, &flk)) {
      if (errno)
         sysErr = errno;
      err = FVE_LOCKFAILED;
      }

#endif // FV_UNIX
   return (err == FVE_OK);
   } // Unlock

//---------------------------------------------------------------------
// Return the (current) size of the file
// (If return value < 0, then an error occured)
//
ulong CFile::Size(void) {
   Assert(fileIsOpen);

   if (stdInOut != notStd)
      return 0;

   ulong where = whereAt;
   Seek(0,     fromTail);
   ulong size  = whereAt;
   Seek(where, fromHead);
   return size;
   } // Size

//---------------------------------------------------------------------
// Seek relative to the beginning, current or end of the file
//
bool CFile::Seek(ulong _offset, EBase _from) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return true;

   switch (_from) {
#if   defined(FV_MSWIN32)
      case fromHead:
         whereAt = ::SetFilePointer(HANDLE(file), _offset, 0, FILE_BEGIN);
         if (whereAt < 0)
            sysErr = logSysErr();
         break;
      case fromCurr:
         whereAt = ::SetFilePointer(HANDLE(file), _offset, 0, FILE_CURRENT);
         if (whereAt < 0)
            sysErr = logSysErr();
         break;
      case fromTail:
         whereAt = ::SetFilePointer(HANDLE(file), _offset, 0, FILE_END);
         if (whereAt < 0)
            sysErr = logSysErr();
         break;

#elif defined(FV_MSWIN16)
      case fromHead:
         whereAt = ::lseek(int(file), _offset, SEEK_SET);
         if (whereAt < 0)
            sysErr = errno;

         break;
      case fromCurr:
         whereAt = ::lseek(int(file), _offset, SEEK_CUR);
         if (whereAt < 0)
            sysErr = errno;
         break;
      case fromTail:
         whereAt = ::lseek(int(file), _offset, SEEK_END);
         if (whereAt < 0)
            sysErr = errno;
         break;

#elif defined(FV_MACINTOSH)
      case fromHead:
         sysErr = ulong(::SetFPos(short(file), fsFromStart, _offset));
         if (OSErr(sysErr) != noErr)
            SetError(FVE_BADSEEK);
         else
            Where();
         break;
      case fromCurr:
         sysErr = ulong(::SetFPos(short(file), fsFromMark, _offset));
         if (OSErr(sysErr) != noErr)
            SetError(FVE_BADSEEK);
         else
            Where();
         break;
      case fromTail:
         sysErr = ulong(::SetFPos(short(file), fsFromLEOF, _offset));
         if (OSErr(sysErr) != noErr)
            SetError(FVE_BADSEEK);
         else
            Where();
         break;

#elif defined(FV_UNIX)
      case fromHead:
         whereAt = ::lseek(int(file), _offset, 0);
         if (whereAt < 0)
            sysErr = errno;
         break;
      case fromCurr:
         whereAt = ::lseek(int(file), _offset, 1);
         if (whereAt < 0)
            sysErr = errno;
         break;
      case fromTail:
         whereAt = ::lseek(int(file), _offset, 2);
         if (whereAt < 0)
            sysErr = errno;
         break;

#endif // FV_UNIX

      default:
         Assert(false);
      } // switch

   if (whereAt < 0)
      return false;
   else
      return true;
   } // Seek

//---------------------------------------------------------------------
// Return the value of the file pointer
// (If return value < 0, then an error occured)
//
ulong CFile::Where(void) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return 0;

#if   defined(FV_MSWIN32)
   whereAt = ::SetFilePointer(HANDLE(file), 0, 0, FILE_CURRENT);
   if (whereAt < 0) {
      sysErr = logSysErr();
      SetError(FVE_BADSEEK);
      }

#elif defined(FV_MSWIN16)
   whereAt = ::lseek(int(file), 0, SEEK_CUR);
   if (whereAt < 0) {
      sysErr = errno;
      SetError(FVE_BADSEEK);
      }

#elif defined(FV_MACINTOSH)
   sysErr = long(::GetFPos(short(file), &whereAt));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_BADSEEK);
      whereAt = -1;
      }

#elif defined(FV_UNIX)
   whereAt = ::lseek(int(file), 0, 1);
   if (whereAt < 0) {
      sysErr = errno;
      err    = FVE_BADSEEK;
      }

#endif // FV_UNIX
   return whereAt;
   } // Where

//---------------------------------------------------------------------
// Flush new data to the file
//---------------------------------------------------------------------
bool CFile::Flush(void) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (stdInOut != notStd)
      return true;

#if   defined(FV_MSWIN32)
   if (!::FlushFileBuffers(HANDLE(file))) {
      sysErr = logSysErr();
      SetError(FVE_FLUSHFAILED);
      return false;
      }

#elif defined(FV_MSWIN16)
   flushall();

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum = short(file);
   sysErr           = long(::PBFlushFileSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr) {
      SetError(FVE_FLUSHFAILED);
      return false;
      }

#elif defined(FV_UNIX)
   Assert(false);

#endif // FV_UNIX

   return true;
   } // Flush

//---------------------------------------------------------------------
// Read a character and return it
//
char CFile::ReadChar(bool _upCase) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if      (!read(&ch, 1))
      return atEOL;
   else if (ch == cr || ch == lf) {
      read(&ch, 1);
      if (ch != lf && ch != cr)
         Seek(-1, fromCurr);
      return atEOL;
      }
   return (_upCase ? CapChr(ch) : ch);

   } // ReadChar

//---------------------------------------------------------------------
// Read a string into _p (farSlang function - do not use in C++)
//
const char DOS_EOF = 0x1a;

Pchar CFile::ReadString(void) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   long where = CFile::Where();
   char buffer[sizeString];
   memset(buffer, 0, sizeString);
   if (!read(&buffer, sizeString-1, 0, true) && err != FVE_SHORTREAD)
      return 0;

#if   defined(FV_MACINTOSH)
   long ief    = ChrPos(DOS_EOF, &buffer);
   if      (ief == 0) {
      SetError(FVE_ENDOFFILE);
      return 0;
      }
   else if (ief >  0)
      buffer[size_t(ief)] = '\0';

#elif defined(FV_UNIX)

#endif // FV_UNIX
   long endStr = 0;
   long icr    = ChrPos(cr, buffer);
   long ilf    = ChrPos(lf, buffer);
   if      (icr >= 0 && (icr < ilf || ilf < 0)) {
      ++endStr;
      buffer[size_t(icr)] = 0;
      if (buffer[size_t(icr+1)] == lf)
         ++endStr;
      }
   else if (ilf >= 0 && (ilf < icr || icr < 0)) {
      ++endStr;
      buffer[size_t(ilf)] = 0;
      if (buffer[size_t(ilf+1)] == cr)
         ++endStr;
      }
   else                                // handle zero-terminated string
      endStr = 1;

   CFile::Seek(where+strlen(buffer)+endStr, fromHead);

   Pchar p = new char[strlen(buffer)+1];
   strcpy(p, buffer);
   return p;
   } // ReadString

//---------------------------------------------------------------------
// Read a string into _p
//
bool CFile::ReadString(Pchar _p, ulong _size) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;
   memset(_p, 0, size_t(_size));

   long where = CFile::Where();
   if (err != FVE_OK)
      return false;
   if (!read(_p, _size-1, 0, true) && err != FVE_SHORTREAD)
      return false;

#if   defined(FV_MACINTOSH)
   long ief    = ChrPos(DOS_EOF, _p);
   if      (ief == 0) {
      SetError(FVE_ENDOFFILE);
      return false;
      }
   else if (ief >  0)
      _p[size_t(ief)] = '\0';

#elif defined(FV_UNIX)

#endif // FV_UNIX
   long endStr = 0;
   long icr    = ChrPos(cr, _p);
   long ilf    = ChrPos(lf, _p);
   if      (icr >= 0 && (icr < ilf || ilf < 0)) {
      ++endStr;
      _p[size_t(icr)] = 0;
      if (_p[size_t(icr+1)] == lf)
         ++endStr;
      }
   else if (ilf >= 0 && (ilf < icr || icr < 0)) {
      ++endStr;
      _p[size_t(ilf)] = 0;
      if (_p[size_t(ilf+1)] == cr)
         ++endStr;
      }
   else                                // handle zero-terminated string
      endStr = 1;

   ulong herr = err;
   CFile::Seek(where+strlen(_p)+endStr, fromHead);
   err = herr;
   return (err == FVE_OK);
   } // ReadString

//---------------------------------------------------------------------
// Read a block into _p
//
bool CFile::ReadBlock(Pvoid _p, ulong _size, Pulong _asize) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   bool ok = read(_p, _size, _asize);

   ulong herr = err;
   CFile::Where();
   err = herr;

   return ok;
   } // ReadBlock

//---------------------------------------------------------------------
// Write a character from _p
//
bool CFile::WriteChar(char _p) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (_p == atEOL)
      if (strlen(delim))
         return write(delim, strlen(delim));
      else
         return write(delim, 1);
   else
      return (write(&_p, 1));
   } // WriteChar

//---------------------------------------------------------------------
// Write a string from _p
//
bool CFile::Printf(Pchar _format, ...) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (!StrEmpty(_format)) {
      // Print formatted message
      va_list args;
      va_start(args, _format);
      char line[sizeString];
      vsprintf(line, _format, args);
      va_end(args);
      Assert(strlen(line) < sizeString);
      return WriteString(line);
      }

   return false;
   } // Printf

//---------------------------------------------------------------------
// Write a string from _p
//
bool CFile::WriteString(Pchar _p) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   if (strlen(delim)) {
      if (write(_p, strlen(_p)))
         return write(delim, strlen(delim));
      else
         return false;
      }
   else
      // Include the trailing zero in the output
      return write(_p, strlen(_p)+1);
   } // WriteString

//---------------------------------------------------------------------
// Write a substring from _p
// (i.e., no end of line characters)
//
bool CFile::WriteSubStr(Pchar _p) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   return write(_p, strlen(_p));
   } // WriteSubStr

//---------------------------------------------------------------------
// Write a block from _p
//
bool CFile::WriteBlock(Pvoid _p, ulong _size) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   err    = FVE_OK;

   return write(_p, _size);
   } // WriteBlock

//---------------------------------------------------------------------
// Read a block from the file
//---------------------------------------------------------------------
bool CFile::read(Pvoid  _buffer,
                 ulong  _size,
                 Pulong _asize,
                 bool   _shortOk) {

   // Same special handling for all platforms if standard input
   if (stdInOut == stdIn) {
      size_t size = fread(_buffer, _size, 1, stdin);
      if      (size == 0) {
         sysErr = FVE_ENDOFFILE;
         SetError(FVE_ENDOFFILE);
         return false;
         }
      else if (!_shortOk && size < _size) {
         sysErr = FVE_ENDOFFILE;
         SetError(FVE_SHORTREAD);
         return false;
         }
      return true;
      }

#if defined(FV_MSWIN32)
   DWORD size;
   if      (!::ReadFile(HANDLE(file), _buffer, _size, &size, 0)) {
      sysErr = ::GetLastError();
      SetError(FVE_BADREAD);
      }
   else if (size == 0)
      SetError(FVE_ENDOFFILE);
   else if (!_shortOk && size < _size)
      SetError(FVE_SHORTREAD);
   if (_asize)
      *_asize  = size;

#elif defined(FV_MSWIN16)
   ulong size = ::read(int(file), _buffer, size_t(_size));
   if      (size == 0)
      SetError(FVE_ENDOFFILE);
   else if (!_shortOk && size < _size)
      SetError(FVE_SHORTREAD);
   if (_asize)
      *_asize = size;

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum    = short(file);
   ioParam.ioBuffer    = Ptr(_buffer);
   ioParam.ioReqCount  = _size;
   ioParam.ioPosMode   = fsFromMark;
   sysErr              = long(::PBReadSync(ParmBlkPtr(&ioParam)));
   if      (ioParam.ioActCount == 0)
      SetError(FVE_ENDOFFILE);
   else if (!_shortOk && ioParam.ioActCount < _size)
      SetError(FVE_SHORTREAD);
   else if (OSErr(sysErr) != noErr)
      SetError(FVE_BADREAD);
   if (_asize)
      *_asize = ioparam.ioActCount;

#elif defined(FV_UNIX)
   ulong size = ::read(int(file), _buffer, size_t(_size));
   if      (size == 0)
      err = FVE_ENDOFFILE;
   else if (size < _size)
      err = FVE_SHORTREAD;
   if (_asize)
      *_asize = size;

#endif // FV_UNIX

   return (err == FVE_OK);
   } // read

//---------------------------------------------------------------------
// Write a block to the file
//---------------------------------------------------------------------
bool CFile::write(Pvoid _buffer, ulong _size) {
   if (!fileCanWrite) {
      SetError(FVE_FILENOTOUTPUT);
      return false;
      }

   // Same special handling for all platforms if standard output
   if (stdInOut == stdOut) {
      size_t size = fwrite(_buffer, _size, 1, stdout);
      if (size < _size) {
         sysErr = FVE_BADWRITE;
         SetError(FVE_BADWRITE);
         return false;
         }
      return true;
      }

#if defined(FV_MSWIN32)
   DWORD size;
   if (!::WriteFile(HANDLE(file), _buffer, _size, &size, 0)) {
      sysErr = GetLastError();
      SetError(FVE_BADWRITE);
      }

#elif defined(FV_MSWIN16)
   ulong size = ::write(int(file), _buffer, size_t(_size));
   if (size != _size) {
      sysErr = FVE_BADWRITE;
      SetError(FVE_BADWRITE);
      }

#elif defined(FV_MACINTOSH)
   IOParam ioParam;
   memset(&ioParam, 0, sizeof(IOParam));
   ioParam.ioRefNum   = short(file);
   ioParam.ioBuffer   = Ptr(_buffer);
   ioParam.ioReqCount = _size;
   ioParam.ioPosMode  = fsFromMark;
   sysErr             = long(::PBWriteSync(ParmBlkPtr(&ioParam)));
   if (OSErr(sysErr) != noErr)
      SetError(FVE_BADWRITE);

#elif defined(FV_UNIX)
   ulong size = ::write(int(file), _buffer, size_t(_size));
   if (size != _size) {
      sysErr = FVE_BADWRITE;
      err    = FVE_BADWRITE;
      }

#endif // FV_UNIX

   return (err == FVE_OK);
   } // write

