
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
//
// FindList.cpp
//
//*********************************************************************

#include "prefixb.h"
#include "book.h"
#include "file.h"
#include "findlist.h"

//---------------------------------------------------------------------
// Constructor
//
CFindItem::CFindItem(PCBook _book,
                     PCRef  _ref,
                     Pchar   _name,
                     ulong   _position):
   CItem   (),
   book    (_book),
   ref     (0),
   name    (0),
   position(_position) {

   ref = new CRef(*_ref);

   if (!StrEmpty(_name)) {
      STRINIT(name, _name);
      }
   } // CFindItem

//---------------------------------------------------------------------
//
CFindItem::CFindItem(PCFindItem _item):
   CItem   (),
   book    (_item->book),
   ref     (_item->ref),
   name    (0),
   position(_item->position) {

   STRINIT(name, _item->name);
   } // CFindItem

//---------------------------------------------------------------------
// Destructor
//
CFindItem::~CFindItem() {
   delete[] name;
   name = 0;
   } // ~CFindItem

//---------------------------------------------------------------------
//
RTTI_CPP(FindItem, Item)

//---------------------------------------------------------------------
// Update the frame name field
//
void CFindItem::SetName(Pchar _name) {
   Assert(_name);
   delete[] name;
   name = new char[strlen(_name)+1];
   strcpy(name, _name);
   } // SetName

//---------------------------------------------------------------------
// Return the address of the frame name field
//
Pchar CFindItem::GetName(bool _alc) {
   if (_alc) {
      if (StrEmpty(name)) {
         Pchar p = new char[1];
         p[0]    = 0;
         return p;
         }

      Pchar p = new char[strlen(name)+1];
      strcpy(p, name);
      return p;
      }

   return name;
   } // GetName

//---------------------------------------------------------------------
// farSlang version
//
Pchar CFindItem::GetName(void) {
   return GetName(true);
   } // GetName

//---------------------------------------------------------------------
// Return the address of the associated book object
//
PCBook CFindItem::GetBook(void) {
   return book;
   } // GetBook

//---------------------------------------------------------------------
// Set the address of the associated book object
//
void CFindItem::SetBook(PCBook _book) {
   book = _book;
   } // SetBook

//---------------------------------------------------------------------
// Return the address of the frame index field
//
PCRef CFindItem::GetRef(void) {
   return ref;
   } // GetRef

//---------------------------------------------------------------------
// Return true if the specified references matches
//
bool CFindItem::RefEquals(PCRef _ref) {
   return (ref == _ref);
   } // RefEquals

//---------------------------------------------------------------------
// Update the frame position field
//
void CFindItem::SetPosition(ulong _position) {
   position = _position;
   } // SetPosition

//---------------------------------------------------------------------
// Return the frame position field
//
ulong CFindItem::GetPosition(void) {
   return position;
   } // GetPosition

//=====================================================================
//---------------------------------------------------------------------
// Constructor
//
CFindList::CFindList(void):
   CList(),
   error(FVE_OK) {
   } // CFindList

//---------------------------------------------------------------------
//
CFindList::CFindList(PCFindList _list):
   CList(),
   error(FVE_OK) {
   PCItem item = _list->Item(0);
   while (item) {
      PCFindItem fitm = TYPECAST(item, FindItem);
      PCFindItem nitm = new CFindItem(fitm);
      FindInsert(nitm);
      item = item->NextSibItem();
      }
   } // CFindList

//---------------------------------------------------------------------
//
CFindList::~CFindList() {
   } // ~CFindList

//---------------------------------------------------------------------
//
RTTI_CPP(FindList, List)

//---------------------------------------------------------------------
// Return current error value
//
ulong CFindList::Error(void) {
   return error;
   } // Error

//---------------------------------------------------------------------
// Set error value
//
void CFindList::Error(ulong _error) {
   error = _error;
   } // Error

//---------------------------------------------------------------------
//
PCFindItem CFindList::FindItem(void) {
   PCItem item = Item();
   return TYPECAST(item, FindItem);
   } // FindItem

//---------------------------------------------------------------------
//
PCFindItem CFindList::FindItem(long _where) {
   PCItem item = Item(_where);
   return TYPECAST(item, FindItem);
   } // FindItem

//---------------------------------------------------------------------
// Find frames in the specified farBook using the specified key
//
bool CFindList::FindKey(PCBook _book,
                        char    _mark,                  // char = aMark
                        Pchar   _key,
                        bool    _trunc) {
   Pchar work = new char[sizeString];
   Strcpy(work, _key, sizeString);

   while (!StrEmpty(work)) {
      PCBook book = TYPECAST(_book, Book);
      PCList me   = this;
      book->FindBookKey(_mark, work, me, _trunc);
      Error(book->Error());
      if ((error/100)*100 == FVE_REMOTEBOOKERROR) {
         // The connection to the host failed
         delete[] work;
         return false;
/*? some old code for error condition
         if (Number() == 0      &&
             (_mark == aName ||
              _mark == anAlias) &&
             CBook::homeBook    &&
             CBook::homeBook != _book)
            // Check homeBook if it is available
            if (CBook::homeBook->FindBookKey(_mark, work, this))
               break;
*/
         }
      if (_trunc && _mark == anAlias) {
         // Shorten the key by one word
         short i = short(strlen(work)-1);
         while (i > 0              &&
               (!IsDelim(work[i])  ||
                work[i] == '.'     ||
                work[i] == ','))
            i--;
         work[i] = '\0';
         Trim(work);
         }
      else
         break;
      }

   delete[] work;
   return true;
   } // FindKey

//---------------------------------------------------------------------
// Search to see if the specified reference is therein
//
bool CFindList::FindRef(PCRef _ref) {
   if (GotoHead()) do {
      PCItem item = Item();
      PCFindItem findItem(TYPECAST(item, FindItem));
      if (findItem->RefEquals(_ref))
         return true;
      } while (GotoNext());

   return false;
   } // FindRef

//---------------------------------------------------------------------
bool CFindList::FindInsert(PCFindItem _item) {
   PCItem item = TYPECAST(_item, Item);
   return Insert(item);
   }  // FindInsert

//---------------------------------------------------------------------
// Append a CFindItem object to the list
//
bool CFindList::FindInsert(PCBook _book,
                           PCRef  _ref,
                           Pchar   _name,
                           ulong   _position) {
   if (_ref->loc) {

//?   _book->AddRef();
      PCItem it = new CFindItem(_book, _ref, _name, _position);
      return Insert(it);
      }
   else
      return false;
   } // FindInsert

//---------------------------------------------------------------------
// Get Names of nodes in the find list
//
bool CFindList::GetNames(PCBook _book) {
   if (GotoHead()) do {
      PCItem item         = Item();
      PCFindItem findItem = TYPECAST(item, FindItem);
      PCRef ref           = new CRef(findItem->GetRef());

/*
      // Read frame abstract
      PCBook book(TYPECAST(_book, Book));
      PCAtom atm = book->ReadPart(ref);
//?   book->OpenNode(ref, readAction);
//?   PCAtom atm = book->ReadPart(0);
//?   PCManifest manifest(TYPECAST(atm, Manifest));
      Error(book->Error());
      if (Error() != FVE_OK ||
          manifest          ||
          !(manifest->Isa(iManifest))) {
         Error(FVE_NOMANIFEST);
         return (Error() == FVE_OK);
         }
//?   book->CloseNode();

      // Update the name field in the find record
      char nodeName[sizeString];
      manifest->GetManifestField("Name", nodeName, sizeString);
      findItem->SetName(nodeName);
*/
      } while (GotoNext());
   return true;
   } // GetNames


