
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// hash.cpp
//*********************************************************************
//
//                              hash
//
//                            jan 1984
// references:
//    'Making Hash With Tables'; Terry Dollhoff,
//     Byte Magazine, January, 1977
//    'Data Stuctures'; Horowitz and Sani
//
//    This module creates a hashed access file to hold up to fSize
//    records, then permits to insert records consisting of string
//    KEYs and long VALs. Once the file is built, the file can be
//    opened and keys can be added, FOUND, and deleted. The space
//    of deleted keys is made available for reuse. Note that the
//    file should always be closed after every session.
//
//    Note that key searching is case sensitive.
//
//*********************************************************************

#include "prefixb.h"
#include "hash.h"
#include "misc.h"

const ulong maxNrBuckets = 516800L;

//---------------------------------------------------------------------
//
CRef::CRef(void):
   CAtom(),
   loc (0) {
   } // CRef

//---------------------------------------------------------------------
//
CRef::CRef(PCRef _ref):
   CAtom(),
   loc (_ref->loc) {
   } // CRef

//---------------------------------------------------------------------
//
CRef::CRef(ulong _loc):
   CAtom(),
   loc (_loc) {
   } // CRef

//---------------------------------------------------------------------
//
CRef::~CRef() {
   } // ~CRef

//---------------------------------------------------------------------
//
RTTI_CPP(Ref, Atom)

//---------------------------------------------------------------------
//
void CRef::Clear(void) {
   loc = 0;
   } // Clear

//---------------------------------------------------------------------
//
void CRef::Loc(ulong _loc) {
   loc = _loc;
   } // Loc

//---------------------------------------------------------------------
//
ulong CRef::Loc(void) {
   return loc;
   } // Loc

//---------------------------------------------------------------------
//
bool CRef::operator ==(CRef _ref) {
   return (loc == _ref.loc);
   } // operator==

//---------------------------------------------------------------------
//
bool CRef::operator !=(CRef _ref) {
   return (loc != _ref.loc);
   } // operator!=

//---------------------------------------------------------------------
//
bool aHashDesc::Read(RCFile _file, ulong _location) {
   if (!_file.Seek(_location, fromHead))
      return false;
   if (!_file.ReadBlock(&nrBuckets, sizeof(ulong)))
      return false;
   Assert(nrBuckets > 0);
   if (!_file.ReadBlock(&nrSlots,   sizeof(ushort)))
      return false;
   Assert(nrSlots > 0);
   if (!_file.ReadBlock(&sizeKey,   sizeof(ushort)))
      return false;
   Assert(sizeKey > 0);
   return true;
   }; // Read

//---------------------------------------------------------------------
//
bool aHashDesc::Write(RCFile _file, ulong _location) {
   if (!_file.Seek(_location, fromHead))
      return false;
   if (!_file.WriteBlock(&nrBuckets, sizeof(ulong)))
      return false;
   if (!_file.WriteBlock(&nrSlots,   sizeof(ushort)))
      return false;
   if (!_file.WriteBlock(&sizeKey,   sizeof(ushort)))
      return false;
   return true;
   }; // Write

//---------------------------------------------------------------------
//
ushort aBucket::Size(RaHashDesc _hd) {
   return sizeof(ulong) +
          sizeof(ulong) +
          _hd.nrSlots * (sizeof(aRef) + _hd.sizeKey);
   } // Size

//---------------------------------------------------------------------
//
paKeyIndex aBucket::Slot(short _iSlot, RaHashDesc _hd) {
   Assert(_iSlot < _hd.nrSlots);

   return paKeyIndex(Pchar(&next) +
                     sizeof(next) +
                     sizeof(use)  +
                     _iSlot * (sizeof(aRef) + _hd.sizeKey));
   } // Slot

//---------------------------------------------------------------------
//
CHash::CHash(ulong _hdrSize, Pchar _name):
   CFile     (_name),
   fileActive(false),
   hashError (false),
   failed    (false),
   state     (looking),
   hdrSize   (_hdrSize),
   bkt       (0),
   bucket    (0),
   keyBkt    (0),
   keyBucket (0),
   nrBkts    (0) {

   // Set up default key size and bucket size
   hd.nrBuckets =  0;
   hd.nrSlots   = 16;
   hd.sizeKey   = 64;

   // Allocate space for the inMemory bucket
   bkt          = new CBlock();
   bkt->Alloc(aBucket::Size(hd));
   bucket       = paBucket(bkt->Lock());
   bkt->Clear(0);
   bucket->next = chainEnd;
   bucket->use  = hNeverFilled;
   } // CHash

//---------------------------------------------------------------------
//
CHash::~CHash() {
   } // ~CHash

//---------------------------------------------------------------------
//
RTTI_CPP(Hash, File)

//---------------------------------------------------------------------
// Use of sqrt routine saves only an insignificant amount of time in
// this context. it does work, however.
//
ulong CHash::iSqrt(ulong _c) {
   ulong next, last;
   next = (_c / 2) + 1;
   for (;;) {
      last = next;
      next = (last + _c / last) / 2;
      long diff = long(next) - long(last);
      if (abs(diff) < 2)
         break;
      }
   return next;
   } // iSqrt

//---------------------------------------------------------------------
//
ulong CHash::fPrime(ulong _fileSize) {
   byte* flags = new byte[ushort(maxNrBuckets/8)+1];
   memset(flags, 0xff, ushort(maxNrBuckets/8));
   ulong root  = iSqrt(_fileSize) + 1;

   for (ulong i = 2; i <= root; ++i)
      if (bitGet(flags[ushort(i/8)], ushort(i%8)))
         for (ulong n = i * i; n < maxNrBuckets; n+=i) {
            flags[ushort(n/8)] = bitSet(flags[ushort(n/8)], ushort(n%8), false);
            }
   for (; !bitGet(flags[ushort(_fileSize/8)], ushort(_fileSize%8)); ++_fileSize);

   delete[] flags;
   return _fileSize;
   } // fPrime

//---------------------------------------------------------------------
// Can generate hash files to handle up to
// 516,800 buckets * SIZE_BUCKET keys.
//
bool CHash::getSize(ulong _nrKeys) {
   _nrKeys     += _nrKeys / 4;                        // add 25 percent
   ulong nrBkts = (_nrKeys + hd.nrSlots - 1) / hd.nrSlots;
   if (nrBkts > maxNrBuckets)
      nrBkts = maxNrBuckets;

   hd.nrBuckets = fPrime(nrBkts);
   return (hd.nrBuckets > 0);
   } // getSize

//---------------------------------------------------------------------
//
bool CHash::makeFile(void) {
   if (!CFile::Create())
      return false;
   if (!OpenUpdate())
      return false;
   return true;
   } // makeFile

//---------------------------------------------------------------------
// Create an empty file with a header block and
// a set of empty buckets
//
bool CHash::makeBuckets(void) {
   // Write an empty header to the file
   if (hdrSize > 0) {
      Pchar header = new char[hdrSize];
      memset(header, 0, hdrSize);
      if (!WriteBlock(header, hdrSize)) {   // User fills this in later
         delete[] header;
         return false;
         }
      delete[] header;
      }

   // Write the file size parameters to the file
   if (!hd.Write(*this, hdrSize))
      return false;

   // Make an empty bucket to write to the file
   PCBlock bkt     = new CBlock();
   bkt->Alloc(aBucket::Size(hd));
   paBucket bucket = paBucket(bkt->Lock());
   bkt->Clear(0);
   bucket->next = chainEnd;          // No chained buckets initially
   bucket->use  = hNeverFilled;

   // Write an empty bucket array to the file
   for (ulong i = 0; i < hd.nrBuckets; ++i)
      if (!WriteBlock(bucket, aBucket::Size(hd)))
         return false;

   Flush();
   return true;
   } // makeBuckets

//---------------------------------------------------------------------
// SetSizes must be called before Create is called, but can be called
// after Open is called, but before any keys are searched.
//
void CHash::SetSizes(long _nrSlots, long _sizeKey, long _nrKeys) {
   Assert(_sizeKey < sizeString-16);

   bool newBucket = false;

   if (_nrSlots) {
      newBucket    = true;
      hd.nrSlots   = _nrSlots;
      }

   if (_sizeKey) {
      newBucket    = true;
      hd.sizeKey   = _sizeKey;
      }

   if (newBucket) {
      hd.nrBuckets = 0;
      getSize(_nrKeys);

      bkt->Free();
      bkt->Alloc(aBucket::Size(hd));
      bucket       = paBucket(bkt->Lock());
      bkt->Clear(0);
      bucket->next = chainEnd;
      bucket->use  = hNeverFilled;
      }
   } // SetSizes

//---------------------------------------------------------------------
// Create a hash file capable of handling at least
// the number of entries specified in _nrKeys
//
bool CHash::Create(void) {
   Assert(hd.nrBuckets > 0);
   fileActive = true;

   // Create the hash file
   if (!makeFile())
      return false;

   // Initialize the hash file
   if (!makeBuckets())
      return false;

   Close();
   return true;
   } // Create

//---------------------------------------------------------------------
// Open the specified hash file.
//
bool CHash::Open(void) {
   fileActive = true;

   if (!OpenUpdate())
      return false;

   // Reset status variables
   inMemory.Clear();
   keyPlace.Clear();
   available.Clear();

   // Get the operating parameters
   if (!hd.Read(*this, hdrSize))
      return false;

   // Initialize the inMemory bucket
   bkt->Clear(0);
   bucket->next = chainEnd;
   bucket->use  = hNeverFilled;

   return true;
   } // Open

//---------------------------------------------------------------------
// Hash the specified key into an unsigned integer.
//
ulong CHash::myHash(Pchar _key) {
   ulong h  = 0;
   while (*_key) {
      char c = MapChar(*_key++);
      if (c)
         h = h * 0x2B + c;
      }
   return h % hd.nrBuckets;
   } // myHash

//---------------------------------------------------------------------
// Check the current slot for a key matching the specified key.
//
SEARCH CHash::checkKey(Pchar _key, short _iSlot, SEARCH _state) {
   paKeyIndex pSlot = bucket->Slot(_iSlot, hd);
   char key[sizeString];
   Strcpy(key, pSlot->key, sizeString);
   if      (strcmp(key, _key) == 0) {
      // Found a matching slot in the bucket
      keyPlace.location = inMemory.location;
      keyPlace.iSlot    = _iSlot;
      return found;
      }
   else if (StrEmpty(pSlot->key) &&
            available.iSlot < 0) {
      // Record only the first empty slot in the bucket
      available.location = inMemory.location;
      available.iSlot    = _iSlot;
      }
   return _state;
   } // checkKey

//---------------------------------------------------------------------
// Search the current bucket for the specified key.
//
SEARCH CHash::searchKey(Pchar _key, Rshort _iSlot, SEARCH _state) {
   for (_iSlot = 0;_iSlot < hd.nrSlots; ++_iSlot) {
      _state = checkKey(_key, _iSlot, _state);
      if (_state != looking)
         return _state;
      }

   if (bucket->use == hNeverFilled)
      return missing;
   else
      return _state;
   } // searchKey

//---------------------------------------------------------------------
// Load a bucket from the hash file. If the first time called, then
// load the bucket whose address is the hash of the specified
// key. If not the first time called, load a bucket from the
// overflow chain, if any.
//
SEARCH CHash::getBucket(Pchar _key, bool _firstTime, SEARCH _state) {
   Assert(hd.nrBuckets > 0);
   if (_firstTime) {
      ulong location = (myHash(_key) * aBucket::Size(hd)) +
                       sizeof(aHashDesc)                  +
                       hdrSize;
      // Check for bucket already in memory
      if (inMemory.location != location) {
         inMemory.location = location;
         Seek(inMemory.location, fromHead);
         if (!ReadBlock(bucket, aBucket::Size(hd)))
            return missing;
         }
      }
   else {
      ulong location = bucket->next;
      
      // Validate that location is within the file
      if (location > Size() - aBucket::Size(hd))
         return missing;                 // Corrupt bucket.next pointer

      if (location != chainEnd) {
         inMemory.location = location;
         Seek(inMemory.location, fromHead);
         if (!ReadBlock(bucket, aBucket::Size(hd)))
            return missing;
         }
      else
         return missing;
      }
   return _state;
   } // getBucket

//---------------------------------------------------------------------
// Find the key in the hash file and return the associated ref.
//
bool CHash::findKey(Pchar _key, RaRef _ref) {
   // Prepare the key for the search
   short sizeKey = SizeKey();
   Pchar key     = new char[sizeKey];
   memset(key, 0, sizeKey);
   if (strlen(_key) >= sizeKey)                 // truncate if too long
      _key[sizeKey-1] = '\0';
   Strcpy(key, _key, sizeKey);

   keyPlace.iSlot  = -1;
   available.iSlot = -1;

   // Search the hash file for a matching slot
   bool firstTime = true;
   short iSlot    = 0;
   for (state = looking; state == looking;) {
      state = getBucket(key, firstTime, state);

      if (state != missing)
         state = searchKey(key, iSlot, state);
      firstTime = false;
      }

   delete[] key;

   if (state == found) {
      hashError = false;
      _ref.Copy(bucket->Slot(iSlot, hd)->ref);
      return true;
      }
   else {
      // MISSING case
      hashError = true;
      _ref.Set(0);
      return false;
      }
   } // findKey

//---------------------------------------------------------------------
//
bool CHash::FindKey(Pchar _key, PCRef _ref) {
   aRef ref(_ref);
   bool rtrn = findKey(_key, ref);
   _ref->loc = ref.loc;
   return rtrn;
   } // FindKey

//---------------------------------------------------------------------
// Update the specified entry
//
bool CHash::UpdateKey(Pchar _key, PCRef _ref) {
   // Prepare the key for the search
   short sizeKey = SizeKey();
   Pchar key     = new char[sizeKey];
   memset(key, 0, sizeKey);
   if (strlen(_key) >= sizeKey)
      _key[sizeKey-1] = '\0';
   Strcpy(key, _key, sizeKey);

   // Check to see if the entry is already
   // present or that we can find it
   aRef ref;
   if ((keyPlace.location == inMemory.location &&
        keyPlace.iSlot  >= 0                   &&
        strcmp(bucket->Slot(keyPlace.iSlot, hd)->key, key) == 0) ||
       findKey(key, ref)) {                               // FOUND case
      bucket->Slot(keyPlace.iSlot, hd)->ref = _ref;
      Seek(inMemory.location, fromHead);
      if (!WriteBlock(bucket, aBucket::Size(hd))) {
         delete[] key;
         return false;
         }
      Flush();
      }
   else {                                               // MISSING case
      delete[] key;
      return false;
      }
   delete[] key;

   return true;
   } // UpdateKey

//---------------------------------------------------------------------
// makeBucket extends the hash chain for the
// current bucket by an additional bucket.
//
ulong CHash::makeBucket(void) {
   // Update the next pointer of the current
   // bucket to reflect the chain extension.
   ulong newBucket = Size();
   bucket->next    = newBucket;
   Seek(inMemory.location, fromHead);
   if (!WriteBlock(bucket, aBucket::Size(hd)))
      return chainEnd;

   // Initialize the inMemory bucket to empty
   bkt->Clear(0);
   bucket->next = chainEnd;
   bucket->use  = hNeverFilled;

   // Write the chain bucket to the end of the file
   Seek(0, fromTail);
   Assert(newBucket == Where());
   if (!WriteBlock(bucket, aBucket::Size(hd)))
      return chainEnd;
   Flush();

   inMemory.location = newBucket;

   // Return CHAIN_END signal if any error
   if (failed)
      return chainEnd;
   else
      return newBucket;
   } // makeBucket

//---------------------------------------------------------------------
// Return the chain pointer. If CHAIN_END, then "rehash" the
// chain and return the file offset of an extension bucket.
//
ulong CHash::reHash(void) {
   if (bucket->next == chainEnd)
      return makeBucket();
   else
      return bucket->next;
   } // reHash

//---------------------------------------------------------------------
// If no bucket available then build a new
// bucket. Otherwise, load available bucket.
//
bool CHash::loadBucket(void) {
   if      (available.location == chainEnd) {
      available.location = makeBucket();
      }
   else if (available.location != inMemory.location) {
      Seek(available.location, fromHead);
      if (!ReadBlock(bucket, aBucket::Size(hd)))
         return false;
      }

   return true;
   } // loadBucket

//---------------------------------------------------------------------
// Find an available slot in the current bucket.
//
short CHash::findSpace(void) {
   for (short iSlot = 0; iSlot < aBucket::Size(hd); ++iSlot)
      if (StrEmpty(bucket->Slot(iSlot, hd)->key))
         return iSlot;
   return -1;
   } // findSpace

//---------------------------------------------------------------------
// Find an available slot and add the new _key and _ref.
//
bool CHash::putSlot(Pchar _key, RaRef _ref) {
   for (;;) {
      available.iSlot = findSpace();
      if (available.iSlot >= 0)
         break;
      else {
         available.location = reHash();
         Assert(available.location != chainEnd);
         if (!loadBucket())
            return false;
         }
      }

   Assert(available.iSlot >= 0);

   // Store the key and reference data into the slot
   paKeyIndex slot = bucket->Slot(available.iSlot, hd);
   memset(slot->key, 0, hd.sizeKey);
   Strcpy(slot->key, _key, hd.sizeKey);
   slot->ref.Copy(_ref);

   // Set use code to reflect new filled status
   bucket->use = hNeverFilled;
   if (available.iSlot == hd.nrSlots-1)
      bucket->use = hOnceFilled;

   Seek(available.location, fromHead);
   if (!WriteBlock(bucket, aBucket::Size(hd)))
      return false;

   Flush();
   return true;
   } // putSlot

//---------------------------------------------------------------------
// Insert the _key and _ref into a slot in a bucket that
// corresponds to the hash value of the _key.
//
bool CHash::InsertKey(Pchar _key, PCRef _ref) {
   // Prepare the key for the insert
   short sizeKey = SizeKey();
   Pchar key     = new char[sizeKey];
   memset(key, 0, sizeKey);
   if (strlen(_key) >= sizeKey)
      _key[sizeKey-1] = '\0';
   Strcpy(key, _key, sizeKey);

   available.location = chainEnd;

   // Locate and update an available slot
   PCRef tref = new CRef();
   if (!FindKey(key, tref)) {
      if (!loadBucket()) {
         delete[] key;
         return false;
         }
      aRef ref(_ref);
      if (!putSlot(key, ref)) {
         delete[] key;
         return false;
         }
      }

   delete[] key;
   return true;
   } // InsertKey

//---------------------------------------------------------------------
//
//
bool CHash::DeleteKey(Pchar _key, PCRef _ref) {
   // Prepare the key for the search
   short sizeKey = SizeKey();
   Pchar key     = new char[sizeKey];
   memset(key, 0, sizeKey);
   if (strlen(_key) >= sizeKey)                 // truncate if too long
      _key[sizeKey-1] = '\0';
   Strcpy(key, _key, sizeKey);

   keyPlace.location = chainEnd;

   // Locate and delete the matching slot
   if (FindKey(key, _ref)) {
      for (short i = keyPlace.iSlot+1; i < hd.nrSlots; ++i)
         memcpy(bucket->Slot(i-1, hd),
                bucket->Slot(i, hd),
                sizeof(aRef) + hd.sizeKey);
      memset(bucket->Slot(hd.nrSlots-1, hd)->key, 0, hd.sizeKey);
      bucket->Slot(hd.nrSlots-1, hd)->ref.Set(0);
      Seek(keyPlace.location, fromHead);
      if (!WriteBlock(bucket, aBucket::Size(hd))) {
         delete[] key;
         return false;
         }

      Flush();
      }
   else {
      delete[] key;
      return false;
      }
   delete[] key;

   return true;
   } // DeleteKey

//---------------------------------------------------------------------
// Serialize the hash file
//
//   return |key|loc                                      for each key,
//          |*|H|bucketNr|bucketLoc|nrSlots|szKey|szBucket     at start
//                                                        of each chain
//                                                  of each hash bucket
//          |*|C|bucketLoc               for start of each chain bucket
//          |*|B                                       for empty bucket
//          |*|E                                        for end of file
//
// Use the following fapis as in this sample code:
//
//     PCHash hash = new CHash(hashFileName);
//     if (hash->Open()) {
//        PCRecord key  = new CRecord();
//        if (hash->GetFirstKey(key)) do {
//           char tipe[2];
//           key->GetField(1, tipe, 2);
//           if (tipe[0] = '*') {
//              // process information record
//              }
//           else {
//              // Process the key
//              }
//           } while (hash->GetNextKey(key));
//        hash->Close();
//        }
//
// Note that it is safe to use the GetFirstKey/GetNextKey at any time
// while the hash file is open.
//
// Since it is possible to run this code across the internet, one
// person could analyse someone else's farBooks. Consequently, this
// code needs to restrict access to those having a certain level of
// permission. Have to work that out.
//
//---------------------------------------------------------------------
//
bool CHash::GetFirstKey(PCRecord _key) {
   // Allocate space for the inMemory bucket
   delete keyBkt;
   keyBkt          = new CBlock();
   keyBkt->Alloc(aBucket::Size(hd));
   keyBkt->Lock();
   keyBkt->Clear(0);
   keyBucket       = paBucket(keyBkt->Pointer());
   keyBucket->next = chainEnd;
   keyBucket->use  = hNeverFilled;
   nrBkts          = hd.nrBuckets;

   char work[sizeString];

   // Load the first (last) bucket into the buffer
   if (nrBkts > 0) {
      nrBkts--;
      Seek(nrBkts * aBucket::Size(hd) +
           hdrSize                    +
           sizeof(aHashDesc),
           fromHead);
      if (!ReadBlock(keyBucket, aBucket::Size(hd)))
         return false;
      keyBucket->use = useType(hd.nrSlots);  // Coopting to track slots
      sprintf(work, "|*|H|%ld|%ld|%d|%d|%d", hd.nrBuckets,
                                             hd.nrBuckets * aBucket::Size(hd),
                                             hd.nrSlots,
                                             hd.sizeKey,
                                             aBucket::Size(hd));
      Assert(strlen(work) < sizeString);
      _key->NewRecord(work);
      return true;
      }

   delete keyBkt;
   keyBkt    = 0;
   keyBucket = 0;

   _key->NewRecord(Pchar("|*|E"));
   return false;
   } // GetFirstKey

//---------------------------------------------------------------------
//
bool CHash::GetNextKey(PCRecord _key) {
   Assert(keyBucket);

   char work[sizeString];

   // Section to obtain the next bucket
   if (keyBucket->use == hNeverFilled) {             // Bucket is empty
      ulong next = keyBucket->next;
      if      (next != chainEnd) {  // Get next bucket in chain
         Seek(next, fromHead);
         if (!ReadBlock(keyBucket, aBucket::Size(hd)))
            return false;
         keyBucket->use = useType(hd.nrSlots);  // Coopting to track slots
         sprintf(work, Pchar("|*|C|%ld"), next);
         Assert(strlen(work) < sizeString);
         _key->NewRecord(work);
         return true;
         }
      else if (nrBkts > 0) {        // Get the next bucket in the table
         nrBkts--;
         Seek(nrBkts * aBucket::Size(hd) +
              hdrSize                    +
              sizeof(aHashDesc),
              fromHead);
         if (!ReadBlock(keyBucket, aBucket::Size(hd)))
            return false;
         keyBucket->use = useType(hd.nrSlots);// Coopting to trak slots
         sprintf(work, Pchar("|*|H|%ld|%ld|%d|%d|%d"), nrBkts,
                                                nrBkts * aBucket::Size(hd),
                                                hd.nrSlots,
                                                hd.sizeKey,
                                                aBucket::Size(hd));
         Assert(strlen(work) < sizeString);
         _key->NewRecord(work);

         return true;
         }
      else {                                             // End of file
         delete keyBkt;
         keyBkt    = 0;
         keyBucket = 0;

         _key->NewRecord(Pchar("|*|E"));
         return false;
         }
      }

   // Section to search the current bucket (got a bucket with slots)
   SEARCH state = looking;
   while (state == looking)         // Search for an slot with a key
      if      (short(keyBucket->use) == 0)
         state = missing;
      else if (keyBucket->Slot(short(keyBucket->use-1), hd)->key[0])
         state = found;
      else
         keyBucket->use = useType(short(keyBucket->use) - 1);

   ulong ndx = ulong(keyBucket->use)-1;

   if (state == found) {           // Extract information from key slot
      sprintf(work, Pchar("|%s|%ld"), keyBucket->Slot(ndx, hd)->key,
                               keyBucket->Slot(ndx, hd)->ref.loc);
      Assert(strlen(work) < sizeString);
      _key->NewRecord(work);
      keyBucket->use = useType(ndx);
      return true;
      }

   _key->NewRecord(Pchar("|*|B"));
   return true;
   } // GetNextKey


