
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


//*********************************************************************
// hash.h
//*********************************************************************
//
// Note: hdrSize parameter in CHash constructor allows space of the
// specified size at the beginning of the file.
// 
//*********************************************************************

#ifndef HASH_H
#define HASH_H

#ifndef FARBOOK_H
#include "block.h"
#include "file.h"
#include "record.h"
#endif

const ulong chainEnd = 0;

//---------------------------------------------------------------------
// Exportable class to identify a location in the farBook file. (Note:
// will need to modify the interface to account for little-endian vs
// big-endian machines. The farBook is the same for all platforms.)
//
_FBDEF(CRef)
class FBC_USE CRef: public CAtom {
   public:
      CRef(void);
      CRef(PCRef _ref);
      CRef(ulong  _loc);
      virtual ~CRef();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      void Clear(void);

      void  Loc (ulong _loc);
      ulong Loc (void);

      // conditional operators
      bool operator ==(CRef _ref);
      bool operator !=(CRef _ref);

      ulong loc;
   }; // CRef

#ifndef _SLANGER

//---------------------------------------------------------------------
// Structure to identify a location in the farBook file. (Note: will
// need to modify the interface to account for little-endian vs big-
// endian machines. The farBook is the same for all platforms.)
//
__FBDEF(aRef)
class aRef {
public:
   inline aRef(void):
      loc (0) {}

   inline aRef(RaRef _ref):
      loc (_ref.loc) {}

   inline aRef(ulong _loc):
      loc (_loc) {}

   inline aRef(PCRef _ref):
      loc (_ref->loc) {}

   inline void Clear(void) {
      loc  = 0;
      }

   inline void Set(ulong _loc) {
      loc  = _loc;
      }

   inline void Copy(RaRef _ref) {
      loc  = _ref.loc;
      }

   // conditional operators
   bool operator ==(aRef _ref) {
      return (loc == _ref.loc);
      } // operator==

   bool operator !=(aRef _ref) {
      return (loc != _ref.loc);
      } // operator!=

   ulong loc;
   }; // aRef

//---------------------------------------------------------------------
// In-memory structure for managing hash buckets.
//
__FBDEF(aHashDesc)
class aHashDesc {
   friend class CHash;
   friend class aBucket;

   public:
       inline aHashDesc(void):
          nrBuckets(0),
          nrSlots  (0),
          sizeKey  (0) {}

   protected:
      inline ushort SizeBucket(void);

      bool Read (RCFile _file, ulong _location);
      bool Write(RCFile _file, ulong _location);

      ulong  nrBuckets;
      ushort nrSlots;
      ushort sizeKey;
   }; // aHashDesc

//---------------------------------------------------------------------
//
__FBDEF(aLocator)
class aLocator {
public:
   inline aLocator(void):
      location(0),
      iSlot   (0) {};

   inline aLocator(RaLocator _loc):
      location(_loc.location),
      iSlot   (_loc.iSlot) {};

   inline void Clear(void) {
      location = 0;
      iSlot    = 0;
      }

   ulong location;
   short iSlot;
   }; // aLocator

//---------------------------------------------------------------------
// Structure to relate a string key to a location in the farBook file.
// Keys point to node descriptors or reference (alias) tables,
// depending on the type of key.
//
__FBDEF(aKeyIndex)
class aKeyIndex {
public:
   aRef  ref;
   char  key[1];
   }; // aKeyIndex

//---------------------------------------------------------------------
// Table of keys and locations of their associated information. Each
// table also carries a pointer to another table that is used to store
// additional keys when the table is filled.
//
enum useType {hNeverFilled = 0, hOnceFilled = 1};

__FBDEF(aBucket)
class aBucket {
   public:
      static ushort Size(RaHashDesc _hd);
      paKeyIndex    Slot(short _iSlot, RaHashDesc _hd);

      ulong     next;               // points to next bucket for rehash
      useType   use;                   // onceFilled = all slots filled
      aKeyIndex slot[1];
   }; // aBucket

#endif // _SLANGER

//---------------------------------------------------------------------
//
_FBDEF(CHash)
class FBC_USE CHash: public CFile {
   public:
      FB_USE CHash(ulong _hdrSize, Pchar _name);
      virtual FB_USE ~CHash();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      FB_USE_(void) SetSizes  (long _nrSlots,
                               long _sizeKey,
                               long _nrKeys);
      inline ushort SizeKey   (void) { return hd.sizeKey; }

      virtual
      FB_USE_(bool) Create    (void);
      virtual
      FB_USE_(bool) Open      (void);

      FB_USE_(ulong) myHash   (Pchar _key);
      FB_USE_(bool) FindKey   (Pchar _key, PCRef _ref);
      FB_USE_(bool) UpdateKey (Pchar _key, PCRef _ref);

      FB_USE_(bool) InsertKey (Pchar _key, PCRef _ref);
      FB_USE_(bool) DeleteKey (Pchar _key, PCRef _ref);

      FB_USE_(bool) GetFirstKey(PCRecord _key);
      FB_USE_(bool) GetNextKey (PCRecord _key);

   protected:
      ulong  iSqrt       (ulong _c);
      ulong  fPrime      (ulong _fileSize);
      bool   getSize     (ulong _fileSize);
      bool   makeFile    (void);
      bool   makeBuckets (void);

      SEARCH checkKey    (Pchar _key, short _iSlot,      SEARCH _state);
      SEARCH searchKey   (Pchar _key, Rshort _iSlot,     SEARCH _state);
      SEARCH getBucket   (Pchar _key, bool   _firstTime, SEARCH _state);

      ulong  makeBucket  (void);
      ulong  reHash      (void);
      bool   loadBucket  (void);
      short  findSpace   (void);
      bool   putSlot     (Pchar _key, RaRef _ref);

      bool   findKey     (Pchar _key, RaRef _ref);

      // Temporary variables for some methods
      bool      fileActive;
      bool      hashError;
      bool      failed;
      SEARCH    state;

      // Permanent variables for the hash object
      ulong     hdrSize;                       // size of user's header
      aHashDesc hd;                        // records hash entity sizes
      aLocator  inMemory;                // identifies bucket in memory
      aLocator  keyPlace;         // identifies bucket and slot of find
      aLocator  available;      // id's first available bucket and slot
      PCBlock   bkt;             // CBlock object containing the bucket
      paBucket  bucket;          // memory buffer for bucket operations
      pCBlock   keyBkt;       // CBlock object containing the keyBucket
      paBucket  keyBucket;     // buffer used by GetFirstKey/GetNextKey
      ulong     nrBkts;               // used by GetFirstKey/GetNextKey
   }; // CHash

#endif // HASH_H

