
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// lclbook.cpp
//*********************************************************************
//
// Notes:
//
// Do not replace farVIEW version without resolving differences.
//
// Keys:
// * retain marks to identify name-type and duplicate-type keys.
// * retain extended form of reference instead of simple ref.
// (what is an "extended form"? PJM 2000/2/6)
//
// Refs table optimization:
//
// I can avoid wasting unneeded refs subtables if I don't allocate them
// for single alias keys. I would need to identify the key as pointing
// directly to the node descriptor. When a second key is created, I
// would have to update the hash key entry and create an arefs subtable.
// The offset field is 32-bits long. This is really wasted space. So,
// using a bit here would have virtually no impact on the address
// space of the system. This can all be managed in CBook. No changes to
// CHash or CRefs are necessary. This means additional time overhead
// when building a new farBook in batch mode, since it means having to
// update the CHash entry for the key when a second alias is received.
// The approach gains more than it costs, however. (PJM 2000/2/1)
//
// Parts space reuse:
//
// I can avoid wasting space in the farBook by reusing old space by
// allocating some extra for a node part according to user preference
// when I write a part out initially. When it is updated, check if
// there is still enough space for the new version. If so, use the old
// space, if not, write the part to a new location at the end of the
// farBook, allocating additional space according to the algorithm.
//
// Keep the step percentage for each part in the farBook in the farBook
// header. That way, no information is needed in the descriptor for the
// part. The problem that I have yet to solve is when the size of the
// part gets too small to support the currently allocated space. Maybe
// I just have to treat it the same way I would treat a part that has
// grown too large.(?) (PJM 2000/2/7)
//
// Buttons and targets:
//
// In HTML, an <href...> has a mark and a URL. A farVIEW button has
// a mark and a targetset name. Targets also have a mark and a targetset
// name. This allows more than one button to link to more than one
// target. When a target is declared, its targetset name is added to the
// dictionary as a form of alias. When a button is declared, no action
// is taken. When you click the button, the system looks up the
// targetset name in the dictionary. That points to a Refs bucket chain
// that contains pointers for the targets.
//
// I expect I will support XML markup of native farVIEW documents. A
// button would be like
//
// <Btn color="..." font="..."targetset= "mytarget">Click Me</Button>
//
// and a target would be like
//
// <Tgt color="..." targetset="mytarget">Come here</Tgt>
//
// or, if no mark is needed,
//
// <Tgt targetset="mytarget"/>
//
// (To be clicked, a button needs a mark, though it would not be
// required.)
//
// BTW, a targetset name can also be a URL, in which case, no target is
// required, of course, and the <Btn> acts like an <href>.
//
// Well, thinking about it, if I support markup in the text, how am I
// going to know when it has changed? I guess it would have to be an
// operation in the editor so that the change is made on the spot. It
// might be good to provide a command to verify marks when a document
// is imported from another editor.
//
// Locking: add a field to the farBook header that identifies the farBook
// as sharable. When an farBook is sharable, it writes a single byte
// preceeding each node descriptor. The single byte is the "lock byte."
// It is used only for locking and unlocking the node, but it is never
// read from nor written to again. This will help to make the farBook
// compatible across platforms. Once the header flag is implemented,
// always set it on when creating a new farBook. This is to protect
// legacy farBooks; viz., mine.
//
//*********************************************************************

#include "prefixb.h"
#include "book.h"
#include "bookinfo.h"
#include "findlist.h"
#include "lclbook.h"
#include "misc.h"
#ifndef FV_NOLOG
#include "syslog.h"
#endif

const bool   checkVER    = true;

const ushort nrMatchKeys =   42;
const ushort blowOut     =    1;
const ushort infoSize    =  512;
const char   lockByte    =    0;

char  version[]          = "3.00";
// char  cpyrite[]          = "Copyright 1987-2012, "
//                            "Paul J Medlock, All Rights Reserved";

//=====================================================================
//                    CLockItem implementation
//=====================================================================

const short iLockItem = 1000;
_FBDEF(CLockItem)

class CLockItem: public CItem {
   public:
      CLockItem(ulong _loc):
         loc (_loc),
         refs(1) {
         }

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      bool match(Pvoid _p) {
         return (loc == ulong(_p));
         }
      ulong loc;
      ulong refs;
   }; // CLockItem

//---------------------------------------------------------------------
//
RTTI_CPP(LockItem, Item)

//=====================================================================
//                    CLocalBook implementation
//=====================================================================

//---------------------------------------------------------------------
// Local farBook object constructor
//
CLocalBook::CLocalBook(Pchar _name):
   CBook  (_name),
   theFile(0),
   locks  () {

   // create the  file object
   theFile = new CRefs(infoSize, _name);
   locks   = new CList();
   } // CLocalBook

//---------------------------------------------------------------------
// farBook object destructor
//
CLocalBook::~CLocalBook() {
   if (theFile->FileIsOpen())
      Close();
   } // ~CLocalBook

//---------------------------------------------------------------------
//
RTTI_CPP(LocalBook, Book)

//---------------------------------------------------------------------
//
bool CLocalBook::UpdateInfo(void) {
   if (theFile->FileIsOpen()) {
      PCAtom f = pCAtom(theFile);
      return (bookInfo->WriteFile(f, /*_isBook=*/true));
      }

   return false;
   } // UpdateInfo

//---------------------------------------------------------------------
//
ulong CLocalBook::Error(void) {
   ulong error = err;
   Error(FVE_OK);
   
   return error;
   } // Error

//---------------------------------------------------------------------
//
ulong CLocalBook::checkErrors(PCFile _f) {
   if (_f->Error()      != FVE_OK &&
       theFile->Error() == FVE_OK)
      Error(_f->Error());

   return err;
   } // checkErrors

//---------------------------------------------------------------------
// Create a new farBook, deleting the old one, if present. Note that
// SetSizes must already have been called.
//
bool CLocalBook::Create(long _nrParts,
                        long _nrRefSlots,
                        long _nrHshSlots,
                        long _sizeKey,
                        long _nrKeys) {
   Error(FVE_OK);

   if (theFile->Exists())
      theFile->Delete();

   // Must set sizes before creating the file
   theFile->SetSizes(_nrRefSlots, _nrHshSlots, _sizeKey, _nrKeys);
   bookInfo->PutField(Pchar("nrParts"), _nrParts);
   bookInfo->PutField(Pchar("nrRefs"),  _nrRefSlots);
   bookInfo->PutField(Pchar("nrHash"),  _nrHshSlots);
   bookInfo->PutField(Pchar("szKey"),   _sizeKey);
   bookInfo->PutField(Pchar("nrKeys"),  _nrKeys);

   if (_nrParts)
      NrParts(_nrParts);

   // Now it's ok to create the file
   if (!(theFile->Create()))
      return false;

   if (theFile->OpenUpdate()) {
      theFile->Lock(0, 1);

      // Output version and deletion chain header
      // form is: |version|copyright|@book|@home|bookName|#parts|#refs|#hash|szKey|shared
      Assert(infoSize <= sizeString);
      theFile->SetEOLDelim(Pchar(""));

      // Create the farBook header
      PCAtom f = pCAtom(theFile);
      if (!(bookInfo->WriteFile(f, /*_isBook=*/true))) {
         Error(theFile->Error());
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.Create: Could not write to %s (%d)"),
                          bookName,
                          err);
#endif
         theFile->Unlock(0, 1);
         theFile->Close();
         return false;
         }

      PCRef ref = new CRef();

      // Output the farBook descriptor
      lDesc.Clear();
      theFile->Seek(0, fromTail);
      ulong where = theFile->Where();
      if (shared)
         theFile->WriteBlock(Pvoid(&lockByte), sizeof(lockByte));
      lDesc.WriteDesc(*theFile);
      ref->Loc(where);
      theFile->InsertKey(Pchar("<Book>"), ref);

      // Output the home descriptor
      theFile->Seek(0, fromTail);
      where = theFile->Where();
      if (shared)
         theFile->WriteBlock(Pvoid(&lockByte), sizeof(lockByte));
      lDesc.WriteDesc(*theFile);
      ref->Loc(where);
      theFile->InsertKey(Pchar("<Home>"), ref);

      theFile->Unlock(0, 1);
      theFile->Close();
      }
   else
      Error(FVE_NOBOOK);

   return (err == FVE_OK);
   } // Create

//---------------------------------------------------------------------
// Open the farBook files to prepare for access. Note that Open() calls
// SetSizes, the user only needs to call SetSizes before creating a new
// farBook.
//
bool CLocalBook::Open(void) {
   Error(FVE_OK);
   if (!theFile->Open()) {
      Error(theFile->Error());
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.Open: Could not open %s (%d)"),
                       bookName,
                       err);
#endif
      Error(FVE_NOBOOK);
      return false;
      }
   Pchar name = theFile->Name(false);
   STRINIT(bookName, name);

   // Lock the farBook to read the header
   theFile->Lock(0, 1);

   char line[sizeString];

   // Validate the farBook version by checking the version.
   // Form is: |3.0x|@book|@home|bookName|#parts|#refs|#hash|szKey|shared
   theFile->SetEOLDelim(Pchar(""));
   PCAtom f  = pCAtom(theFile);
   bool ok   = bookInfo->ReadFile(f, /*_isBook=*/true);
   Error(theFile->Error());

   // Unlock the farBook for other users
   theFile->Unlock(0, 1);

   if (!ok) {
      Error(theFile->Error());
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.Open: Could not open %s (%d)"),
                       bookName,
                       err);
#endif
      return false;
      }

   bookInfo->GetField(Pchar("version"), line, sizeString);
   if (checkVER && strncmp(line, version, 3)) {
      char work[sizeString];
      sprintf(work, Pchar("%s (%s) is the wrong version"), bookName, line);
      Assert(strlen(work) < sizeString);
      Error(FVE_OLDBOOK);
      theFile->Close();
      }

   // Get the number of descriptor parts supported
   NrParts(bookInfo->GetIntField(Pchar("nrParts")));
   // Adjust the operating sizes for the refs and hash objects
   long nrRefSlots = bookInfo->GetLongField(Pchar("nrRefs"));
   long nrHshSlots = bookInfo->GetLongField(Pchar("nrHash"));
   long sizeKey    = bookInfo->GetLongField(Pchar("szKey"));
   long nrKeys     = bookInfo->GetLongField(Pchar("nrKeys"));
   theFile->SetSizes(nrRefSlots, nrHshSlots, sizeKey, nrKeys);

   // Get the share type of the farBook
   shared = (bookInfo->GetIntField(Pchar("shared")) == 1);

   return (err == FVE_OK);
   } // Open

//---------------------------------------------------------------------
//
bool CLocalBook::Close(void) {
   if (locks->GotoHead()) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.Close: Locks list "
                       "contains locks (%ld)"),
                       locks->Number());
#endif
      Assert(theFile->FileIsOpen());
      do {
         PCRef ref = new CRef(pCLockItem(pCItem(locks->Item()))->loc);
         Unlock(ref);
         locks->Remove();
         } while (locks->GotoHead());
      }

   if (theFile->FileIsOpen()) {
      // Update the book information record if any changes
      if (bookInfo->IsDirty(shallow)) {
         PCAtom f = pCAtom(theFile);
         bookInfo->WriteFile(f, /*_isBook=*/true);
         }
      return theFile->Close();
      }
   else
      return true;
   } // Close

//---------------------------------------------------------------------
// Lock the iNode specified by _id.nodeID.
//
bool CLocalBook::Lock(PCRef _ref, AWaitKind _wait) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D', 
                       Pchar("CLocalBook.Lock: "
                       "Previous error (%ld) not handled."),
                       err);
#endif
      Assert(false);
      }

#ifndef FV_NOLOG
//?SysLog::Log('D', Pchar("Locking %ld"), _ref->loc);
#endif

   if (!_ref)
      return false;

   // Check for a current lock on this location
   if (locks->FindFirst(Pvoid(_ref->loc))) {
      Error(FVE_OK);
      PCItem it      = locks->Item();
      PCLockItem lit = TYPECAST(it, LockItem);
      lit->refs++;
      return true;
      }

   // Try to lock the specified location
   const short waitTime = 100;
   if (_wait == waitForLock)
      for (short t = waitTime; t > 0; t--)
         if (theFile->Lock(_ref->loc, 1)) {
            PCItem it = new CLockItem(_ref->loc);
            locks->Insert(it);
            return true;
            }
         else
            Pause(100);
   else if (theFile->Lock(_ref->loc, 1)) {
      PCItem it = new CLockItem(_ref->loc);
      locks->Insert(it);
      return true;
      }

   return false;
   } // Lock

//---------------------------------------------------------------------
// Unlock the iNode specified by _id.nodeID.
//
bool CLocalBook::Unlock(PCRef _ref) {
   if (!_ref)
      return false;
      
   // Check for a current lock on this location
   if (!(locks->FindFirst(Pvoid(_ref->loc)))) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.Unlock: "
                  "Could not find the specified ref (%d)."),
                  _ref->loc);
#endif
      Assert(false);
      }

   PCItem it      = locks->Item();
   PCLockItem lit = TYPECAST(it, LockItem);
   lit->refs--;
   if (lit->refs > 0)
      return true;

   locks->Remove();

#ifndef FV_NOLOG
//?SysLog::Log('D', Pchar("Unlocked %ld"), _ref->loc);
#endif

   Error(theFile->Error());
   bool ok   = theFile->Unlock(_ref->loc, 1);
   return ok;
   } // Unlock

//---------------------------------------------------------------------
//
PCRefs CLocalBook::GetFile(void) {
   return theFile;
   } // GetFile

//---------------------------------------------------------------------
//
bool CLocalBook::GetDesc(ulong _loc, PCRecord _desc) {
   _desc->Clear();
   
   aDesc desc;

   // Validate the location provided
   if (_loc == 0 || _loc > theFile->Size() -
                    (aDesc::NrParts() * sizeof(ALNdx))) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.GetDesc: Location out of bounds (%d)"),
                  _loc);
#endif
      return false;
      }

   // Read the deacriptor
   if (shared)
      theFile->Seek(_loc+1, fromHead);
   else
      theFile->Seek(_loc,   fromHead);
   if (!(desc.ReadDesc(*(theFile.getPtr())))) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.GetDesc: Could not read descriptor (%d)"),
                  _loc);
#endif
      return false;
      }

   // Format the return value
   short nr = desc.NrParts();
   Assert(nr);
   _desc->NewRecord(nr, Pchar("|"));
   for (short i = 0; i < nr; i++) {
      PCRecord part = new CRecord(3, Pchar(";"));
      part->PutField(1, desc.Type(i));
      part->PutField(2, desc.Where(i));
      part->PutField(3, desc.Size(i));
      _desc->PutField(i+1, part->GetDelimStr(false));
      }

   return true;
   } // GetDesc

//---------------------------------------------------------------------
//                            link i/o
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// Check the farBook dictionary to determine if the key is present.
//
bool CLocalBook::IfBookKey(AMark _mark, Pchar _key, bool _trunc) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.IfBookKey: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   if (StrEmpty(_key)) {
      Error(FVE_EMPTYKEY);
      return false;
      }

   ulong szKey = theFile->SizeKey();

   // Check for key too long while truncation not permitted
   if ((!_trunc) && strlen(_key) >= szKey-1) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.IfBookKey: Key too long (%s)"), _key);
#endif
      Error(FVE_LONGKEY);
      return false;
      }

   // Prepare the key
   char key[sizeString];
   Strcpy(key, _key, sizeString);
   prepKey(_mark, key, sizeString, szKey);

//?if (_mark == isPrepped)
//?   _mark = _key[0];

   // Search the dictionary
   PCRef ref = new CRef();
   return (theFile->FindKey(key, ref));
   } // IfBookKey

//---------------------------------------------------------------------
// Use the key to find all descriptors that are identified
// by the key, and add them to the findList. Return true if
// any descriptors are found.
//
bool CLocalBook::FindBookKey(AMark   _mark,
                             Pchar   _key,
                             PCList _pFind,
                             bool    _trunc) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.FindBookKey: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   if (StrEmpty(_key)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.FindBookKey: Key empty"));
#endif
      Error(FVE_EMPTYKEY);
      return false;
      }

   ulong szKey = theFile->SizeKey();

   // Check for key too long while truncation not permitted
   if ((!_trunc) && strlen(_key) >= szKey-1) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.FindBookKey: Key too long (%s)"), _key);
#endif
      Error(FVE_LONGKEY);
      return false;
      }

   PCFindList pFind = TYPECAST(_pFind, FindList);

   // Include Name keys with Alias keys
   if (_mark == anAlias)
      FindBookKey(aName, _key, _pFind, _trunc);

   // Prepare the key
   char key[sizeString];
   Strcpy(key, _key, sizeString);
   prepKey(_mark, key, sizeString, szKey);

   if (_mark == isPrepped)
      _mark = _key[0];

   PCRef refRef  = new CRef();
   PCRef refNode = new CRef();
   PCBook me     = this;
   if (_mark == aName) {
      if (theFile->FindKey(key, refNode))
         pFind->FindInsert(me, refNode, Pchar(""), 0);
      }
   else {
      if (theFile->FindKey(key, refRef) &&
          theFile->GetFirstRef(refRef->loc, refNode)) do
         if (!(pFind->FindInsert(me, refNode, Pchar(""), 0)))
            break;
         while (theFile->GetNextRef(refNode));
      }

   return (pFind->Number() != 0);
   } // FindBookKey

//---------------------------------------------------------------------
// Add the specified key to the farBook dictionary.
//
// Marks are treated as follows:
//   * aName type keys must be unique.
//   * other types need not be unique.
//
bool CLocalBook::MakeBookKey(AMark  _mark,
                             Pchar  _key,
                             PCRef _ref,
                             bool   _trunc) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.MakeBookKey: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   if (StrEmpty(_key)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.MakeBookKey: Key empty"));
#endif
      Error(FVE_EMPTYKEY);
      return false;
      }

   // If the link is in the dictionary, don't add it again!
   PCList fList = new CFindList();
   if (FindBookKey(_mark, _key, fList, _trunc)) {
      pCFindList findList = pCFindList(fList.getPtr());
      if (findList->FindRef(_ref))
         return true;
      }
   fList = 0;

   ulong szKey = theFile->SizeKey();

   // Check for key too long while truncation not permitted
   if ((!_trunc) && strlen(_key) >= szKey-1) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.MakeBookKey: Key too long (%s)"), _key);
#endif
      Error(FVE_LONGKEY);
      return false;
      }

   // Prepare the key
   char key[sizeString];
   Strcpy(key, _key, sizeString);
   prepKey(_mark, key, sizeString, szKey);

   if (_mark == isPrepped)
      _mark = _key[0];

   PCRef lref = new CRef(0);
   if (!Lock(lref, waitForLock)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.MakeBookKey: Could not lock farBook"));
#endif
      return false;
      }

   // Record the key in the dictionary
   PCRef ref = new CRef();
   if (_mark == aName) {
      theFile->InsertKey(key, _ref);
      if (theFile->Error() != FVE_OK)
         Error(FVE_DUPKEY);
      }
   else {
      // look for the key in the dictionary
      if (theFile->FindKey(key, ref)) {
         if (!(theFile->InsertRef(ref->loc, _ref))) {
#ifndef FV_NOLOG
            SysLog::Log('D', Pchar("CLocalBook.MakeBookKey: "
                             "Could not insert ref for new key (%s)"),
                             key);
#endif
            Unlock(lref);
            return false;
            }
         }
      else {                                       // this is a new key
         ref->loc = theFile->InsertRef(ref->loc, _ref);
         if (!(ref->loc)) {
#ifndef FV_NOLOG
            SysLog::Log('D', Pchar("CLocalBook.MakeBookKey: "
                             "Could not insert ref for existing key (%s)"),
                             key);
#endif
            Unlock(lref);
            return false;
            }
         theFile->InsertKey(key, ref);
         }
      if (theFile->Error() != FVE_OK)
         Error(FVE_DUPKEY);
      }

   Unlock(lref);

   return (err == FVE_OK);
   } // MakeBookKey

//---------------------------------------------------------------------
// Delete the key from the farBook index
//
bool CLocalBook::DeleteBookKey(AMark  _mark,
                               Pchar  _key,
                               PCRef _ref,
                               bool   _trunc) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.DeleteBookKey: "
                  "Previous error (%ld) not handled."),
#endif
                  err);
      Assert(false);
      }


   if (StrEmpty(_key)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.DeleteBookKey: Key empty"));
#endif
      Error(FVE_EMPTYKEY);
      return false;
      }

   ulong szKey = theFile->SizeKey();

   // Check for key too long while truncation not permitted
   if ((!_trunc) && strlen(_key) >= szKey-1) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.DeleteBookKey: Key too long (%s)"), _key);
#endif
      Error(FVE_LONGKEY);
      return false;
      }

   // Prepare the key
   char key[sizeString];
   Assert(strlen(_key) < sizeString-2);
   Strcpy(key, _key, sizeString);
   prepKey(_mark, key, sizeString, szKey);

   if (_mark == isPrepped)
      _mark = _key[0];

   PCRef lref = new CRef(0);
   if (!Lock(lref, waitForLock)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.DeleteBookKey: Could not lock farBook"));
#endif
      return false;
      }

   PCRef ref = new CRef();
   if      (_mark == aName)
      theFile->DeleteKey(key, _ref);
   else if (theFile->FindKey(key, ref) &&
            theFile->DeleteRef(ref->loc, _ref))
      theFile->DeleteKey(key, ref);

   Unlock(lref);

   return (err == FVE_OK);
   } // DeleteBookKey

//---------------------------------------------------------------------
// Rename the current iNode
//
bool CLocalBook::RenameNode(Pchar _oldName, Pchar _newName, bool _trunc) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.RenameNode: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   if (StrEmpty(_oldName)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.RenameNode: Old name empty"));
#endif
      Error(FVE_EMPTYKEY);
      return false;
      }

   if (StrEmpty(_newName)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.RenameNode: New name empty"));
#endif
      Error(FVE_EMPTYKEY);
      return false;
      }

   ulong szKey = theFile->SizeKey();

   // Check for keys too long while truncation not permitted
   if (!_trunc) {
      if (strlen(_oldName) >= szKey-1) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.RenameNode: Old name too long (%s)"),
                          _oldName);
#endif
         Error(FVE_LONGKEY);
         return false;
         }
      if (strlen(_newName) >= szKey-1) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.RenameNode: New name too long (%s)"),
                          _newName);
#endif
         Error(FVE_LONGKEY);
         return false;
         }
      }

   PCRef ref = new CRef();

   // Prepare the old name
   char oldName[sizeString];
   Assert(strlen(_oldName) < sizeString-2);
   Strcpy(oldName, _oldName, sizeString);
   prepKey(aName, oldName, sizeString, szKey);

   // Prepare the new name
   char newName[sizeString];
   Assert(strlen(_newName) < sizeString-2);
   Strcpy(newName, _newName, sizeString);
   prepKey(aName, newName, sizeString, szKey);

   PCRef lref = new CRef(0);
   if (!Lock(lref, waitForLock)) {
#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("CLocalBook.RenameNode: Could not lock farBook"));
#endif
      return false;
      }

   theFile->DeleteKey(oldName, ref);
   if (theFile->Error() == FVE_OK) {
      theFile->InsertKey(newName, ref);
      if (theFile->Error() != FVE_OK)
         Error(FVE_DUPKEY);
      }
   else
      Error(FVE_NOFIND);

   Unlock(lref);

   return (err == FVE_OK);
   } // RenameNode

//---------------------------------------------------------------------
//                            iNode i/o
//---------------------------------------------------------------------

//---------------------------------------------------------------------
// Do not call this method directly in a shared farBook. Use
// ReadPartWithLock, WritePartWithLock and CreateNodeWithLock instead.
//
bool CLocalBook::openNode(PCRef _ref, ABookAction _action) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.openNode: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   iDesc  = _ref;
   action = _action;

   if     (action == readAction) {
      // Prepare for iNode read
      // Read descriptor from farBook
      if (iDesc.loc == 0 ||
          iDesc.loc > theFile->Size() -
                      (aDesc::NrParts() * sizeof(ALNdx))) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.openNode: Location out of range (%d)"),
                          iDesc.loc);
#endif
         Error(FVE_LOCKRANGEERR);
         return false;
         }

      if (shared)
         theFile->Seek(iDesc.loc+1, fromHead);
      else
         theFile->Seek(iDesc.loc,   fromHead);
      if (!(lDesc.ReadDesc(*(theFile.getPtr())))) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.openNode: Could not read desc at %d"),
                          iDesc.loc);
#endif
         return false;
         }
      }
   else if (action == writeAction) {
      // Prepare for iNode write
      Assert(iDesc.loc);        // new iNode request - use createAction

      // Read the descriptor from the farBook
      if (shared)
         theFile->Seek(iDesc.loc+1, fromHead);
      else
         theFile->Seek(iDesc.loc,   fromHead);
      if (!(lDesc.ReadDesc(*(theFile.getPtr())))) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.openNode: Could not read desc at %d"),
                          iDesc.loc);
#endif
         return false;
         }
      }
   else if (action == createAction) {
      // Locate to the end of the farBook
      theFile->Seek(0, fromTail);
      iDesc.Set(theFile->Where());

      // Add the lock byte if needed
      if (shared)
         theFile->WriteBlock(Pvoid(&lockByte), sizeof(lockByte));

      // Write the descriptor to the farBook (just a placeholder at this
      // point so that we can safely write keys for this iNode)
      lDesc.Clear();
      if (!(lDesc.WriteDesc(*(theFile.getPtr())))) {
#ifndef FV_NOLOG
         SysLog::Log('D', Pchar("CLocalBook.openNode: Could not write desc to %d"),
                          iDesc.loc);
#endif
         return false;
         }
      }

   _ref->Loc(iDesc.loc);
   return true;
   } // openNode

//---------------------------------------------------------------------
// Close the iNode in the farBook to update the descriptor
// (MUST BE CALLED after writing or creating a iNode)
//
// Do not call this method directly in a shared farBook. Use
// ReadPartWithLock, WritePartWithLock and CreateNodeWithLock instead.
//
bool CLocalBook::closeNode(void) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.closeNode: "
                  "Error (%ld) detected."),
                  err);
#endif
      }

   if (action != writeAction &&
       action != createAction)
      return true;

   // Write the updated descriptor to the farBook
   if (shared)
      theFile->Seek(iDesc.loc+1, fromHead);
   else
      theFile->Seek(iDesc.loc,   fromHead);
   if (!(lDesc.WriteDesc(*(theFile.getPtr()))))
      return false;

   return true;
   } // closeNode

//---------------------------------------------------------------------
// Create an object and read its data
//
// Do not call this method directly in a shared farBook. Use
// ReadPart(WithLock), WritePart(WithLock) and CreateNode(WithLock)
// instead.
//
PCAtom CLocalBook::readPart(void) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.readPart: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   PCAtom part = new CBlock();

   // Get the data for the object
   ulong where = lDesc.Where(ContentPart);
   if (where <= iDesc.loc || where >= theFile->Size()) {
      Error(FVE_LOCKRANGEERR);
      return 0;                                 // Valid range check
      }

   theFile->Seek(where, fromHead);
   PCAtom f = pCAtom(theFile);
   if (!(part->ReadFile(f, true))) {
      Error(part->Error());
      return 0;
      }

   return part;
   } // readPart

//---------------------------------------------------------------------
// Lock, open, and read the specified part
//
PCAtom CLocalBook::ReadPart(PCRef _ref) {
   PCRef ref = new CRef(_ref->loc+1);

   if (!Lock(ref, waitForLock))
      return 0;

   if (!openNode(_ref, readAction)) {
      Unlock(ref);
      return 0;
      }

   PCAtom p = readPart();
   ulong e = err;

   closeNode();
   Unlock(ref);
   Error(e);

   return p;
   } // ReadPart

//---------------------------------------------------------------------
// the fields in the table in sizes specify the amount of space to
// allocate to store an object of the specified size. The algorithm
// searches the record shown below left-to-right for the ith value that
// is greater or equal to the specified size. Then it obtains the
// corresponding ith field from the sizes record to use as the
// percentage of excess to be allocated.
//
// |  1|   2|   3|    4|    5|     6|      7|
// |256|1024|4096|16384|65536|262144|1048576|etc
//
// If a corresonding percentage is not available (e.g., the  ith field
// is empty), then zero is returned.
//
ulong CLocalBook::calcSize(ulong _size) {
   ulong size = 256;
   ushort nr  = sizes->Number();
   if (nr == 0)
      return Max(256, (_size*5)/4);
   else
      for (ushort i = 1; i <= nr; i++) {
         if (_size <= size) {
            ulong calc = Max(256, (_size * (100 + sizes->GetLongField(i))) / 100);
            return calc;
            }
         size *= 4;
         }

   return 0;
   } // calcSize

//---------------------------------------------------------------------
// Write the data in the specified object
//
// Note: if at some point you want to provide versioning services for
// data stored in the farBook, you need to revisit the space reuse
// support here.
//
// Do not call this method directly in a shared farBook. Use
// ReadPartWithLock, WritePartWithLock and CreateNodeWithLock instead.
//
bool CLocalBook::writePart(PCAtom _part) {
   Error(theFile->Error());
   if (err != FVE_OK) {
#ifndef FV_NOLOG
      SysLog::Log('D',
                  Pchar("CLocalBook.writePart: "
                  "Previous error (%ld) not handled."),
                  err);
#endif
      Assert(false);
      }

   Assert(action != readAction);
   Assert(_part);

   // Record the type of the atom in the descriptor
   lDesc.Type(ContentPart, iBlock);

   // Write the data to the farBook file
   PCAtom f = pCAtom(theFile);

   // Determine if new space needs to be allocated for the part
   ulong sizePart = _part->BookSize();
   if (lDesc.Size(ContentPart) < sizePart) {
      lDesc.Size(ContentPart, calcSize(sizePart));

      // Lock the entire farBook when extending it
      PCRef ref = new CRef(0);
      if (!Lock(ref, waitForLock))
         return false;

      theFile->Seek(0, fromTail);
      lDesc.Where(ContentPart, theFile->Where());

      _part->WriteFile(f, true);

      // Now write filler to the farBook
      ulong sizeSpace = lDesc.Size(ContentPart);
      if (sizeSpace) {
         sizeSpace -= sizePart;
         char fill[sizeLine];
         memset(fill, 0, sizeLine);
         while (sizeLine <= sizeSpace) {
            theFile->WriteBlock(fill, sizeLine);
            sizeSpace -= sizeLine;
            }
         if (sizeSpace > 0)
            theFile->WriteBlock(fill, sizeSpace);
         }

      Unlock(ref);

#ifndef FV_NOLOG
      SysLog::Log('B',
                  Pchar("CLocalBook.writePart: "
                  "type=%d, "
                  "loc=%ld, "
                  "%ld used, "
                  "%ld allocated)"),
                  _part->Ima(),
                  iDesc.loc,
                  sizePart,
                  lDesc.Size(ContentPart));
#endif
      }
   else { // reuse the space to store the part
      theFile->Seek(lDesc.Where(ContentPart), fromHead);
      _part->WriteFile(f, true);
      Error(_part->Error());
      }

   return (err == FVE_OK);
   } // writePart

//---------------------------------------------------------------------
// Lock, open, and write the specified part
//
bool CLocalBook::WritePart(PCRef _ref, PCAtom _part) {
   PCRef ref = new CRef(_ref->loc+1);

   if (!Lock(ref, waitForLock))
      return false;

   if (!openNode(_ref, writeAction)) {
      Unlock(ref);
      return false;
      }

   bool b = writePart(_part);

   closeNode();
   Unlock(ref);

   return b;
   } // WritePart

//---------------------------------------------------------------------
// Lock, open, and create the specified node
//
bool CLocalBook::CreateNode(PCRef _ref, PCAtom _body) {

   PCRef ref = new CRef(0);

   if (!Lock(ref, waitForLock))
      return false;

   if (!openNode(_ref, createAction)) {
      Unlock(ref);
      return false;
      }

   bool b = writePart(_body);

   closeNode();
   Unlock(ref);

   return b;
   } // CreateNode

//---------------------------------------------------------------------
// Get content if in cache, else use ReadPartWithLock to get it
//
PCAtom CLocalBook::GetContent(PCRef _ref) {
   return ReadPart(_ref);
   } // GetContent

