
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//********************************************************************
// misc.cpp
//*********************************************************************

#include "prefixb.h"
#include "atom.h"
#include "common.h"
#include "misc.h"
#include "record.h"

//=====================================================================
// Date and time services
//

//---------------------------------------------------------------------
// Convert year, month and day into date string (mm/dd/yyyy)
//
void DateStr(long _year, long _month, long _day,
             Pchar _date, short _size) {
   Assert(_size >= 11);
   sprintf(_date, "%ld/%02ld/%ld", _month, _day, _year);
   } // DateStr

//---------------------------------------------------------------------
// Convert hour, minute and second into time string (hh/mm/ss)
//
void TimeStr(long _hour, long _minute, long _second,
             Pchar _time, short _size) {
   Assert(_size >= 11);
   sprintf(_time, "%ld:%02ld:%02ld", _hour, _minute, _second);
   } // TimeStr

#ifdef FV_UNIX
//---------------------------------------------------------------------
// Convert seconds since 00:00:00 UTC, Jan 1, 1970 to date and time
// for Unix file times. Return as record of two fields: date and time.
//
PCRecord SecsToDateAndTime(time_t _secs) {
   char work[sizeLine];
   PCRecord dt = new CRecord(2, Pchar("|"));

   tm* t        = ::localtime(&_secs);

   // Get the date
   DateStr(t->tm_year + 1900, t->tm_mon, t->tm_mday, work, sizeLine);
   dt->PutField(1, work);

   // Get the time
   TimeStr(t->tm_hour, t->tm_min, t->tm_sec, work, sizeLine);
   dt->PutField(2, work);

   return dt;
   } // SecsToDateAndTime

#endif // FV_UNIX

//=====================================================================
// This pair of routines implements a simple but effective encryption
// algorithm for long strings or files. They use an arbitrarily long
// secret key to encrypt by scrambling the plaintext and descrambling
// the encrypted text. This approach is best used with longer files
// and binary files. Short strings can be easily decrypted by inspection.
// Note that it depends on the factorial of the number of bytes in the
// scramble.

//---------------------------------------------------------------------
//
void Encrypt(Pchar _key, Pchar _text) {
   int k = strlen(_key);
   int l = strlen(_text);

   // Encryption loop
   // encrypt _text using _key
   int j = 0;
   for (int i = 0; i < l; i++) {
      int n     = (j % k);
      int m     = ((i + _key[n]) % l);
      char temp = _text[m];
      _text[m]  = _text[i];
      _text[i]  = temp;
      j++;
      }
   } // Encrypt

//---------------------------------------------------------------------
//
void Decrypt(Pchar _key, Pchar _text) {
   int k = strlen(_key);
   int l = strlen(_text);

   // Decryption loop
   // decrypt _text using _key
   int j = l-1;
   for (int i = l-1; i >= 0; i--) {
      int n     = (j % k);
      int m     = ((i + _key[n]) % l);
      char temp = _text[m];
      _text[m]  = _text[i];
      _text[i]  = temp;
      j--;
      if (j < 0)
         j = l-1;
      }
   } // Decrypt

//=====================================================================
// File name services
//

//---------------------------------------------------------------------
//
void CutDrive(Pchar _name) {
   Assert(_name);

   if (!StrEmpty(_name))
      if (ChrPos(drivDelim[0], _name) > 0) {
         char work[sizeString];
         After(Pchar(drivDelim), _name, work, sizeString);
         strcpy(_name, work);
         }
   } // CutDrive

//---------------------------------------------------------------------
//
void GetDrive(Pchar _drive, Pchar _name, short _size) {
   Assert(_drive);
   Assert(_name);

   _drive[0] = 0;

   if (!StrEmpty(_name))
      if (ChrPos(drivDelim[0], _name) > 0)
         Before(Pchar(drivDelim), _name, _drive, _size);
   } // GetDrive

//---------------------------------------------------------------------
// Remove the path from the specified file name.
//
void CutFilePath(Pchar _name) {
   if (StrEmpty(_name))
      return;
      
   char name[sizeString];
   Strcpy(name, _name, sizeString);
   ChrRep('/', pathDelim[0], name);

   int len = strlen(name);
   if (!len)
      return;

   for (int i = len-1; i; i--)
      if (name[i] == pathDelim[0]) {
         strcpy(_name, name+i+1);
         return;
         }
   } // CutFilePath

//---------------------------------------------------------------------
// Parse a file name into _path and _name (the file name
// is not affected). If _path is NULL, the path is discarded.
//
void GetFilePath(Pchar _path, Pchar _name, short _size) {
   char name[sizeString];
   if (StrEmpty(_name))
      name[0] = 0;
   else {
      Strcpy(name, _name, sizeString);
      ChrRep('/', pathDelim[0], name);
      }

   PCRecord rec = new CRecord(name, Pchar(pathDelim));

   // Copy each path element to _path
   if (_path) {
      _path[0] = '\0';
      char work[sizeString];
      short nr = rec->Number();
      for (short i = 1; i < nr; i++) {
         rec->GetField(i, work, sizeString);
         Strcat(_path, work,      _size);
         Strcat(_path, Pchar(pathDelim), _size);
         }
      }
   } // GetFilePath

//---------------------------------------------------------------------
//
short findExtension(Pchar _name) {
   short i;
   for (i = short(strlen(_name))-1; i >= 0; i--)
      if (_name[i] == *extDelim)
         break;
   if (i > 0 && strlen(_name) - i > 0)
      return i;
   else
      return 0;
   } // findExtension

//---------------------------------------------------------------------
// Remove the extension from a file name
//
void CutExtension(Pchar _name) {
   short i = findExtension(_name);
   if (i > 0)
      _name[i] = 0;
   } // CutExtension

//---------------------------------------------------------------------
// Return the extension from the specified file name
//
void GetExtension(Pchar _name, Pchar _ext, short _size) {
   short i = findExtension(_name);
   if (i > 0) {
      Strcpy(_ext, _name+i+1, _size);
      }
   else
      _ext[0] = 0;
   } // GetExtension

//---------------------------------------------------------------------
// Combine the name parts given and return the result in _name. Don't

// use _path if _name contains a path delimiter. Likewise, don't use
// _extension if _name contains an extension delimiter.
//
void CatName(Pchar _path, Pchar _name, Pchar _ext, short _size) {
   Assert(!StrEmpty(_name));

   char path[sizeString];
   if (StrEmpty(_path))
      path[0] = 0;
   else {
      Strcpy(path, _path, sizeString);
      ChrRep('/', pathDelim[0], path);
      }

   char work[sizeString];
   work[0] = '\0';

   // append the path
   if (!StrEmpty(path) && ChrPos(pathDelim[0], _name) < 0) {
      Strcat(work, path, sizeString);
      // add a path delimiter if needed
      if (work[strlen(work)-1] != pathDelim[0])
          Strcat(work, Pchar(pathDelim), sizeString);
      }

   // append the name
   if (_name)
      Strcat(work, _name, sizeString);

   // append the extension
   if (!StrEmpty(_ext) && ChrPos(extDelim[0], _name) < 0) {
      // add an extension delimiter if needed
      if (_ext[0] != extDelim[0])
         Strcat(work, extDelim, sizeString);
      Strcat(work, _ext, sizeString);
      }
   Strcpy(_name, work, _size);
   } // CatName

//---------------------------------------------------------------------
// Parse the file name into path, name, and extension
// Assumes that _path and _ext are large enough. You
// can, however, pass NULL for _path and/or _ext just
// to discard those parts.
//
void ParseName(Pchar _path, Pchar _name, Pchar _ext,
               short _szPath, short _szExt) {
   if (_path)
      _path[0] = 0;
   if (_ext)
      _ext[0]  = 0;

   char name[sizeString];
   if (StrEmpty(_name))
      name[0] = 0;
   else {
      Strcpy(name, _name, sizeString);
      ChrRep('/', pathDelim[0], name);
      }

   Pchar p;

   // obtain the extension
   p = strrchr(name, extDelim[0]);
   if (p) {
     if (_ext)
        Strcpy(_ext, p+1, _szExt);
     *p = 0;
     }

   // obtain the path
   p = strrchr(name, pathDelim[0]);
   if (p) {
      *p = 0;
      if (_path) {
         Strcpy(_path, name,      _szPath);
         Strcat(_path, Pchar(pathDelim), _szPath);
         }
      strcpy(_name, p+1);
      }
   else
      strcpy(_name, name);
   } // ParseName

//---------------------------------------------------------------------
// Return valid file name
//
void MakeFileName(Pchar _name, short _size) {
   GetID(_name, _size);
   } // MakeFileName

//---------------------------------------------------------------------
// Concatenate the directory name to the path
//
void CatPath(Pchar _path, Pchar _name, short _size) {
   if (StrEmpty(_path))
      return;
   if (StrEmpty(_name))
      return;

   if      (StrEmpty(_path))
      ChrCpy(_path, pathDelim[0]);

   else if (_path[strlen(_path)-1] != pathDelim[0])
      ChrCat(_path, pathDelim[0], _size);

   Strcat(_path, _name, _size);
   } // CatPath

//=====================================================================
// Drive and directory services
//

//---------------------------------------------------------------------
// Return true if the data string matches the pattern or if no pattern
// is specified. Supports "*" to match the rest of the string up to a
// period or the end of the string, and "?" to match any character.
// (Standard DOS file match with wild cards.)
//
bool MatchPattern(const Pchar _pat, const Pchar _data) {
   // Check for null cases
   if (StrEmpty(_pat))
      return true;

   if (StrEmpty(_data))
      return false;

   // Prepare the pattern
   char pat[sizeString];
   Strcpy(pat, _pat, sizeString);
   CapStr(pat);

   // Prepare the data
   char dat[sizeString];
   Strcpy(dat, _data, sizeString);
   CapStr(dat);

   // Match the pattern with the data, char-for-char
   Pchar pp    = pat;                                  // pattern index
   Pchar dp    = dat;                                     // data index
   while (*pp && *dp)
      // Check for multi-char wildcard
      if      (*pp == '*') {
         // Check for mis-formed pattern
         if (*(pp+1) && *(pp+1) != '.')
            return false;

         // Search data for '.' or eos
         while (*dp && *dp != '.')
            dp++;
         if (*dp)
            pp++;
         }

      // Check for single-char wildcard
      else if (*pp == '?')
         if (*dp) {
            pp++;
            dp++;
            }
         else
            return false;

      // Check for per-char not match
      else if (*pp != *dp)
         return false;

      // characters match
      else {
         pp++;
         dp++;
         }

   if (*pp || *dp)                // both pointers should come out even
      return false;
   else
      return true;
   } // MatchPattern

#if   defined(FV_MACINTOSH)
//---------------------------------------------------------------------
// Initialize the specified MacOS FSSpec
// struct using the specified file name.
//
bool InitSpec(Pchar _path, FSSpec &_spec) {
   uchar      path[sizeString];
   char       name[sizeString];

   // Parse the specified path
   Strcpy(name, _path, sizeString);
   GetFilePath(Pchar(path+1), name, sizeString);
   path[0] = strlen(Pchar(path+1));

   // Set up to retrieve information about the specified path
   CInfoPBRec cpb;
   memset(&cpb, 0, sizeof(CInfoPBRec));
   cpb.dirInfo.ioNamePtr   = path;
   cpb.dirInfo.ioFDirIndex = -1;

   // Make sure that the specified path is actually a path
   OSErr err = ::PBGetCatInfoSync(&cpb);
   if (err == noErr) {                               // Is a valid path
      // Check that path is a folder
      if (cpb.hFileInfo.ioFlAttrib & ioDirMask) {
         // Check if the specified file is in the folder
         Strcpy(Pchar(path+1), name, sizeString-1);
         path[0] = strlen(name);

         memset(&cpb, 0, sizeof(CInfoPBRec));
         cpb.dirInfo.ioNamePtr = path;
         cpb.dirInfo.ioDrParID = cpb.dirInfo.ioDrDirID;

         err = ::PBGetCatInfoSync(&cpb);
         // if no error, then the folder we are looking for
         // exists and we now have the information about
         //	the parID we need for the spec.
         }
      // else the partial path is a file
      }

   _spec.vRefNum = cpb.dirInfo.ioVRefNum;
   _spec.parID   = cpb.dirInfo.ioDrDirID;

   return (err == noErr);
   } // InitSpec
#endif // FV_MACINTOSH

//---------------------------------------------------------------------
// Return the current drive and path in _path
//
void GetCurrentDir(Pchar _path, int _size) {

#if   defined(FV_MSWINDOWS)
   ::GetCurrentDirectory(_size, _path);

#elif defined(FV_MACINTOSH)
   uchar vName[sizeString];
   short vRefNum;
   long  dirID;
   ::HGetVol(&vName[0], &vRefNum, &dirID);

   // Set up to retrieve information
   CInfoPBRec cpb;
   uchar dName[sizeString];
   memset(&cpb, 0, sizeof(CInfoPBRec));
   cpb.dirInfo.ioVRefNum = vRefNum;
   cpb.dirInfo.ioDrParID = dirID;
   cpb.dirInfo.ioNamePtr = dName;

   // Start at the folder level, and loop until we reach the root level
   memset(_path, 0, _size);
   while (cpb.dirInfo.ioDrDirID != fsRtDirID) {
      cpb.dirInfo.ioFDirIndex = -1;
      cpb.dirInfo.ioDrDirID   = cpb.dirInfo.ioDrParID;

      OSErr err = ::PBGetCatInfoSync(&cpb);

      PtoCstr(dName);
      Strcat(Pchar(dName), Pchar(pathDelim), sizeString);
      Strcat(Pchar(dName), _path,            sizeString);
      Strcpy(_path,        Pchar(dName),     _size);
      CtoPstr(Pchar(dName));
      }

#elif defined(FV_UNIX)
   ::getcwd(_path, _size);

#endif // FV_UNIX
   } // GetCurrentDir

//---------------------------------------------------------------------
// Change to the drive specified in the first
// character of the name given. ex. c:\windows\system.
// Return true if successful. Return false if the
// second character is not a colon, or if unsuccessful.
//
bool ChangeDisk(Pchar _name) {

#if   defined(FV_MSWINDOWS)
   if (strlen(_name) < 2 || _name[1] != ':')
      return true;

   char diskLetter = _name[0];
   CapChr(diskLetter);
   int disk  = int(diskLetter) - int('A');
#if defined __BORLANDC__
   ::setdisk(disk);

#else
   _chdrive(disk);
#endif
   return true;

#elif defined(FV_MACINTOSH)
   Assert(false);
   return false;

#elif defined(FV_UNIX)
   Assert(false);
   return false;

#endif // FV_UNIX
   } // ChangeDisk

//---------------------------------------------------------------------
// Make the specified directory.
// Return true if successful.
//
bool MakeDir(Pchar _name, Rulong _err) {
   if (StrEmpty(_name))
      return false;
      
   char name[sizeString];
   Strcpy(name, _name, sizeString);
   ChrRep('/', pathDelim[0], name);

#if   defined(FV_MSWINDOWS)
   if (::mkdir(name) != 0) {
      _err = errno;
      return false;
      }

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   if (::mkdir(name, 0770) != 0) {
      _err = errno;
      return false;
      }

#endif // FV_UNIX
   return true;
   } // MakeDir

//---------------------------------------------------------------------
// Remove the specified directory.
// Return true if successful.
//
bool RemoveDir(Pchar _name, Rulong _err) {
   if (StrEmpty(_name))
      return false;
      
   char name[sizeString];
   Strcpy(name, _name, sizeString);
   ChrRep('/', pathDelim[0], name);

#if   defined(FV_MSWINDOWS)
   if (::rmdir(name) != 0) {
      _err = errno;
      return false;
      }

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   if (::rmdir(name) != 0) {
      _err = errno;
      return false;
      }

#endif // FV_UNIX
   return true;
   } // RemoveDir

//---------------------------------------------------------------------
// Change to the specified directory.
// Return true if successful.
//
bool ChangeDir(Pchar _name, Rulong _err) {
   if (StrEmpty(_name))
      return false;

   char name[sizeString];
   Strcpy(name, _name, sizeString);
   ChrRep('/', pathDelim[0], name);

#if   defined(FV_MSWINDOWS)

   if (::chdir(name) != 0) {
      _err = errno;
      return false;
      }

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   if (::chdir(name) != 0) {
      _err = errno;
      return false;
      }

#endif // FV_UNIX
   return true;
   } // ChangeDir

//---------------------------------------------------------------------
// Return true if the specified file is in the specified path. Set
// fname to the full name of the file.
//
bool FindFileInPath(Pchar _name, Pchar _path,
                    Pchar _fname, short _size) {
   if (StrEmpty(_name))
      return false;

   ulong err;

   if (FileExists(_name, err)) {
      Strcpy(_fname, _name, _size);
      return true;
      }

   // prepare the path string
   PCRecord path = new CRecord(_path, Pchar(";"));
   char pathName[sizeString];
   char name[sizeString];

   if (path->Number() == 1) {
      // path is specified
      path->GetField(1, pathName, sizeString);
      Strcpy(_fname, _name, _size);
      CatName(pathName, _fname, 0, _size);

      if (FileExists(_fname, err))                    // found the file
         return true;
      }
   // traverse the path list (record)
   else for (short i = 2; i <= path->Number(); i++) {
      path->GetField(i, pathName, sizeString);
      Strcpy(_fname, _name, _size);
      CatName(pathName, _fname, 0, _size);

      if (FileExists(_fname, err))                    // found the file
         return true;
      }

   return false;
   } // FindFileInPath

//---------------------------------------------------------------------
// Return true if file exists.
//
bool FileExists(Pchar _name, Rulong _err) {
   if (StrEmpty(_name))
      return false;
      
   char name[sizeString];
   Strcpy(name, _name, sizeString);
   ChrRep('/', pathDelim[0], name);

#if   defined(FV_MSWINDOWS)
   if (::access(name, 0) != 0) {
      _err = errno;
      return false;
      }

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   if (::access(name, 0) != 0) {
      _err = errno;
      return false;
      }

#endif // FV_UNIX
   return true;
   } // FileExists

//---------------------------------------------------------------------
//
FileType GetFileType(Pchar _fileName) {
#if defined(FV_MSWIN32)
   WIN32_FIND_DATA findData;
   HANDLE searchHandle = ::FindFirstFile(_fileName, &findData);
   if ((findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ==
        FILE_ATTRIBUTE_DIRECTORY)
      return dirFile;
   else
      return datFile;
#endif // FV_MSWIN32
   } // GetFileType

//---------------------------------------------------------------------
// Get information about specified file and return it in the specified
// record. Format is
//   |fileName|crdate|crtime|crsecs|moddate|modtime|modsecs|size|dir/data
//
bool GetFileInfo(Pchar _fileName, PCRecord _data) {
#if defined(FV_MSWIN32)
   WIN32_FIND_DATA findData;
   HANDLE searchHandle = ::FindFirstFile(_fileName, &findData);
   if (searchHandle == INVALID_HANDLE_VALUE)
      return false;

   // Build the file info record
   // get the file name
   _data->PutField(SRF_NAME, findData.cFileName);

   FILETIME localTime;
   WORD date, time;

   // Get the create date and time
   ::FileTimeToLocalFileTime(&findData.ftCreationTime, &localTime);
   ::FileTimeToDosDateTime(&localTime, &date, &time);

   char swork[128];

   // Convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_CRDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                               // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_CRTIME, swork);


   LARGE_INTEGER secs;
   memcpy(&secs, &findData.ftCreationTime, sizeof(LARGE_INTEGER));
   _data->PutField(SRF_CRSECS, secs.QuadPart);

   // Get the modify date and time
   ::FileTimeToLocalFileTime(&findData.ftLastWriteTime, &localTime);
   ::FileTimeToDosDateTime(&localTime, &date, &time);

   // Convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_MODDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                                // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_MODTIME, swork);

   memcpy(&secs, &findData.ftLastWriteTime, sizeof(LARGE_INTEGER));
   _data->PutField(SRF_MODSECS, secs.QuadPart);

   // Note that size is assumed to be less than 2 ^ 32
   _data->PutField(SRF_SIZE, long(findData.nFileSizeLow));

   // Get the file type
   if ((findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ==
        FILE_ATTRIBUTE_DIRECTORY)
      _data->PutField(SRF_TYPE, Pchar("dir"));
   else
      _data->PutField(SRF_TYPE, Pchar("data"));

#endif // FV_MSWIN32
   return false;
   } // GetFileInfo

//---------------------------------------------------------------------
// Return true if successful. Does not move a file.
//
bool RenameFile(Pchar _oldName, Pchar _newName, Rulong _err) {
   if (StrEmpty(_oldName))
      return false;
   if (StrEmpty(_newName))
      return false;
      
   char oldName[sizeString];
   Strcpy(oldName, _oldName, sizeString);
   ChrRep('/', pathDelim[0], oldName);

   char newName[sizeString];
   Strcpy(newName, _newName, sizeString);
   ChrRep('/', pathDelim[0], newName);

#if   defined(FV_MSWINDOWS)
   if (::rename(oldName, newName) != 0) {
      _err = errno;
      return false;
      }


#elif defined(FV_MACINTOSH)
   Assert(0);

#elif defined(FV_UNIX)
// check that this is correct for UNIX
   if (::rename(oldName, newName) != 0) {
      _err = errno;
      return false;
      }

#endif // FV_UNIX
   return true;
   } // RenameFile

//---------------------------------------------------------------------
// Rename the file using the specified backup extension. If a file
// already exists, then delete it first. Return true if successful.
//
bool BackupFile(Pchar _name, Pchar _backupExt, Rulong _err) {
   if (StrEmpty(_name))
      return false;

   char name[sizeLine];
   Strcpy(name, _name, sizeLine);
   ChrRep('/', pathDelim[0], name);

   CutExtension(name);
   CatName(0, name, _backupExt, sizeLine);
   RemoveFile(name, _err);
   return RenameFile(_name, name, _err);
   } // BackupFile

//---------------------------------------------------------------------
// Return true if successful.
//
bool RemoveFile(Pchar _name, Rulong _err) {
   if (StrEmpty(_name))
      return false;
      
   char name[sizeLine];
   Strcpy(name, _name, sizeLine);
   ChrRep('/', pathDelim[0], name);

#if   defined(FV_MSWINDOWS)
   if (FileExists(name, _err))
      if (::remove(name) != 0) {
         _err = errno;
         return false;
         }

#elif defined(FV_MACINTOSH)
   Assert(false);

#elif defined(FV_UNIX)
   if (FileExists(name, _err))
      if (::remove(name) != 0) {
         _err = errno;
         return false;
         }

#endif // FV_UNIX
   return true;
   } // RemoveFile

//=====================================================================
// File search services
//

//---------------------------------------------------------------------
// (note: this class requires locking)
//
CSearchFile::CSearchFile(void):
   CAtom(),

#if   defined(FV_MSWINDOWS)
   fm   (0),
   ft   (0) {
   searchHandle = 0;

#elif defined(FV_MACINTOSH)
   memset(&cipbr,    0, sizeof(CInfoPBRec));
   memset(fileName,  0, sizeof(Str255));

#elif defined(FV_UNIX)
   dir     (0),
   fileType(datFile) {
//?entry   (?),
//?buff    (?)
   
   pattern[0] = 0;

#endif // FV_UNIX
   } // CSearchFile

//---------------------------------------------------------------------
//
RTTI_CPP(SearchFile, Atom)

//---------------------------------------------------------------------
//
bool CSearchFile::FindFirstName(Pchar _fileName, FileType _fileType) {

#if   defined(FV_MSWINDOWS)
   Assert(!searchHandle);

   // Determine the type of file being searched
   if      (_fileType == datFile) {
      fm = FILE_ATTRIBUTE_DIRECTORY;
      ft = FILE_ATTRIBUTE_DIRECTORY;
      }
   else if (_fileType == dirFile) {
      fm = FILE_ATTRIBUTE_DIRECTORY;
      ft = 0;
      }
   else { // if (_fileType == anyFile)
      fm = 0;
      ft = 1;
      }

   // Search the directory for the file
   searchHandle = ::FindFirstFile(_fileName, &findData);
   if (searchHandle != INVALID_HANDLE_VALUE) {
      bool b = true;
      while (b && ((findData.dwFileAttributes & fm) == ft))
         b = ::FindNextFile(searchHandle, &findData);
      if (b)
         return true;
      }

#elif defined(FV_MACINTOSH)
   fileType = _fileType;

   HFileInfo *fpb = (HFileInfo*) &cipbr;
   Assert(fpb->ioFDirIndex == 0);

   // Get the path part of the file name
   char path[sizeString];
   Strcpy(pattern, _fileName, sizeString);
   GetFilePath(path, pattern, sizeString);

   // handle case when default folder is specified
   if (strlen(path) == 0)
      GetCurrentDir(path, sizeString);

   // Convert the path into a dirID
   FSSpec spec;
   memset(&spec, 0, sizeof(FSSpec));
   InitSpec(path, spec);

   memset(&cipbr, 0, sizeof(CInfoPBRec));
   fpb->ioVRefNum   = spec.vRefNum;
   fpb->ioDirID     = spec.parID;
   fpb->ioNamePtr   = fileName;
   fpb->ioFDirIndex = 1;           // set up to get the first file name

   // Search the folder for the first specified file name
   short rc = ::PBGetCatInfoSync(&cipbr);
   PtoCstr(fileName);
   while (!rc)
      if (((fileType == anyFile)                                       ||
           (fileType == datFile && (fpb->ioFlAttrib & ioDirMask) == 0) ||
           (fileType == dirFile && (fpb->ioFlAttrib & ioDirMask) != 0)) &&
          MatchPattern(pattern, Pchar(fileName)))
         return true;
      else {
         fpb->ioFDirIndex++;
         rc = ::PBGetCatInfoSync(&cipbr);
         }

#elif defined(FV_UNIX)
   Assert(!dir);

   fileType = _fileType;

   // Get the path part of the file name
   char path[sizeString];
   Strcpy(pattern, _fileName, sizeString);
   GetFilePath(path, pattern, sizeString);

   dir = ::opendir(path);
   if (!dir)
      return false;

   // (Note: if you have to use readdir(), you will have to protect it)
   dirent* result;
   while (::readdir_r(dir, &entry, &result) &&
          ::stat(entry.d_name, &buff)) {

      if ((fileType == datFile && S_ISREG(buff.st_mode)) ||
          (fileType == dirFile && S_ISDIR(buff.st_mode)) ||
          (fileType == anyFile && (S_ISREG(buff.st_mode) ||
                                   S_ISDIR(buff.st_mode)))) {
         if (strcmp(entry.d_name, Pchar("."))  &&
             strcmp(entry.d_name, Pchar("..")) &&
             MatchPattern(pattern, entry.d_name)) {
            // accept this file entry
            return true;
            }
         }
      }

#endif // FV_UNIX
   return false;
   } // FindFirstName

//---------------------------------------------------------------------
bool CSearchFile::FindNextName(void) {

#if   defined(FV_MSWINDOWS)
   if (!searchHandle)
      return false;

   bool b = ::FindNextFile(searchHandle, &findData);
   if (b) {
      while (b && ((findData.dwFileAttributes & fm) == ft))
         b = ::FindNextFile(searchHandle, &findData);
      if (b)
         return true;
      }

#elif defined(FV_MACINTOSH)
   HFileInfo *fpb = (HFileInfo*) &cipbr;
   Assert(fpb->ioFDirIndex);

   // Search the folder for the next specified file name
   fpb->ioFDirIndex++;
   short rc = ::PBGetCatInfoSync(&cipbr);
   PtoCstr(fileName);
   while (!rc)
      if (((fileType == anyFile)                                       ||
           (fileType == datFile && (fpb->ioFlAttrib & ioDirMask) == 0) ||
           (fileType == dirFile && (fpb->ioFlAttrib & ioDirMask) != 0)) &&
          MatchPattern(pattern, Pchar(fileName)))
         return true;
      else {
         fpb->ioFDirIndex++;
         rc = ::PBGetCatInfoSync(&cipbr);
         }

#elif defined(FV_UNIX)
   if (!dir)
      return false;

   // (Note: if you have to use readdir(), you will have to protect it)
   dirent* result;
   while (::readdir_r(dir, &entry, &result) &&
          ::stat(entry.d_name, &buff)) {

      if ((fileType == datFile && S_ISREG(buff.st_mode)) ||
          (fileType == dirFile && S_ISDIR(buff.st_mode)) ||
          (fileType == anyFile && (S_ISREG(buff.st_mode) ||
                                   S_ISDIR(buff.st_mode)))) {
         if (strcmp(entry.d_name, Pchar("."))  &&
             strcmp(entry.d_name, Pchar("..")) &&
             MatchPattern(pattern, entry.d_name)) {
            // accept this file entry
            return true;
            }
         }
      }

#endif // FV_UNIX
   return false;
   } // FindNextName

//---------------------------------------------------------------------
// Return record describing current file
// form is:
//   |filename
//   |crdate|crtime|crsecs
//   |moddate|modtime|modsecs
//   |size|dir/data
//
void CSearchFile::GetSearchData(PCRecord _data) {

   _data->NewRecord(SRF_NRFIELDS, Pchar("|"));

#if   defined(FV_MSWINDOWS)
   // Get the file name
   _data->PutField(SRF_NAME, findData.cFileName);

   FILETIME localTime;
   WORD date, time;

   // Get the create date and time
   ::FileTimeToLocalFileTime(&findData.ftCreationTime, &localTime);
   ::FileTimeToDosDateTime(&localTime, &date, &time);

   char swork[128];

   // Convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_CRDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                               // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_CRTIME, swork);

   LARGE_INTEGER secs;
   memcpy(&secs, &findData.ftCreationTime, sizeof(LARGE_INTEGER));
   _data->PutField(SRF_CRSECS, secs.QuadPart);

   // Get the modify date and time
   ::FileTimeToLocalFileTime(&findData.ftLastWriteTime, &localTime);
  ::FileTimeToDosDateTime(&localTime, &date, &time);

   // Convert date to mm/dd/yyyy format
   DateStr(((date >> 9) & 0x7f) + 1980,                         // year
            (date >> 5) & 0x0f,                                // month
            (date >> 0) & 0x1f,                                  // day
           swork, 32);
   _data->PutField(SRF_MODDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr((time >> 11) & 0x1f,                                 // hour
           (time >>  5) & 0x3f,                                // minute
           (time >>  0) & 0x1f,                               // second
           swork, 32);
   _data->PutField(SRF_MODTIME, swork);

   memcpy(&secs, &findData.ftLastWriteTime, sizeof(LARGE_INTEGER));
   _data->PutField(SRF_MODSECS, secs.QuadPart);

   // Note that size is assumed to be less than 2 ^ 32
   _data->PutField(SRF_SIZE, long(findData.nFileSizeLow));

   // Get the file type
   if ((findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ==
        FILE_ATTRIBUTE_DIRECTORY)
      _data->PutField(SRF_TYPE, Pchar("dir"));
   else
      _data->PutField(SRF_TYPE, Pchar("data"));

#elif defined(FV_MACINTOSH)
   HFileInfo *fpb = (HFileInfo*) &cipbr;
   Assert(fpb->ioFDirIndex);

   // Get the file name
   _data->PutField(SRF_NAME, Pchar(fileName));

   DateTimeRec dt;

   // Get the create date and time
   ::SecondsToDate(fpb->ioFlCrDat, &dt);

   char swork[32];

   // Convert date to mm/dd/yyyy format
   DateStr(dt.year + 1904,
           dt.month,
           dt.day, swork);
   _data->PutField(SRF_CRDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr(dt.hour,
           dt.minute,
           dt.second, swork);
   _data->PutField(SRF_CRTIME, swork);

   _data->PutField(SRF_CRSECS, fpb->ioFlCrDat);

   // Get the modify date and time
   ::SecondsToDate(fpb->ioFlMdDat, &dt);

   // Convert date to mm/dd/yyyy format
   DateStr(dt.year + 1904,
           dt.month,
           dt.day, swork);
   _data->PutField(SRF_MODDATE, swork);

   // Convert time to hh/mm/ss format
   TimeStr(dt.hour,
           dt.minute,
           dt.second, swork);
   _data->PutField(SRF_MODTIME, swork);

   _data->PutField(SRF_MODSECS, fpb->ioFlMdDat);

   // Get the file size
   _data->PutField(SRF_SIZE, fpb->ioFlLgLen);

#elif defined(FV_UNIX)
   Assert(dir);

   // Get the file name
   _data->PutField(SRF_NAME, entry.d_name);

   PCRecord dt;
   char work[sizeLine];

   // Get the create date and time
   dt = SecsToDateAndTime(buff.st_ctime);
   dt->GetField   (1, work, sizeLine);
   _data->PutField(SRF_CRDATE, work);
   dt->GetField   (2, work, sizeLine);
   _data->PutField(SRF_CRTIME, work);
   _data->PutField(SRF_CRSECS, buff.st_ctime);

   // Get the modify date and time
   dt = SecsToDateAndTime(buff.st_mtime);
   dt->GetField   (1, work, sizeLine);
   _data->PutField(SRF_MODDATE, work);
   dt->GetField   (2, work, sizeLine);
   _data->PutField(SRF_MODTIME, work);
   _data->PutField(SRF_MODSECS, buff.st_mtime);

   // Get the file size
   _data->PutField(SRF_SIZE, buff.st_size);

#endif // FV_UNIX
   } // GetSearchData

//---------------------------------------------------------------------
// Close a file name search
//
void CSearchFile::CloseFind(void) {

#if   defined(FV_MSWINDOWS)
   if (searchHandle)
      ::FindClose(searchHandle);
   searchHandle = 0;

#elif defined(FV_MACINTOSH)
   memset(&cipbr,    0, sizeof(CInfoPBRec));
   memset(fileName, 0, sizeof(Str255));

#elif defined(FV_UNIX)
   if (dir)
      ::closedir(dir);
   dir = 0;

#endif // FV_UNIX
   } // CloseFind


