
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// odb.cpp
//*********************************************************************
//
#include "odb.h"
#include "refs.h"
#include "search.h"
#include "syslog.h"

//---------------------------------------------------------------------
//
CODB::CODB(Pchar DBName):
   CAtom (),
   domain(0),
   locked(false)   {
   STRINIT(domain, Pchar(""));
   book = new CLocalBook(DBName);
   result = new CFindList();
   err = 0;
   } // CODB

//---------------------------------------------------------------------
//
CODB::~CODB() {
   } // ~CODB

//---------------------------------------------------------------------
//
RTTI_CPP(ODB, Atom)

//---------------------------------------------------------------------
// Set the domain for the session
//
void CODB::SetDomain(Pchar dom) {
   STRINIT(domain, dom);
   } // SetDomain
   
//---------------------------------------------------------------------
// Create a new ODB
//
bool CODB::Create(long _nrParts,
                  long _nrRefSlots,
                  long _nrHshSlots,
                  long _sizeKey,
                  long _nrKeys) {
   bool ok = book->Create(_nrParts,
                          _nrRefSlots,
                          _nrHshSlots,
                          _sizeKey,
                          _nrKeys);
   err = book->Error();
   return ok;
   } // Create

//---------------------------------------------------------------------
// Open the ODB
//
bool CODB::Open(void) {
   bool ok = book->Open();
   err     = book->Error();

   return ok;
   } // Open

//---------------------------------------------------------------------
// Close the ODB
//
bool CODB::Close(void) {
   if (locked)
      Unlock();

   bool ok = book->Close();
   err     = book->Error();

   return ok;
   } // Close

//---------------------------------------------------------------------
// Lock the ODB
//
bool CODB::Lock(void) {
   if (locked)
      return true;
      
   PCRef ref = new CRef();
   locked    = book->Lock(ref, waitForLock);
   err       = book->Error();

   return locked;
   } // Lock

//---------------------------------------------------------------------
// Unlock the ODB
//
bool CODB::Unlock(void) {
   if (!locked)
      return true;

   PCRef ref = new CRef();
   locked    = !book->Unlock(ref);
   err       = book->Error();

   return !locked;
   } // Unlock

//---------------------------------------------------------------------
// private method to return true if the key is already associated with
// a topic
//
bool CODB::keyPresent(Pchar topicName) {
   char key[sizeString];

   MakeKey(key, domain, UniqueType, topicName, sizeString);
   PCList find = new CFindList();
   return book->FindBookKey(aName, key, find, false);
   } // keyPresent

//---------------------------------------------------------------------
// private method to find the location of the named topic
//
PCRef CODB::getRef(Pchar topicName) {
   // Find the named topic in the book
   PCFindList findList = new CFindList();

   char key[sizeString];

   MakeKey(key, domain, UniqueType, topicName, sizeString);
   if (!findList->FindKey(book, aName, key)) {
      err = findList->Error();
      return 0;
      }

   if (findList->Number() == 0) {
      err = FVE_NOFIND;
      return 0;
      }

   // Obtain the location of the topic
   PCItem item = findList->Item();
   PCFindItem findItem = TYPECAST(item, FindItem);
   PCRef findRef = new CRef(findItem->GetRef());

   return findRef;
   } // getRef

//---------------------------------------------------------------------
// private method to return true if the specified key is associated
// with the topic reference
//
bool CODB::refExists(PCRef ref, Pchar key) {
   // Find the topics associated with the key
   PCFindList findList = new CFindList();
   if (!findList->FindKey(book, anAlias, key))
      return false;

   if (findList->Number() == 0)
      return false;

   // Obtain the location of the topic
   PCItem item = findList->Item();
   PCFindItem findItem = TYPECAST(item, FindItem);
   PCRef findRef = new CRef(findItem->GetRef());

   return (findRef == ref);
   } // refExists

//---------------------------------------------------------------------
// Check if the named topic exists
//
bool CODB::TopicExists(Pchar topicName) {
   return keyPresent(topicName);
   } // topicExists

//---------------------------------------------------------------------
// Add a topic
//
bool CODB::AddTopic(Pchar topic, Pchar topicName) {
   err = 0;

   // Does the key already exist?
   if (keyPresent(topicName))
      return false;

   PCRef ref = new CRef(0);
   PCAtom atom = new CBlock(topic);

   if (!book->CreateNode(ref, atom)) {
      err = book->Error();
      return false;
      }

   char key[sizeString];

   MakeKey(key, domain, UniqueType, topicName, sizeString);
   if (!book->MakeBookKey(aName, key, ref, true)) {
      err = book->Error();
      return false;
      }

   return true;
   } // AddTopic

//---------------------------------------------------------------------
// Add a topic still in a string list
//
bool CODB::AddTopic(PCStrList topic, Pchar topicName) {
   PCBlock blk = new CBlock();
   topic->ToBlock(blk);

   return AddTopic(blk, topicName);
   } // AddTopic

//---------------------------------------------------------------------
// Add a topic already in a block
//
bool CODB::AddTopic(PCBlock topic, Pchar topicName) {
   err = 0;

   // Does the key already exist in the db?
   if (keyPresent(topicName))
      return false;

   PCRef ref = new CRef(0);
   PCAtom atom = TYPECAST(topic, Atom);

   if (!book->CreateNode(ref, atom)) {
      err = book->Error();
      return false;
      }

   char key[sizeString];

   MakeKey(key, domain, UniqueType, topicName, sizeString);
   if (!book->MakeBookKey(aName, key, ref, true)) {
      err = book->Error();
      return false;
      }

   return true;
   } // AddTopic
#ifndef FV_NOXML

//---------------------------------------------------------------------
//
bool CODB::AddTopic(PCNode topic, Pchar topicName) {
   PCBlock blk = new CBlock();
   topic->Save(blk);
   return AddTopic(blk, topicName);
   } // Addtopic
#endif

//---------------------------------------------------------------------
// Add a reference for the specified topic
//
// note: would be nice to check if the specified reference already
// identifies the specified topic
//
bool CODB::AddRef(Pchar topicName, Pchar reference) {
   err = 0;

   // Find the named topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   char key[sizeString];

   // Add the topic reference to the book
   MakeKey(key, domain, AliasType, reference, sizeString);

   // Does the reference already identify the specified topic?
   if (refExists(ref, key))
      return true;

   if (!book->MakeBookKey(anAlias, key, ref, true)) {
      err = book->Error();
      return false;
      }

   return true;
   } // AddRef

//---------------------------------------------------------------------
// Add a set of references for the specified topic
//
// note: would be nice to check if any of the specified references
// already identify the specified topic
//
bool CODB::AddRefs(Pchar topicName, PCStrList referenceList) {
   err = 0;

   // Find the named topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   char key[sizeString];

   if (referenceList->GotoHead()) do {
      Pchar reference = referenceList->StrItem();
      MakeKey(key, domain, AliasType, reference, sizeString);

      // Does the reference already identify the specified topic?
      if (refExists(ref, key))
         return true;

      if (!book->MakeBookKey(anAlias, key, ref, true)) {
         err = book->Error();
         return false;
         }
      } while (referenceList->GotoNext());

   return true;
   } // AddRefs

//---------------------------------------------------------------------
// Establish a relationship between a parent and a child
//
// note: would be nice to check if the relationship already exists.
//
bool CODB::AddLink(Pchar parentName, Pchar childName) {
   err = 0;

   // Find the parent topic in the book
   PCRef parentRef = getRef(parentName);
   if (!parentRef)
      return false;

   // Find the child topic in the book
   PCRef childRef = getRef(childName);
   if (!childRef)
      return false;

   char key[sizeString];

   // Make the parentOf alias for the parent topic
   MakeKey(key, domain, ParentType, childName, sizeString);
   if (!book->MakeBookKey(anAlias, key, parentRef, true)) {
      err = book->Error();
      return false;
      }

   // Make the childOf alias for the child topic
   MakeKey(key, domain, ChildType, parentName, sizeString);
   if (!book->MakeBookKey(anAlias, key, childRef, true)) {
      err = book->Error();
      return false;
      }

   return true;
   } // AddLink

//---------------------------------------------------------------------
// Establish a relationship between a parent and a set of children
//
// note: would be nice to check if the relationships already exist.
//
bool CODB::AddLinks(Pchar parentName, PCStrList childNameList) {
   err = 0;

   // Find the parent topic in the book
   PCRef parentRef = getRef(parentName);
   if (!parentRef)
      return false;

   if (childNameList->GotoHead()) do {
      Pchar childName = childNameList->StrItem();

      // Find the child topic in the book
      PCRef childRef = getRef(childName);
      if (!childRef)
         return false;

      char key[sizeString];

      // Make the parentOf alias for the child topic
      MakeKey(key, domain, ParentType, childName, sizeString);
      if (!book->MakeBookKey(anAlias, key, parentRef, true)) {
         err = book->Error();
         return false;
         }

      // Make the childOf alias for the child topic
      MakeKey(key, domain, ChildType, parentName, sizeString);
      if (!book->MakeBookKey(anAlias, key, childRef, true)) {
         err = book->Error();
         return false;
         }
      } while (childNameList->GotoNext());

   return true;
   } // AddLinks

//---------------------------------------------------------------------
//
bool CODB::RemoveRef(Pchar topicName, Pchar reference) {
   err = 0;

   // Find the named topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   char key[sizeString];

   // Add the topic reference to the book
   MakeKey(key, domain, AliasType, reference, sizeString);

   // Does the reference already identify the specified topic?
   if (refExists(ref, key))
      return true;

   if (!book->DeleteBookKey(anAlias, key, ref, true)) {
      err = book->Error();
      return false;
      }

   return true;
   } // RemoveRef

//---------------------------------------------------------------------
//
bool CODB::RemoveRefs(Pchar topicName, PCStrList referenceList) {
   err = 0;

   // Find the named topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   char key[sizeString];

   if (referenceList->GotoHead()) do {
      Pchar reference = referenceList->StrItem();
      MakeKey(key, domain, AliasType, reference, sizeString);

      // Does the reference already identify the specified topic?
      if (refExists(ref, key))
         return true;

      if (!book->DeleteBookKey(anAlias, key, ref, true)) {
         err = book->Error();
         return false;
         }
      } while (referenceList->GotoNext());

   return true;
   } // RemoveRefs

//---------------------------------------------------------------------
// Update a topic
//
bool CODB::UpdateTopic(Pchar topic, Pchar topicName) {
   err = 0;

   // Find the topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   // Write the updated topic to the book
   PCAtom atom = new CBlock(topic);
   return book->WritePart(ref, atom);
   } // UpdateTopic

//---------------------------------------------------------------------
// Update a topic
//
bool CODB::UpdateTopic(PCBlock topic, Pchar topicName) {
   err = 0;

   // Find the topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   // Write the updated topic to the book
   PCAtom atom = TYPECAST(topic, Atom);
   return book->WritePart(ref, atom);
   } // UpdateTopic
#ifndef FV_NOXML

//---------------------------------------------------------------------
//
bool CODB::UpdateTopic(PCNode topic, Pchar topicName) {
   PCBlock blk = new CBlock();
   topic->Save(blk);
   return UpdateTopic(blk, topicName);
   } // UpdateTopic
#endif

//---------------------------------------------------------------------
// Remove the topic by changing its name to something unique but 
// forgettable. This should be pretty much guaranteed to be unique!?
//
bool CODB::RemoveTopic(Pchar topicName) {
   char uuid[sizeLine];
   Strcpy(uuid, Pchar("3.1415926.$$$."), sizeLine);
   GetID(uuid+strlen(uuid), sizeLine-strlen(uuid));
   bool ok = book->RenameNode(topicName, uuid);
if (!ok)
SysLog::Log('D', Pchar("RemoveTopic: <%s> to <%s> error=%d"), topicName, uuid, book->Error());
   return ok;
   } // RemoveTopic

//---------------------------------------------------------------------
// Get the named topic
//
PCBlock CODB::GetTopic(Pchar topicName) {
   // Find the topic in the book
   PCRef ref = getRef(topicName);
   if (!ref)
      return false;

   // Read the topic from the book
   PCAtom atom = book->ReadPart(ref);
   if (!atom)
      return 0;

   return TYPECAST(atom, Block);
   } // GetTopic

//---------------------------------------------------------------------
// Protected method used by Query method.
//
bool CODB::find(Pchar query, char mark, Pchar type) {
   if (StrEmpty(type)) {
      // Evaluate the query expression to find topics in the book
      PCSearch search = new CSearch(book);
      result = search->Search(domain, query, false);
      if (!result) {
         err = search->Error();
         return false;
         }

      }
   else {
      // Find the specified topic(s) in the book
      char key[sizeString];
      MakeKey(key, domain, type, query, sizeString);
      if (!result->FindKey(book, mark, key)) {
         err = result->Error();
         return false;
         }
      }

   if (result->Number() == 0) {
      err = FVE_NOFIND;
      return false;
      }

   return true;
   } // find

//---------------------------------------------------------------------
// Obtain a set of topics given a key or expression
//
bool CODB::Query(Pchar query, ODB_LINKTYPE type) {
   err = 0;

   result->Flush();

   // Translate the linkType to book key type
   char mark;
   char tp[2];
   tp[0] = 0;

   switch (type) {
      case ODB_NAME:
         return find(query, aName,   UniqueType);
      case ODB_PARENT:
         return find(query, anAlias, ParentType);
      case ODB_CHILD:
         return find(query, anAlias, ChildType);
      case ODB_REF:
         return find(query, aName,   UniqueType) ||
                find(query, anAlias, AliasType);
      case ODB_QUERY:
         return find(query, ' ', Pchar(""));
      default:
         assert(false);
      }
   } // Query

//---------------------------------------------------------------------
// Count the number of topics in the result set
//
long CODB::Length(void) {
   if (!result)
      return 0;

   err = 0;

   return result->Number();
   } // Length

//---------------------------------------------------------------------
// Obtain the indexed topic from the result set. The index of the first
// topic is zero.
//
PCBlock CODB::Topic(long index) {
   if (!result)
      return 0;

   err = 0;

   if (index < 0)
      return 0;

   if (!result->GotoItem(index+1))
      return 0;

   // Obtain the location of the topic
   PCItem item = result->Item();
   PCFindItem findItem = TYPECAST(item, FindItem);
   PCRef ref = new CRef(findItem->GetRef());

   // read the topic from the book
   PCAtom atom = book->ReadPart(ref);
   if (!atom) {
      err = book->Error();
      return 0;
      }

   PCBlock blk = TYPECAST(atom, Block);
   return blk;
   } // Topic

