
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// record.cpp
//*********************************************************************

#include "prefixb.h"
#include "block.h"
#include "file.h"
//#include "heapwrap.h"
#include "list.h"
#include "record.h"

//---------------------------------------------------------------------
// Constructor
//
CRecord::CRecord(void):
   CItem      (),
   delmAdded  (false),
   checkQuotes(true),
   tmplt      (0) {
   chars    = new char[1];
   chars[0] = 0;
   } // CRecord

//---------------------------------------------------------------------
// Copy Constructor
//
CRecord::CRecord(PCRecord _rec):
   CItem      (),
   delmAdded  (false),
   checkQuotes(true),
   tmplt      (0),
   chars     (0) {
   if (_rec->tmplt) {
      tmplt  = new char[strlen(_rec->tmplt)+1];
      strcpy(tmplt, _rec->tmplt);
      }
   if (_rec->chars) {
      chars = new char[strlen(_rec->chars)+1];
      strcpy(chars, _rec->chars);
      }
   } // CRecord

//---------------------------------------------------------------------
// Constructor
//
CRecord::CRecord(Pchar _data):
   CItem      (),
   delmAdded  (false),
   checkQuotes(true),
   tmplt      (0) {

   Assert(_data);

   size_t len = strlen(_data);
   Assert(_data && len <= maxSizeRecord);
   chars = new char[len+1];
   strcpy(chars, _data);
   } // CRecord

//---------------------------------------------------------------------
// Constructor
//
CRecord::CRecord(Pchar _data, Pchar _delm):
   CItem      (),
   delmAdded  (true),
   checkQuotes(true),
   tmplt      (0) {

   Assert(!StrEmpty(_delm));

   if (_data) {
      size_t len = strlen(_data);
      Assert(len <= maxSizeRecord);
      chars = new char[len+2];
      ChrCpy(chars, _delm[0]);
      Strcat(chars, _data, len+2);
      }
   } // CRecord

//---------------------------------------------------------------------
// Constructor
//
CRecord::CRecord(short _nrFields, Pchar _delm):
   CItem      (),
   delmAdded  (true),
   checkQuotes(true),
   tmplt      (0) {

   Assert(0 < _nrFields && _nrFields <= maxSizeRecord);
   Assert(!StrEmpty(_delm));

   chars            = new char[_nrFields+1];
   memset(chars, _delm[0], _nrFields);
   chars[_nrFields] = 0;
   } // CRecord

//---------------------------------------------------------------------
// Make a record from a string containing fixed fields.
// _sizes must contain a record of field sizes
//    example:
//       _s contains:
//          field: 1           2       3                   4
//          size:  12345678901212345678123456789012345678901234567
//          _s:    C10Cscreen  wrong.2 she.15              C01A012
//       _sizes contains:
//          |12|8|20|7
//       converts to:
//          |C10Cscreen|wrong.2|she.15|C01A012
//
// Note: _sizes can contain size subrecords.
//
CRecord::CRecord(Pchar _data, PCRecord _sizes):
   CItem      (),
   delmAdded  (false),
   checkQuotes(true),
   tmplt      (0) {

   Assert(_data && strlen(_data) <= maxSizeRecord);
   Assert(_sizes->chars);

   // Create an empty record
   short nr     = _sizes->Number();
   chars     = new char[nr+1];
   memset(chars, _sizes->chars[0], nr);
   chars[nr] = 0;

   // Move the fields of the fixed-format record into the text record
   short ofst   = 0;
   char work[maxSizeField];
   for (short i = 1; i <= nr; ++i) {
      if (_sizes->IsNumber(i)) {
         short size = _sizes->GetIntField(i);
         memset  (work, 0, size_t(maxSizeField));
         strncpy (work, _data + ofst, size);
         Trim    (work);
         ofst += size;
         }
      else {           // assume that the field is a subrecord of sizes
         _sizes->GetField(i, work, maxSizeField);
         PCRecord field = new CRecord();
         PCRecord wrk   = new CRecord(work);
         ofst += field->NewRecord(_data, wrk);
         field->GetDelimStr(work, maxSizeField);
         }
      PutField(i, work);
      }
   } // CRecord

//---------------------------------------------------------------------
// Destructor
//
CRecord::~CRecord() {
   delete[] tmplt;
   tmplt = 0;
   delete[] chars;
   chars = 0;
   } // ~CRecord

//---------------------------------------------------------------------
//
RTTI_CPP(Record, Item)

//---------------------------------------------------------------------
//
bool CRecord::match(Pvoid _p) {
   char work[maxSizeField];
   matchRec* r = (matchRec*)_p;
   GetField(r->iField, work, size_t(maxSizeField));
   return (stricmp(work, r->value) == 0);
   } // match

//---------------------------------------------------------------------
// Note: empty specified template clears the record template
//
void CRecord::SetTemplate(Pchar _tmplt) {
   delete[] tmplt;
   tmplt = 0;

   if (_tmplt) {
      tmplt = new char[strlen(_tmplt)+1];
      strcpy(tmplt, _tmplt);
      }
   } // SetTemplate

//---------------------------------------------------------------------
//
Pchar CRecord::GetTemplate(bool _alc) {
   if (!_alc)
      return tmplt;

   if (StrEmpty(tmplt)) {
      Pchar p = new char[1];
      Assert(p);
      p[0]    = 0;
      return p;
      }

   Pchar p = new char[strlen(tmplt)+1];
   Assert(p);
   strcpy(p, tmplt);
   return p;
   } // GetTemplate

//---------------------------------------------------------------------
// farSlang version
//
Pchar CRecord::GetTemplate(void) {
   return GetTemplate(true);
   } // GetTemplate

//---------------------------------------------------------------------
//
bool CRecord::GetFieldName(short _iField, Pchar _name, short _size) {
   Assert(_name);

   _name[0] = 0;
   if (StrEmpty(tmplt))
      return false;

   PCRecord t = new CRecord(tmplt);
   return t->GetField(_iField, _name, _size);
   } // GetFieldName

//---------------------------------------------------------------------
// farSlang version - caller must delete[] returned memory
//
Pchar CRecord::GetFieldName(short _iField) {
   if (StrEmpty(tmplt)) {
      Pchar p = new char[1];
      p[0]    = 0;
      return p;
      }

   PCRecord t = new CRecord(tmplt);
   return t->GetField(_iField);
   } // GetFieldName

//---------------------------------------------------------------------
// Clear the fields of the record - but not the template. Note that the
// template, if it exists, controls the number of fields in the record.
//
void CRecord::Clear(void) {
   if (tmplt) {
      delete[] chars;
      chars = new char[strlen(tmplt)+1];
      Assert(chars);
      strcpy(chars, tmplt);
      }

   short nr = Number();
   for (short i = 1; i <= nr; i++)
      PutField(i, Pchar(""));
   } // Clear

//---------------------------------------------------------------------
// Make a record from the specified record
//
void CRecord::NewRecord(PCRecord _record) {
   delmAdded = _record->delmAdded;
   delete[] tmplt;
   tmplt     = 0;
   if (_record->tmplt) {
      tmplt  = new char[strlen(_record->tmplt)+1];
      Assert(tmplt);
      strcpy(tmplt, _record->tmplt);
      }

   delete[] chars;
   chars = 0;
   if (_record->chars) {
      chars = new char[strlen(_record->chars)+1];
      Assert(chars);
      strcpy(chars, _record->chars);
      }
   } // NewRecord

//---------------------------------------------------------------------
// Make a record from the specified data string
//
void CRecord::NewRecord(Pchar _data) {
   delmAdded = false;
   delete[] chars;
   if (StrEmpty(_data)) {
      chars    = new char[2];
      chars[0] = ' ';
      chars[1] = 0;
      }
   else {
      chars    = new char[strlen(_data)+1];
      strcpy(chars, _data);
      }
   } // NewRecord

//---------------------------------------------------------------------
// Make a record from the string
//
void CRecord::NewRecord(Pchar _data, Pchar _delm) {
   delmAdded = true;
   delete[] chars;
   if (StrEmpty(_data)) {
      chars    = new char[2];
      chars[0] = _delm[0];
      chars[1] = 0;
      }
   else {
      chars    = new char[strlen(_data)+2];
      chars[0] = _delm[0];
      chars[1] = 0;
      strcat(chars, _data);
      }
   } // NewRecord

//---------------------------------------------------------------------
// Make an empty record. Note that, since the delimiter is specified,
// getDelimStr returns without the leading delimiter.
//
void CRecord::NewRecord(short _nrFields, Pchar _delm) {
   Assert(0 < _nrFields && _nrFields <= maxNrFields);
   Assert(!StrEmpty(_delm));

   delmAdded = true;
   delete[] chars;
   chars    = new char[_nrFields+1];
   Assert(chars);
   memset(chars, _delm[0], _nrFields);
   chars[_nrFields] = 0;
   } // NewRecord

//---------------------------------------------------------------------
// Make a record from a string containing fixed fields.
// _sizes must contain a record of field sizes
//    example:
//       _s contains:
//          field: 1           2       3                   4
//          size:  12345678901212345678123456789012345678901234567
//          _s:    C10Cscreen  wrong.2 she.15              C01A012
//       _sizes contains:
//          |12|8|20|7
//       converts to:
//          |C10Cscreen|wrong.2|she.15|C01A012
//
// Note: _sizes can contain size subrecords.
//
short CRecord::NewRecord(Pchar _string, PCRecord _sizes) {
   Assert(_string && strlen(_string) <= maxSizeRecord);
   Assert(_sizes->chars);

   NewRecord(_sizes->Number(), _sizes->chars);
   delmAdded = false;

   short ofst  = 0;
   char work[maxSizeRecord];
   for (short i = 1; i <= _sizes->Number(); ++i) {
      if (_sizes->IsNumber(i)) {
         short size = _sizes->GetIntField(i);
         memset(work, 0, maxSizeField);
         strncpy(work, _string+ofst, size);
         Trim(work);
         ofst    += size;
         }
      else {            // assume that the field is a subrecord of sizes
         _sizes->GetField(i, work, maxSizeField);
         PCRecord field = new CRecord();
         PCRecord wrk   = new CRecord(work);
         ofst          += field->NewRecord(_string, wrk);
         field->GetDelimStr(work, maxSizeRecord);
         }
      PutField(i, work);
      }
   return ofst;
   } // NewRecord

//---------------------------------------------------------------------
// Return the value of the delimiter
//
char CRecord::GetDelimiter(void) {
   return chars[0];
   } // GetDelimiter

//---------------------------------------------------------------------
// Return the value of the record as a delimited string
//
void CRecord::GetDelimStr(Pchar _string, short _size) {
   memset(_string, 0, _size);
   if (StrEmpty(chars))
      return;
   else if (delmAdded)
      Strcpy(_string, chars+1, _size-1)
   else
      Strcpy(_string, chars,   _size);
   } // GetDelimStr

//---------------------------------------------------------------------
// Return the value of the record as a delimited string (w/ leading
// delimiter)
//
void CRecord::GetDelimStrD(Pchar _string, short _size) {
   memset(_string, 0, _size);
   if (StrEmpty(chars))
      return;
   else
      Strcpy(_string, chars,   _size);
   } // GetDelimStrD

//---------------------------------------------------------------------
// Return the value of the record as a delimited string
//
Pchar CRecord::GetDelimStr(bool _alc) {
   if (StrEmpty(chars))
      if (_alc) {
         Pchar p =  new char[1];
         Assert(p);
         p[0]    = 0;
         return p;
         }
      else
         return chars;

   else if (delmAdded)
      if (_alc) {
         Pchar p = new char[strlen(chars)];
         Assert(p);
         strcpy(p, chars+1);
         return p;
         }
      else
         return chars+1;

   else if (_alc) {
      Pchar p = new char[strlen(chars)+1];
      Assert(p);
      strcpy(p, chars);
      return p;
      }

   return chars;
   } // GetDelimStr

//---------------------------------------------------------------------
// Return the value of the record as a delimited string (w/ leading
// delimiter)
//
Pchar CRecord::GetDelimStrD(bool _alc) {
   if (StrEmpty(chars))
      if (_alc) {
         Pchar p =  new char[1];
         Assert(p);
         p[0]    = 0;
         return p;
         }
      else
         return chars;

   else if (_alc) {
      Pchar p = new char[strlen(chars)+1];
      Assert(p);
      strcpy(p, chars);
      return p;
      }

   return chars;
   } // GetDelimStrD

//---------------------------------------------------------------------
// farSlang version
//
Pchar CRecord::GetDelimStr(void) {
   return GetDelimStr(true);
   } // GetDelimStr

//---------------------------------------------------------------------
// farSlang version
//
Pchar CRecord::GetDelimStrD(void) {
   return GetDelimStrD(true);
   } // GetDelimStrD

//---------------------------------------------------------------------
// Return true if all fields are empty
//
bool CRecord::Empty(void) {
   // test for an empty record
   if (StrEmpty(chars))
      return true;

   // test each field for empty
   for (short i = 1; i <= Number(); i++)
      if (!FieldEmpty(i))
         return false;

   return true;
   } // Empty

//---------------------------------------------------------------------
// Count the number of fields. If a field contains a quote, then ignore
// separators till the next quote.
//
short CRecord::Number(void) {
   // test for an empty record
   if (StrEmpty(chars))
      return 0;

   // Count field separators in record. Ignore separators that are
   // within quotes
   short iField = 0;
   bool inQuote = false;
   char quotChr = 0;
   for (Pchar p = chars; *p; p++)
      if      (inQuote && *p == quotChr) {
         quotChr = 0;
         inQuote = false;
         }
      else if (!inQuote)
         if      (checkQuotes && *p == '"' && *chars != '"') {
            quotChr = '"';
            inQuote = true;
            }
         else if (checkQuotes && *p == '\'' && *chars != '\'') {
            quotChr = '\'';
            inQuote = true;
            }
         else if (checkQuotes && *p == '`' && *chars != '`') {
            quotChr = '`';
            inQuote = true;
            }
         else if (*p == *chars)
            iField++;
   return iField;
   } // Number

//---------------------------------------------------------------------
// Return the offset of the begin and end characters of the specified
// field. If a field contains a quote, then ignore separators till the
// next quote.
//
void CRecord::FindField(short _iField, Rshort _iBegin, Rshort _iEnd) {
   Assert(_iField);

   short iField;
   if (_iField < 0)
      iField = Number() + _iField + 1;
   else
      iField = _iField;

   bool inQuote = false;
   char quotChr = 0;
   Pchar beg    = chars;
   for (; *beg; beg++)
      if      (iField == 0)
         break;
      else if (inQuote && *beg == quotChr) {
         quotChr = 0;
         inQuote = false;
         }
      else if (!inQuote)
         if     (checkQuotes && *beg == '"' && *chars != '"') {
            quotChr = '"';
            inQuote = true;
            }
         else if (checkQuotes && *beg == '\'' && *chars != '\'') {
            quotChr = '\'';
            inQuote = true;
            }
         else if (checkQuotes && *beg == '`' && *chars != '`') {
            quotChr = '`';
            inQuote = true;
            }
         else if (*beg == *chars)
            iField--;

   _iBegin = beg - chars;
   if (iField) {
      _iEnd = -1;
      return;
      }

   Pchar end;
   inQuote = false;
   quotChr = 0;
   for (end = beg; *end; end++)
      if     (inQuote && *end == quotChr) {
         quotChr = 0;
         inQuote = false;
         }
      else if (!inQuote)
         if     (checkQuotes && *end == '"' && *chars != '"') {
            quotChr = '"';
            inQuote = true;
            }
         else if (checkQuotes && *end == '\'' && *chars != '\'') {
            quotChr = '\'';
            inQuote = true;
            }
         else if (checkQuotes && *end == '`' && *chars != '`') {
            quotChr = '`';
            inQuote = true;
            }
         else if (*end == *chars)
            break;

   _iEnd = end - chars - 1;
   } // FindField

//---------------------------------------------------------------------
// Return the character offset of the specified field.
//
short CRecord::OfsField(short _iField) {
   Assert(abs(_iField) <= Number());

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);
   return iBegin;
   } // OfsField

//---------------------------------------------------------------------
// Return the length of the specified field.
//
short CRecord::LenField(short _iField) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   // return the size
   return short(iEnd - iBegin + 1);
   } // LenField

//---------------------------------------------------------------------
// Trim the specified field
//
short CRecord::TrimField(short _iField, whichEnd _end) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());
   Pchar work = GetField(_iField);
   Trim(work, _end);
   short length = strlen(work);
   PutField(_iField, work);
   delete[] work;

   return length;
   } // TrimField

//---------------------------------------------------------------------
// Trim all the fields in the record
//
void CRecord::TrimAll(whichEnd _end) {
   short nr = Number();
   for (short i = 1; i <= nr; i++)
      TrimField(i, _end);
   } // TrimAll

//---------------------------------------------------------------------
//
bool CRecord::FieldEmpty(short _iField) {
   Assert(abs(_iField) <= maxNrFields);

   // test for an empty record
   if (StrEmpty(chars))
      return true;

   // test for field request beyond record
   if (abs(_iField) > Number())
      return true;

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   // Test the field for empty
   if (iEnd >= iBegin)
      return false;

   return true;
   } // FieldEmpty

//---------------------------------------------------------------------
// Get the value of the specified field
//
bool CRecord::GetField(short _iField, Pchar _field, short _size) {
   Assert(abs(_iField) <= maxNrFields);

   _field[0] = 0;

   // test for an empty record
   if (StrEmpty(chars))
      return false;

   // test for field request beyond record
   if (abs(_iField) > Number())
      return false;

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   Assert(iBegin > 0);
   
   // copy the field to the value buffer
   if (iEnd >= iBegin) {
      strncpy(_field, chars+iBegin, Min(_size-1, iEnd-iBegin+1));
      _field[Min(_size-1, iEnd-iBegin+1)] = '\0';
      }
   return true;
   } // GetField

//---------------------------------------------------------------------
// farSlang version - caller must delete[] returned memory
//
Pchar CRecord::GetField(short _iField) {
   Assert(abs(_iField) <= maxNrFields);

   // test for an empty record
   if (StrEmpty(chars)) {
      Pchar p = new char[1];
      Assert(p);
      p[0]    = 0;
      return p;
      }

   // test for field request beyond record
   if (abs(_iField) > Number()) {
      Pchar p = new char[1];
      Assert(p);
      p[0]    = 0;
      return p;
      }

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   Assert(iBegin > 0);
   
   // copy the field to the value buffer
   if (iEnd >= iBegin) {
      Pchar p = new char[iEnd-iBegin+2];
      Assert(p);
      strncpy(p, chars + iBegin, iEnd - iBegin + 1);
      p[iEnd-iBegin+1] = 0;
      return p;
      }

   Pchar p = new char[1];
   Assert(p);
   p[0]    = 0;
   return p;
   } // GetField

//---------------------------------------------------------------------
// Get the value of the specified field.
//
bool CRecord::GetField(Pchar _name, Pchar _field, short _size) {
   Assert(!StrEmpty(_name));
   _field[0] = 0;

   if (tmplt == 0)
      return false;

   short i = GetFieldNr(_name);
   if (i == 0)
      return false;
   else
      return GetField(i, _field, _size);
   } // GetField

//---------------------------------------------------------------------
// farSlang version
//
Pchar CRecord::GetField(Pchar _name) {
   Assert(!StrEmpty(_name));

   if (tmplt == 0) {
      Pchar p = new char[1];
      Assert(p);
      p[0]    = 0;
      return p;
      }

   short i = GetFieldNr(_name);
   if (i == 0) {
      Pchar p = new char[1];
      Assert(p);
      p[0]    = 0;
      return p;
      }

   return GetField(i);
   } // GetField

//---------------------------------------------------------------------
// Return true if field is number
//
bool CRecord::IsNumber(short _iField) {
   Assert(_iField > 0 && _iField <= Number());

   bool isNumber = false;

   Pchar work = GetField(_iField);
   if (!StrEmpty(work)) {
      Trim(work);
      isNumber = ::IsNumber(work);
      }
   delete[] work;

   return isNumber;
   } // IsNumber

//---------------------------------------------------------------------
// Return true if field is number
//
bool CRecord::IsNumber(Pchar _name) {
   Assert(!StrEmpty(_name));

   short i = GetFieldNr(_name);
   if (i == 0)
      return false;
   else
      return IsNumber(i);
   } // IsNumber

//---------------------------------------------------------------------
// Return the boolean value of the specified field. If value is "true",
// then return true, return false otherwise. Not case sensitive.
//
bool CRecord::GetBoolField(short _iField) {
   bool value = false;

   Pchar work = GetField(_iField);

   if (!StrEmpty(work)) {
      Trim(work);
      value = stricmp(work, Pchar("true")) == 0;
      }

   delete[] work;

   return value;
   } // GetBoolField

//---------------------------------------------------------------------
// Return the boolean value of the specified field
//
bool CRecord::GetBoolField(Pchar _name) {
   Assert(!StrEmpty(_name));

   short i = GetFieldNr(_name);

   if (i == 0)
      return 0;
   else
      return GetBoolField(i);
   } // GetBoolField

//---------------------------------------------------------------------
// Return the integer numeric value of the specified field
//
short CRecord::GetIntField(short _iField) {
   short value = 0;

   Pchar work = GetField(_iField);

   if (!StrEmpty(work)) {
      Trim(work);
      if (::IsNumber(work))
         value = atoi(work);
      }
   delete[] work;

   return value;
   } // GetIntField

//---------------------------------------------------------------------
// Return the integer numeric value of the specified field
//
short CRecord::GetIntField(Pchar _name) {
   Assert(!StrEmpty(_name));

   short i = GetFieldNr(_name);
   if (i == 0)
      return 0;
   else
      return GetIntField(i);
   } // GetIntField

//---------------------------------------------------------------------
// Return the long numeric value of the specified field
//
long CRecord::GetLongField(short _iField) {
   long value = 0L;

   Pchar work = GetField(_iField);

   if (!StrEmpty(work)) {
      Trim(work);
      if (::IsNumber(work))
         value = StrToLong(work);
      }
   delete[] work;

   return value;
   } // GetLongField

//---------------------------------------------------------------------
// Return the long numeric value of the specified field
//
long CRecord::GetLongField(Pchar _name) {
   Assert(!StrEmpty(_name));

   short i = GetFieldNr(_name);
   if (i == 0)
      return 0;
   else
      return GetLongField(i);
   } // GetLongField

//---------------------------------------------------------------------
// Return the double numeric value of the specified field
//
double CRecord::GetDoubleField(short _iField) {
   double value = 0.0;

   Pchar work = GetField(_iField);

   if (!StrEmpty(work)) {
      Trim(work);
      value = StrToDouble(work);
      }
   delete[] work;

   return value;
   } // GetDoubleField

//---------------------------------------------------------------------
// Return the double numeric value of the specified field
//
double CRecord::GetDoubleField(Pchar _name) {
   Assert(!StrEmpty(_name));

   short i = GetFieldNr(_name);
   if (i == 0)
      return 0;
   else
      return GetDoubleField(i);
   } // GetDoubleField

//---------------------------------------------------------------------
// Replace the specified field with the specified value.
//
bool CRecord::PutField(short _iField, Pchar _field) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());

   // test for an empty record
   if (StrEmpty(chars))
      return false;

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   Assert(iBegin > 0);
   
   if (iEnd >= 0) {
      Pchar p = new char[strlen(chars) + strlen(_field)+1];
      Assert(p);
      strcpy(p, chars);
      strcpy(p+iBegin, _field);
      strcat(p, chars+iEnd+1);
      delete[] chars;
      chars = p;
      }
   return true;
   } // PutField

//---------------------------------------------------------------------
// Replace the specified field with the specified value.
//
bool CRecord::PutField(Pchar _name, Pchar _field) {
   Assert(!StrEmpty(_name));
   Assert(_field && strlen(_field) < maxSizeField);

   short i = GetFieldNr(_name);
   if (i == 0)
      return false;

   return PutField(i, _field);
   } // PutField

//---------------------------------------------------------------------
// Replace the specified field with the specified boolean value.
//
bool CRecord::PutBoolField(short _iField, bool _field) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());

   char work[maxSizeField];

   if (_field)
      strcpy(work, Pchar("true"));
   else
      strcpy(work, Pchar("false"));

   Assert(strlen(work) < maxSizeField);

   dirty = true;

   return PutField(_iField, work);
   } // PutBoolField

//---------------------------------------------------------------------
// Replace the specified field with the specified boolean value.
//
bool CRecord::PutBoolField(Pchar _name, bool _field) {
   Assert(!StrEmpty(_name));
   Assert(tmplt);

   short i = GetFieldNr(_name);

   if (i == 0)
      return false;

   return PutBoolField(i, _field);
   } // PutBoolField

//---------------------------------------------------------------------
// Replace the specified field with the specified numeric value.
//
bool CRecord::PutField(short _iField, long  _field) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());
   char work[maxSizeField];
   sprintf(work, Pchar("%ld"), _field);
   Assert(strlen(work) < maxSizeField);
   return PutField(_iField, work);
   } // PutField

//---------------------------------------------------------------------
// Replace the specified field with the specified value.
//
bool CRecord::PutField(Pchar _name, long _field) {
   Assert(!StrEmpty(_name));
   Assert(tmplt);

   short i = GetFieldNr(_name);
   if (i == 0)
      return false;

   return PutField(i, _field);
   } // PutField

/*?
//---------------------------------------------------------------------
// Replace the specified field with the specified value.
//
bool CRecord::PutLLField(short _iField, LONGLONG _field) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());
   char work[maxSizeField];
   _i64toa(_field, work, 10);
   Assert(strlen(work) < maxSizeField);     // the barndoor may be open
   return PutField(_iField, work);
   } // PutLLField

//---------------------------------------------------------------------
// Replace the specified field with the specified value.
//
bool CRecord::PutLLField(Pchar _name, LONGLONG _field) {
   Assert(!StrEmpty(_name));
   Assert(tmplt);

   short i = GetFieldNr(_name);
   if (i == 0)
      return false;

   return PutField(i, _field);
   } // PutLLField
?*/

//---------------------------------------------------------------------
// Insert a field before the specified field.
//
bool CRecord::InsertField(short _iField, Pchar _field) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number()+1);

   // test for an empty record
   if (StrEmpty(chars))
      return false;

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   if (iEnd >= iBegin) {
      Assert(iBegin > 0);
      
      // Inserting a field
      char work[maxSizeRecord];
      char dlm;
      if (*chars)
         dlm = *chars;
      else
         dlm = 0xfe;
      strncpy(work, chars, iBegin - 1);
      work[iBegin - 1] = '\0';
      ChrCat(work, dlm, maxSizeRecord);
      Strcat(work, _field, maxSizeRecord);
      Strcat(work, chars + iBegin - 1, maxSizeRecord);
      delete[] chars;
      chars = new char[strlen(work)+1];
      strcpy (chars, work);
      }
   else if (iBegin == short(strlen(chars))) {
      // Appending a field
      char work[maxSizeRecord];
      memset(work, 0, maxSizeRecord);
      char dlm;
      if (*chars)
         dlm = *chars;
      else
         dlm = 0xfe;
      Strcpy(work, chars, maxSizeRecord);
      ChrCat(work, dlm, maxSizeRecord);
      Strcat(work, _field, maxSizeRecord);
      delete[] chars;
      chars = new char[strlen(work)+1];
      strcpy (chars, work);
      }
   return true;
   } // InsertField

//---------------------------------------------------------------------
// Insert a field before the specified field.
//
bool CRecord::InsertField(Pchar _name1, Pchar _name2, Pchar _field) {
   Assert(_name1);
   Assert(!StrEmpty(_name2));
   Assert(GetFieldNr(_name2) == 0);
   Assert(_field && strlen(_field) < maxSizeField);

   // Create a template if needed
   if (tmplt == 0) {
      Assert(StrEmpty(_name1));
      tmplt    = new char[2];
      Assert(tmplt);
      tmplt[0] = '\255';
      tmplt[1] = '\0';
      }

   // Find the field index
   short i;
   if (StrEmpty(_name1))
      i = short(Number()+1);
   else {
      i = GetFieldNr(_name1);
      if (i == 0)
         return false;
      }

   Pchar hold = chars;
   chars     = tmplt;
   bool rtrn  = InsertField(i, _name2);
   tmplt      = chars;
   chars     = hold;
   if (rtrn == false)
      return rtrn;
   else
      return InsertField(i, _field);
   } // InsertField

//---------------------------------------------------------------------
// Insert a numeric field before the specified field.
//
bool CRecord::InsertField(Pchar _name1, Pchar _name2, long _field) {
   char work[maxSizeField];
   sprintf(work, Pchar("%ld"), _field);
   Assert(strlen(work) < maxSizeField);
   return InsertField(_name1, _name2, work);
   } // InsertField

//---------------------------------------------------------------------
// Append a field to the end of the record.
//
bool CRecord::AppendField(Pchar _field) {
   Assert(_field && strlen(_field) < maxSizeField);

   char dlm;
   if (*chars)
      dlm = *chars;
   else
      dlm = 0xfe;

   short size   = strlen(chars) + 1 + strlen(_field) + 1;
   Pchar newRec = new char[size];
   Strcpy(newRec, chars, size);
   ChrCat(newRec, dlm, size);
   Strcat(newRec, _field, size);
   delete[] chars;
   chars = newRec;

   return true;
   } // AppendField

//---------------------------------------------------------------------
// Append a field to the end of the record.
//
bool CRecord::AppendField(Pchar _name, Pchar _field) {
   Assert(_field && strlen(_field) < maxSizeField);

   if (!StrEmpty(_name) && GetFieldNr(_name))
      return false;

   // Initialize if needed
   if (tmplt == 0) {
      tmplt     = new char[strlen(_name)+2];
      Assert(tmplt);
      tmplt[0]  = 0xfe;
      tmplt[1]  = 0;
      strcat(tmplt, _name);
      delete[] chars;
      chars    = new char [strlen(_field)+2];
      chars[0] = 0xfe;
      chars[1] = 0;
      strcat(chars, _field);
      return true;
      }

   // Find the field index
   short i = short(Number()+1);

   Pchar hold = chars;
   chars     = tmplt;
   InsertField(i, _name);
   tmplt      = chars;
   chars     = hold;
   return InsertField(i, _field);
   } // AppendField

//---------------------------------------------------------------------
// Remove the specified field.
//
bool CRecord::DeleteField(short _iField) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());

   // test for an empty record
   if (StrEmpty(chars))
      return false;

   // obtain the boundaries of the field
   short iBegin, iEnd;
   FindField(_iField, iBegin, iEnd);

   Assert(iBegin > 0);

   if (iEnd >= iBegin)
      strcpy(chars+iBegin-1, chars+iEnd+1);
   else // field is empty, so eat a delimiter
      strcpy(chars+iBegin-1, chars+iBegin);
   return true;
   } // DeleteField

//---------------------------------------------------------------------
// Remove the specified field.
//
bool CRecord::DeleteField(Pchar _name) {
   Assert(!StrEmpty(_name));

   // test for an empty template
   if (StrEmpty(tmplt))
      return false;

   // test for an empty record
   if (StrEmpty(chars))
      return false;

   // obtain the number of the field
   short i = GetFieldNr(_name);
   if (i == 0)
      return false;

   // Delete the field from the
   // template and the record
   Pchar temp = chars;
   chars = tmplt;
   if (DeleteField(i)) {
      tmplt  = chars;
      chars = temp;
      return DeleteField(i);
      }
   else {
      tmplt  = chars;
      chars = temp;
      return false;
      }
   } // DeleteField

//---------------------------------------------------------------------
// Return the identifier of the template field that contains
// the specified value. (Ignores case and spaces.)
// Return zero if no field contains the specified
// value.
//
short CRecord::GetFieldNr(Pchar _name) {
   Assert(!StrEmpty(_name));

   // test for an empty template
   if (StrEmpty(tmplt))
      return 0;

   char temp[sizeString];
   Strcpy(temp, tmplt,    sizeString);
   ChrCat(temp, tmplt[0], sizeString);
   CapStr(temp);

   char name[sizeString];
   name[0] = tmplt[0];
   name[1] = '\0';
   Strcat(name, _name,    sizeString);
   ChrCat(name, tmplt[0], sizeString);
   CapStr(name);

   Pchar p = strstr(temp, name);
   if (p) {
      short i = 1;
      Pchar q = temp;
      while (q != p)
         if (*q++ == *temp)
            i++;
      return i;
      }
   return 0;
   } // GetFieldNr

//---------------------------------------------------------------------
// Returns the index of the field whose value matches
// the specified value.
//
short CRecord::FindField(Pchar _value, whichEnd _end) {
   for (short i = Number(); i; i--)
      if (Match(i, _value, _end))
         return i;
   return 0;
   } // FindField

//---------------------------------------------------------------------
// Returns true if the value of the specified _field matches the
// specified _value. Not case-sensitive. Note that if the specified
// _end is "frontEnd," _value must match the first part of the _field,
// and if the _end is "backEnd," _value must match the backend of
// _field.
//
bool CRecord::Match(short _iField, Pchar _value, whichEnd _end) {
   Assert(abs(_iField) <= maxNrFields);
   Assert(abs(_iField) <= Number());

   char arg[maxSizeField];
   Strcpy(arg, _value, maxSizeField);
   Trim(arg);
   CapStr(arg);
   short szArg = strlen(arg);
   char value[maxSizeField];
   if (GetField(_iField, value, maxSizeField)) {
      Trim(value);
      CapStr(value);
      short szVal = strlen(value);
      short i = StrPos(arg, value);
      if      (i < 0)
         return false;
      else if (szArg == szVal)
         return true;
      else if (_end == frontEnd && i == 0)
         return true;
      else if (_end == backEnd  && i + szArg == szVal)
         return true;
      }
   return false;
   } // Match

//---------------------------------------------------------------------
// Copy the template and the data in the block into the record.
//
bool CRecord::BlockTo(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   Pchar p       = block->Lock();
   Assert(p);

   // Copy the template
   if (*p) {
      ulong size = strlen(p);
      Assert(p+size < block->Pointer()+block->GetSize());

      delete[] tmplt;
      tmplt = new char[size+1];
      Assert(tmplt);
      strcpy(tmplt, p);
      p += size;
      }
   p++;

   // Copy the data
   if (*p) {
      ulong size = strlen(p);
      Assert(p+size < block->Pointer()+block->GetSize());

      delete[] chars;
      chars = new char[size+1];
      Assert(chars);
      strcpy(chars, p);
      }

   return true;
   } // BlockTo

//---------------------------------------------------------------------
// Copy the template and data in the record into the block.
//
bool CRecord::ToBlock(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   // Compute the size to allocate in the block
   ulong size = strlen(chars)+2;
   if (tmplt)
      size += strlen(tmplt);

   // Allocate and clear the space
   block->Alloc(size);
   block->Clear(0);

   Pchar p = block->Lock();

   // Copy the template
   if (tmplt)
      strcpy(p, tmplt);
   else
      p[0] = 0;
   p += strlen(p)+1;

   // Copy the data
   strcpy(p, chars);

   return true;
   } // ToBlock

//---------------------------------------------------------------------
//
ulong CRecord::BookSize(void) {
   return strlen(tmplt) + 1 + strlen(chars) + 1;
   } // BookSize

//---------------------------------------------------------------------
// Read a record from the specified file
//
bool CRecord::ReadFile(PCAtom _input, bool _isBook) {
   _isBook;

   delete[] tmplt;
   tmplt  = 0;
   delete[] chars;
   chars = 0;

   char line[sizeString];

   // Read the template from the file
   PCFile input = TYPECAST(_input, File);
   if (!(input->ReadString(line, sizeString)))
      return false;

   tmplt = new char[strlen(line)+1];
   Assert(tmplt);
   strcpy(tmplt, line);

   // Read the data from the file
   if (!(input->ReadString(line, sizeString))) {
      delete[] tmplt;
      tmplt = 0;
      return false;
      }

   chars = new char[strlen(line)+1];
   Assert(chars);
   strcpy(chars, line);

   return true;
   } // ReadFile

//---------------------------------------------------------------------
// Write the record to the specified file
//
bool CRecord::WriteFile(PCAtom _output, bool _isBook) {
   _isBook;

   if (chars == 0)
      return FVE_NODATA;

   PCFile output = TYPECAST(_output, File);

   // Write the template to the file
   if (tmplt == 0) {
      if (!(output->WriteString(Pchar(""))))
         return false;
      }
   else
      if (!(output->WriteString(tmplt)))
         return false;

   // Write the data to the file
   return TYPECAST(_output, File)->WriteString(chars);
   } // WriteFile

//---------------------------------------------------------------------
// Read a text line from the specified input file.
//
bool CRecord::ReadThis(PCAtom _input) {
   char text[sizeString];

   // Read a string from the file
   if (!(TYPECAST(_input, File)->ReadString(text, sizeString)))
      return false;

   NewRecord(text);
   return true;
   } // ReadThis

//---------------------------------------------------------------------
// Write the delimited string contents of this
// string item to the specified file.
//
bool CRecord::WriteThis(PCAtom _output) {
   return TYPECAST(_output, File)->WriteString(GetDelimStr(false));
   } // WriteThis

//---------------------------------------------------------------------
// Read a record from the specified file
//
bool CRecord::Load(Pchar _fileName) {
   PCFile input = new CFile(_fileName);

   if (!input->OpenInput())
      return false;

   Pchar s   = input->ReadString();
   delete[] chars;
   chars    = s;

   delmAdded = false;

   input->Close();
   return true;
   } // Load

//---------------------------------------------------------------------
// Write the contents of the record to the specified file.
//
bool CRecord::Save(Pchar _fileName) {
   PCFile output = new CFile(_fileName);

   if (!output->OpenOutput())
      return false;

   output->WriteString(chars);
   output->Close();

   return true;
   } // Save


