
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// refs.cpp
//*********************************************************************

#include "prefixb.h"
#include "refs.h"

#ifndef FV_NOLOG
#include "syslog.h"
#endif

const ulong rEmpty     = ulong(0);                      // Must be zero
const ulong deleted    = ulong(-1);
const short nrRefSlots = 8;

//---------------------------------------------------------------------
//
short aRefBucket::Size(short _nrSlots) {
   return short(sizeof(ulong) + size_t(_nrSlots) * sizeof(aRef));
   } // Size

//---------------------------------------------------------------------
//
paRef aRefBucket::Slot(short _iSlot, short _nrSlots) {
   Assert(_iSlot < _nrSlots);

   return paRef(Pchar(&next) + sizeof(ulong) + _iSlot * sizeof(aRef));
   } // Slot

//---------------------------------------------------------------------
// Constructor
//
CRefs::CRefs(ulong _hdrSize, Pchar _name):
   CHash    (_hdrSize + sizeof(long), _name),
   nrSlots  (nrRefSlots),
   location (0),
   iSlot    (0),
   refbkt   (0),
   refbucket(0) {

   // Allocate imMemory space for the ref refbucket
   refbkt          = new CBlock();
   refbkt->Alloc(aRefBucket::Size(nrSlots));
   refbucket       = paRefBucket(refbkt->Lock());
   refbkt->Clear(0);
   refbucket->next = chainEnd;
   } // CRefs

//---------------------------------------------------------------------
// Destructor
//
CRefs::~CRefs() {
   } // ~CRefs

//---------------------------------------------------------------------
//
RTTI_CPP(Refs, Hash)

//---------------------------------------------------------------------
// Must be called before Create() but after Open() is ok.
//
void CRefs::SetSizes(long _nrRefSlots,
                     long _nrHshSlots,
                     long _sizeKey,
                     long _nrKeys) {
   CHash::SetSizes(_nrHshSlots, _sizeKey, _nrKeys);

   if (_nrRefSlots) {
      nrSlots      = _nrRefSlots;
      // Allocate space for the ref refbucket
      refbkt->Alloc(aRefBucket::Size(nrSlots));
      refbucket       = paRefBucket(refbkt->Lock());
      refbkt->Clear(0);
      refbucket->next = chainEnd;
      }
   } // SetSizes

//---------------------------------------------------------------------
// Create the reference file
//
bool CRefs::Create(void) {
   Assert(nrSlots > 0);

   CHash::Create();
   if (!OpenUpdate())
      return false;

   // Record the operating parameters in the file
   Seek(hdrSize - sizeof(short), fromHead);
   if (!WriteBlock(&nrSlots, sizeof(short)))
      return false;

   Close();
   return true;
   } // Create

//---------------------------------------------------------------------
//
bool CRefs::Open(void) {
   if (!CHash::Open())
      return false;

   // Get the number of slots in the file
   Seek(hdrSize - sizeof(short), fromHead);
   if (!ReadBlock(&nrSlots, sizeof(short)))
      return false;
   Assert(nrSlots > 0);

   // Initialize the inMemory refbucket
   refbkt->Free();
   refbkt->Alloc(aRefBucket::Size(nrSlots));
   refbucket       = paRefBucket(refbkt->Lock());
   refbkt->Clear(0);
   refbucket->next = chainEnd;

   return true;
   } // Open

//---------------------------------------------------------------------
// Find the first reference in the reference chain.
// Return false if no references.
//
bool CRefs::GetFirstRef(ulong _location, PCRef _ref) {
   Assert(nrSlots > 0);

   if (_location == 0)
      return false;

   location = _location;
   iSlot    = -1;

   // read the first refbucket
   Seek(location, fromHead);
   if (!ReadBlock(refbucket, aRefBucket::Size(nrSlots)))
      return false;

   // Get a reference
   return GetNextRef(_ref);
   } // GetFirstRef

//---------------------------------------------------------------------
// Find the next reference in the reference chain. Return false if no
// more references.
//
bool CRefs::GetNextRef(PCRef _ref) {
   Assert(nrSlots > 0);

   do {
      iSlot++;
      if (iSlot >= nrSlots)
         if (refbucket->next) {
            location = refbucket->next;
            iSlot    = 0;

            // read the next refbucket
            Seek(location, fromHead);
            if (!ReadBlock(refbucket, aRefBucket::Size(nrSlots)))
                return false;
            }
         else // followed chain to its end
            return false;
      } while (refbucket->Slot(iSlot, nrSlots)->loc == deleted ||
             refbucket->Slot(iSlot, nrSlots)->loc == rEmpty);

   // return contents of the current slot
   aRef ref  = *refbucket->Slot(iSlot, nrSlots);
   _ref->loc = ref.loc;
   return true;
   } // GetNextRef

//---------------------------------------------------------------------
// Make a new refbucket and return its location
//
bool CRefs::makeBucket(RaRef _ref) {
   Assert(nrSlots > 0);

   // Make a new refbucket with the reference
   PCBlock b = new CBlock();
   b->Alloc(aRefBucket::Size(nrSlots));
   paRefBucket buck = paRefBucket(b->Lock());
   b->Clear(0);

   // Add the reference to the new refbucket
   *buck->Slot(0, nrSlots) = _ref;

   // Write the refbucket to the tail of the file
   Seek(0, fromTail);
   refbucket->next = Where();
   if (!WriteBlock(buck, aRefBucket::Size(nrSlots))) {
      location = 0;
      return false;
      }

   // link the current refbucket to the new refbucket
   if (location) {
      Seek(location, fromHead);
      if (!WriteBlock(refbucket, aRefBucket::Size(nrSlots))) {
         location = 0;
         return false;
         }
      }

   // update the current refbucket to the new refbucket
   location = refbucket->next;
   memcpy(refbucket, buck, aRefBucket::Size(nrSlots));

   return true;
   } // makeBucket

//---------------------------------------------------------------------
// Add the specified reference to the current refbucket chain
//
bool CRefs::addToBucket(RaRef _ref) {
   while (refbucket->Slot(iSlot, nrSlots)->loc != rEmpty  &&// occupied
          refbucket->Slot(iSlot, nrSlots)->loc != deleted && // deleted
         *refbucket->Slot(iSlot, nrSlots)      != _ref) {
      ++iSlot;
      if (iSlot >= nrSlots) {
         if (refbucket->next) {
            location = refbucket->next;
            iSlot    = 0;

#ifndef FV_NOLOG
            SysLog::Log('D', Pchar("refs::addToBucket 1"));
#endif

            // Read the next refbucket
            Seek(location, fromHead);
            if (!ReadBlock(refbucket, aRefBucket::Size(nrSlots)))
               return false;
            }
         else {
#ifndef FV_NOLOG
            SysLog::Log('D', Pchar("refs::addToBucket 2"));
#endif

            // Make a new refbucket to extend the chain and quit
            return makeBucket(_ref);
            }
         }
      }

#ifndef FV_NOLOG
   SysLog::Log('D', Pchar("refs::addToBucket 3"));
#endif

   if (*refbucket->Slot(iSlot, nrSlots) != _ref) {// Don't save dup refs
      // Update the available slot
      *refbucket->Slot(iSlot, nrSlots) = _ref;
      Seek(location, fromHead);
      if (!WriteBlock(refbucket, aRefBucket::Size(nrSlots)))
         return false;

#ifndef FV_NOLOG
      SysLog::Log('D', Pchar("refs::addToBucket 4"));
#endif
      }

#ifndef FV_NOLOG
   SysLog::Log('D', Pchar("refs::addToBucket 5"));
#endif

   return true;
   } // addToBucket

//---------------------------------------------------------------------
// Insert the specified reference into the specified reference chain.
// Return the location of the chain.
//
ulong CRefs::InsertRef(ulong _location, PCRef _ref) {
   location = _location;
   aRef ref(_ref->Loc());

   if (location) {
      iSlot = 0;

      // Read the refbucket
      Seek(location, fromHead);
      if (!ReadBlock(refbucket, aRefBucket::Size(nrSlots)))
         return 0;

      // Add the reference to the refbucket
      addToBucket(ref);
      }
   else                                                      // new key
      // Add a new refbucket
      if (!makeBucket(ref))
         return 0;

   return location;
   } // InsertRef

//---------------------------------------------------------------------
// Delete the specified reference from the specified reference chain
//
bool CRefs::DeleteRef(ulong _location, PCRef _ref) {
   if (_location == 0)
      return false;

   // Load the head refbucket
   PCRef tref = new CRef();
   if (!GetFirstRef(_location, tref))
      return false;

   // Find the slot matching the specified reference
   iSlot    = 0;
   aRef ref(_ref->Loc());
   while (*refbucket->Slot(iSlot, nrSlots) != ref) {
      ++iSlot;

      // Check for the end of the refbucket
      if (iSlot >= nrSlots) {
         location = refbucket->next;
         // Read the next refbucket if there is another
         if (location) {
            iSlot = 0;

            // Read the next refbucket
            Seek(location, fromHead);
            if (!ReadBlock(refbucket, aRefBucket::Size(nrSlots)))
               return false;
            }
         // Ran out of buckets to search
         else                             // Reference not in the chain
            return false;
         }
      }

   // Remove the reference from the slot
   if (refbucket->slot[iSlot] == ref) {
      // Update the slot by compressing the reference out
      for (short i = iSlot; i < nrSlots-1; ++i)
         *refbucket->Slot(i, nrSlots) = *refbucket->Slot(short(i+1), nrSlots);
      refbucket->Slot(nrSlots-1, nrSlots)->loc = rEmpty;

      // Write the updated refbucket back to the file
      Seek(location, fromHead);
      if (!WriteBlock(refbucket, aRefBucket::Size(nrSlots)))
         return false;

      // Return true if the reference is fully deleted
      if (location                         == _location &&
          refbucket->Slot(0, nrSlots)->loc == rEmpty)
         return true;
      }
   return false;
   } // DeleteRef

