
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// search.cpp
//*********************************************************************
//
// The CSearch class is patterned after the CEvaluate class. It parses
// an embedded infix search expression string into a postfix expression
// string, then evaluates the resulting expression.
//
// Evaluation comprises a set of searches against the specified farBook
// according to the specified expression each search resulting in a
// list of node names, and logical operations between result lists to
// reduce in a single list.
//
// The recognized operator set includes the following:
//
// Precedence
//    3: (, )
//    2: *, &, and
//    1: +, |, or,  -, butnot
//
// notes:
//    The parse() method pushes empty CSearchItem objects onto the
//    evaluation stack as it parses. When the searchAnd(), searchOr(),
//    and searchButNot() methods pop an CSearchItem object off the
//    stack, they check to see if the list is null. If so, using the
//    arg, they perform a find to obtain a list. Each method combines
//    the two lists in a different way:
//
//    searchAnd():    insert the objects in the first list into a hash
//                    table, then insert the objects in the second list
//                    into the hash table, counting duplicates. When
//                    done, dump all the objects with a count greater
//                    than one in the hash table into a new findlist.
//
//                    first list:  object->keep = false
//                    second list: if object in table,
//                                 then object->keep = true
//
//    searchOr():     insert the objects in the first list into a hash
//                    table, then insert the objects in the second list
//                    into the hash table. When done, dump all the
//                    objects in the hash table into a new findlist.
//
//                    first list:  object->keep = true;
//                    second list: object->keep = true;
//
//    searchButNot(): insert the objects in the first list into a hash
//                    table, then delete the objects in the hash table
//                    that match the objects in the second list. When
//                    done, dump all the objects remaining in the hash

//                    table into a new findlist.
//
//                    first list:  object->keep = true;
//                    second list: if object in table,
//                                 then object->keep = false
//
//    need a preference that limits the number of hits from a search?
//
// class CSearchItem: public CItem {
//    public:
//       CSearchItem(Pchar _arg);
//       virtual ~CSearchItem();
//    protected:
//       bool       keep;
//       Pchar      arg;
//       PCFindList list;
//    }; // CSearchItem
//
//*********************************************************************

#include "lclbook.h"
#include "hshlist.h"
#include "misc.h"
#include "search.h"
#ifndef FV_NOLOG
#include "syslog.h"
#endif

const char spaces[2] = {' ', 0};

//=====================================================================
// SearchItem implementation
//
//---------------------------------------------------------------------
// Constructor
//
CSearchItem::CSearchItem(Pchar _key):
   CItem(),
   key  (0),
   list (0) {

   Assert(!StrEmpty(_key));
   STRINIT(key, _key)
   } // CSearchItem

//---------------------------------------------------------------------
// Destructor
//
CSearchItem::~CSearchItem() {
   delete[] key;
   key = 0;
   } // ~CSearchItem

//---------------------------------------------------------------------
//
RTTI_CPP(SearchItem, Item)

//=====================================================================
// SearchHash implementation
//
const ushort iSearchHash = 1000;

_FBDEF(CSearchHash)
class CSearchHash: public CHshList {
   public:
      CSearchHash(short _wide = 32);
      virtual ~CSearchHash();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      void AndInsert   (PCItem _item, bool _first);
      void OrInsert    (PCItem _item, bool _first);
      void ButNotInsert(PCItem _item, bool _first);

      PCFindList GetFindList(void);
   }; // CSearchHash

//---------------------------------------------------------------------
// Constructor
//
CSearchHash::CSearchHash(short _wide):
   CHshList(_wide) {
   } // CSearchHash

//---------------------------------------------------------------------
// Destructor
//
CSearchHash::~CSearchHash() {
   } // ~CSearchHash

//---------------------------------------------------------------------
//
RTTI_CPP(SearchHash, HshList)

//---------------------------------------------------------------------
// If the first insert, set item->keep to false, and insert the item
// into the table. If not the first insert, if the specified item is
// already in the table, set item->keep to true. Do not insert the
// specified item.
//
void CSearchHash::AndInsert(PCItem _item, bool _first) {
   PCFindItem find = TYPECAST(_item, FindItem);
   ulong loc       = find->GetRef()->loc;

   if (_first) {
      find->SetPosition(0);
      Insert(Pvoid(loc), _item);
      }
   else {
      PCItem item = GetFirstItem(Pvoid(loc));
      if (item) {
         find = TYPECAST(item, FindItem);
         find->SetPosition(1);
         }
      }
   } // AndInsert

//---------------------------------------------------------,---------
// If the first insert, set item->keep to true, and insert the item
// into the table. If not the first insert, if the specified item is
// not already in the table, set item->keep to true, and insert the
// item.
//
void CSearchHash::OrInsert(PCItem _item, bool _first) {
   PCFindItem find = TYPECAST(_item, FindItem);
   ulong loc       = find->GetRef()->loc;
   find->SetPosition(1);

   if (_first)
      Insert(Pvoid(loc), _item);
   else {
      PCItem item = GetFirstItem(Pvoid(loc));
      if (!item)
         Insert(Pvoid(loc), _item);
      }
   } // OrInsert

//---------------------------------------------------------------------
// If the first insert, set item->keep to true, and insert the item
// into the table. If not the first insert, if the specified item is
// already in the table, set item->keep to false. Do not insert the
// specified item.
//
void CSearchHash::ButNotInsert(PCItem _item,  bool _first) {
   PCFindItem find = TYPECAST(_item, FindItem);
   ulong loc       = find->GetRef()->loc;

   if (_first) {
      find->SetPosition(1);
      Insert(Pvoid(loc), _item);
      }
   else {
      PCItem item = GetFirstItem(Pvoid(loc));
      if (item) {
         find = TYPECAST(item, FindItem);
         find->SetPosition(0);
         }
      }
   } // ButNotInsert

//---------------------------------------------------------------------
// Traverse the hash table to move its contents into a single
// findlist.
//
PCFindList CSearchHash::GetFindList(void) {
   PCFindList findList = new CFindList();

   short i = GetWidth();
   while (i > 0) {
      i--;
      PCList list = GetList(i);
      if (list && list->GotoHead()) do {
         PCItem item     = list->Item();
         list->Remove();

         PCFindItem find = TYPECAST(item, FindItem);
         if (find->GetPosition()) {
            find->SetPosition(0);
            findList->Insert(item);
            }
         } while (list->GotoHead());
      }

   return findList;
   } // GetFindList
   
//=====================================================================
// Search implementation
//
//---------------------------------------------------------------------
// Constructor
//
CSearch::CSearch(PCBook _book):
   CAtom     (),
   usingIndex(false),
   pbook     (_book),
   book      (0),
   domain    (0),
   atEnd     (false),
   op        (NOTKNOWN),
   inFixExp  (0),
   pInFixExp (0),
   postFixExp(0),
   stoken    (0),
   err       (FVE_OK),
   ch        (spacer) {
   } // CSearch

//---------------------------------------------------------------------
// Destructor
//
CSearch::~CSearch() {
   delete[] domain;
   domain = 0;
   delete[] inFixExp;
   inFixExp = 0;
   delete[] postFixExp;
   postFixExp = 0;
   delete[] stoken;
   stoken = 0;
   } // ~CSearch

//---------------------------------------------------------------------
//
RTTI_CPP(Search, Atom)

//=====================================================================
// Expression evaluator
//=====================================================================

//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
// Evaluate the expression returning the computed value
//
PCFindList CSearch::Search(Pchar _domain,
                           Pchar _expression,
                           bool  _fulText) {
   STRINIT(domain, _domain)

   if (StrEmpty(_expression))
      return 0;

   //  Check if the search is against a remote farBook
   if (pbook->Isa(iRemoteBook)) {
      PCList p = new CFindList();
      pbook->Search(_domain,  _expression, _fulText, p);
      return TYPECAST(p, FindList);
      }

   // Perform the search against the local farBook
   char work[sizeString];
   Strcpy(work, _expression, sizeString);

   // Compile from infix to postfix
   if (!parse(work, work, sizeString))
      return 0;

   // Use the book index if it is available
   char bookName[sizeString];
   Strcpy(bookName, pbook->Name(false), sizeString);
   CutExtension(bookName);
   CatName(0, bookName, Pchar("ndx"), sizeString);
   ulong err = 0;
   if (FileExists(bookName, err)) {
      usingIndex = true;
#ifndef FV_NOSTEM
      stem = new CStem();
#endif
      book = new CLocalBook(bookName);
      book->Open();
      }
   else
      book = pbook;

   // Hide embedded spaces
   bool inq = false;
   for (Pchar p = work; *p; p++)
      if (inq) {                                       // inside quotes
         if      (*p == '"')
            inq = false;
         else if (*p == ' ')
            *p = char(0x7f);
         }
      else if (*p == '"')                             // outside quotes
         inq = true;
   // Convert to a record on white space
   Squeeze(work, spacer);
   PCRecord exp = new CRecord(work, Pchar(spaces));

   PCList stack = new CList();

   bool ok = true;

   char token[sizeString];
   while (exp->Number()) {
      // Get the next token
      exp->GetField(1, token, sizeString);
      exp->DeleteField(1);

      // Re-expose embedded spaces
      for (Pchar p = token; *p; p++)
         if (*p == char(0x7f))
            *p = ' ';

      // Handle quoted strings with embedded spaces
      if      (token[0] == strQuote1) {
         if (getString(exp, token, sizeString)) {
            PCSearchItem qry = new CSearchItem(token);
            if (!query(qry)) {
               ok = false;
               break;
               }
            PCItem item     = TYPECAST(qry, Item);
            stack->Insert(item, atTop);
            }
         else {
            ok = false;
            break;
            }
         }
      else if (token[0] == strQuote2) {
         if (getString(exp, token, sizeString)) {
            PCSearchItem qry = new CSearchItem(token);
            if (!query(qry)) {
               ok = false;
               break;
               }
            PCItem item     = TYPECAST(qry, Item);
            stack->Insert(item, atTop);
            }
         else {
            ok = false;
            break;
            }
         }
      else if (token[0] == strQuote3) {
         if (getString(exp, token, sizeString)) {
            PCSearchItem qry = new CSearchItem(token);
            if (!query(qry)) {
               ok = false;
               break;
               }
            PCItem item     = TYPECAST(qry, Item);
            stack->Insert(item, atTop);
            }
         else {
            ok = false;
            break;
            }
         }

      // Handle binary operators
      else if (strcmp(token, Pchar("&"))  == 0) {
         if (!queryAnd(stack)) {
            ok = false;
            break;
            }
         }
      else if (strcmp(token, Pchar("|"))  == 0) {
         if (!queryOr(stack)) {
            ok = false;
            break;
            }
         }
      else if (strcmp(token, Pchar("-"))  == 0) {
         if (!queryButNot(stack)) {
            ok = false;
            break;
            }
         }
      // Handle an unrecognized argument
      else
         Assert(false);
      } // while (exp.Number())

   if (usingIndex)
      book->Close();

   if (!ok)
      return 0;

   // Return result value in place of the input expression
   if      (stack->Number() == 0)
      // Replace the _exp string with the (empty) result
      return 0;

   else if (stack->Number() == 1) {
      stack->GotoHead();
      PCItem item      = stack->Item();
      PCSearchItem qry = TYPECAST(item, SearchItem);

      // Check for indexed search
      if (book != pbook) {
         PCFindList list = qry->list;
         if (list->GotoHead()) do {
            PCItem item     = list->Item();
            PCFindItem find = TYPECAST(item, FindItem);
            find->SetBook(pbook);
            } while (list->GotoNext());
         }

      return qry->list;
      }
   else      // Oops! The expression did not resolve to a single result
      return 0;
   } // Search

//---------------------------------------------------------------------
// If the specified query item does not have a find list, get one and
// fill it from the specified farBook.
//
bool CSearch::query(PCSearchItem _query) {
   if (_query->list)                                 // already has one
      return true;

   _query->list = new CFindList();

   char queryKey[sizeString];
   Strcpy(queryKey, _query->key, sizeString);
#ifndef FV_NOSTEM
   if (usingIndex)
      stem->StemWord(queryKey, int(sizeString));
#endif

   // Build an alias-type target key
   char key[sizeString];
   MakeKey(key, domain, AliasType, queryKey, sizeString);

   // Find the targets
   _query->list->FindKey(book, anAlias, key);

   // Check for errors
   Error(_query->list->Error());
   if (err != FVE_OK) {
      char msg[sizeString];
      sprintf(msg,
              Pchar("Error %ld encountered while reading target nodes"),
              err);
#ifndef FV_NOLOG
      SysLog::Log('D', msg);
#endif
      return false;
      }

   // Build a keyword-type target key (for backward compatibility)
   MakeKey(key, domain, KeywordType, queryKey, sizeString);

   // Find the targets
   _query->list->FindKey(book, anAlias, key);

   // Check for errors
   Error(_query->list->Error());
   if (err != FVE_OK) {
      char msg[sizeString];
      sprintf(msg,
              Pchar("Error %ld encountered while reading target nodes"),
              err);
#ifndef FV_NOLOG
      SysLog::Log('D', msg);
#endif
      return false;
      }

   return true;
   } // query

//---------------------------------------------------------------------
// Evaluate query1 & query2
//
bool CSearch::queryAnd(PCList _stack) {
   if (_stack->Number() < 2)
      return false;

   // Pop query2 off the stack
   _stack->GotoHead();
   PCItem item2       = _stack->Item();
   _stack->Remove();
   PCSearchItem query2 = TYPECAST(item2, SearchItem);

   // Pop query1 off the stack
   _stack->GotoHead();
   PCItem item1       = _stack->Item();
   _stack->Remove();
   PCSearchItem query1 = TYPECAST(item1, SearchItem);

   PCSearchHash table  = new CSearchHash();

   // Derive the intersection of query1 and query2
   if (query1->list->GotoHead()) do {
      PCItem item = query1->list->Item();
      query1->list->Remove();
      table->AndInsert(item, true);
      } while (query1->list->GotoHead());

   if (query2->list->GotoHead()) do {
      PCItem item = query2->list->Item();
      query2->list->Remove();
      table->AndInsert(item, false);
      } while (query2->list->GotoHead());

   // Push (query1 & query2) onto the stack
   query1->list = table->GetFindList();
   _stack->Insert(item1, atTop);

   return true;
   } // queryAnd

//---------------------------------------------------------------------
// Evaluate query1 | query2
//
bool CSearch::queryOr(PCList _stack) {
   if (_stack->Number() < 2)
      return false;

   // Pop query2 off the stack
   _stack->GotoHead();
   PCItem item2       = _stack->Item();
   _stack->Remove();
   PCSearchItem query2 = TYPECAST(item2, SearchItem);

   // Pop query1 off the stack
   _stack->GotoHead();
   PCItem item1       = _stack->Item();
   _stack->Remove();
   PCSearchItem query1 = TYPECAST(item1, SearchItem);

   PCSearchHash table  = new CSearchHash();

   // Derive the union of query1 and query2
   if (query1->list->GotoHead()) do{
      PCItem item = query1->list->Item();
      query1->list->Remove();
      table->OrInsert(item, true);
      } while (query1->list->GotoHead());

   if (query2->list->GotoHead()) do {
      PCItem item = query2->list->Item();
      query2->list->Remove();
      table->OrInsert(item, false);
      } while (query2->list->GotoHead());

   // Push (query1 | query2) onto the stack
   query1->list = table->GetFindList();
   _stack->Insert(item1, atTop);

   return true;
   } // queryOr

//---------------------------------------------------------------------
// Evaluate query1 - query2
//
bool CSearch::queryButNot(PCList _stack) {
   if (_stack->Number() < 2)
      return false;

   // Pop query2 off the stack
   _stack->GotoHead();
   PCItem item2       = _stack->Item();
   _stack->Remove();
   PCSearchItem query2 = TYPECAST(item2, SearchItem);

   // Pop query1 off the stack
   _stack->GotoHead();
   PCItem item1       = _stack->Item();
   _stack->Remove();
   PCSearchItem query1 = TYPECAST(item1, SearchItem);

   PCSearchHash table  = new CSearchHash();

   // Derive the difference of query1 and query2
   if (query1->list->GotoHead()) do {
      PCItem item = query1->list->Item();
      query1->list->Remove();
      table->ButNotInsert(item, true);
      } while (query1->list->GotoHead());

   if (query2->list->GotoHead()) do {
      PCItem item = query2->list->Item();
      query2->list->Remove();
      table->ButNotInsert(item, false);
      } while (query2->list->GotoHead());

   // Push (query1 - query2) onto the stack
   query1->list = table->GetFindList();
   _stack->Insert(item1, atTop);

   return true;
   } // queryButNot

//---------------------------------------------------------------------
// Extract quoted string from expression.
//
bool CSearch::getString(PCRecord _exp, Pchar _token, ushort _size) {
   if (_token[0] != strQuote1 &&
       _token[0] != strQuote2 &&
       _token[0] != strQuote3)
      return true;

   while (_exp->Number() && (strlen(_token)           == 1 ||
                             _token[strlen(_token)-1] != _token[0])) {
      ChrCat(_token, spacer, _size);
      char field[sizeString];
      _exp->GetField(1, field, sizeString);
      _exp->DeleteField(1);
      if (strlen(_token) + strlen(field) < _size)
         Strcat(_token, field, _size)
      else {
         Assert(false);
         }
      }

   Trim(_token);
   if (_token[strlen(_token)-1] != _token[0])
      return false;
   else {
      _token[strlen(_token)-1] = 0;
      strcpy(_token, _token+1);
      Trim(_token);
      return true;
      }
   } // getString

//=====================================================================
// Parser
//=====================================================================

//---------------------------------------------------------------------
// Expression parser converts infix notation to postfix notation.
//
// Operator Precedence
//
//    3: (, )
//    2: *, &, 'and'
//    1: +, |, 'or', -, 'butnot'
//
bool CSearch::parse(Pchar _inFixExp, Pchar _postFixExp, ushort _size) {
   Assert(_postFixExp);
   Assert(_size != 0);

   // Check for something to do
   if (StrEmpty(_inFixExp)) {
      memset(_postFixExp, 0, _size);
      return true;
      }

   // Prepare the input and output expression areas
   inFixExp = new char[strlen(_inFixExp)+1];
   strcpy(inFixExp, _inFixExp);
   memset(_postFixExp, 0, _size);

   delete[] postFixExp;
   postFixExp = new char[sizeString];
   memset(postFixExp, 0, sizeString);

   // Initialize the tokenizer
   openTokenizer(inFixExp);

   // Parse the expression
   do {
      getToken();
      parse1();
      } while (!atEnd);

   // Return the postfix expression result
   if (strlen(postFixExp) >= _size) {
      delete[] inFixExp;
      inFixExp   = 0;
      delete[] postFixExp;
      postFixExp = 0;
      return false;
      }
   else {
      strncpy(_postFixExp, postFixExp, _size);
      delete[] inFixExp;
      inFixExp   = 0;
      delete[] postFixExp;
      postFixExp = 0;
      return true;
      }
   } // parse

//---------------------------------------------------------------------
// Parse exp1 ::= exp2 {OP exp2}
// (where OP in {+, |, 'or', -, 'butnot'})
//
void CSearch::parse1(void) {
   char oper[8];

   parse2();

   while (true) {
      switch (op) {
         case OR:
            strcpy(oper, Pchar("|"));
            break;
         case BUTNOT:
            strcpy(oper, Pchar("-"));
            break;
         default:
            return;
         }

      getToken();
      parse2();
      outputPostfix(oper);
      }
   } // parse1

//---------------------------------------------------------------------
// Parse exp2 ::= exp3 {OP exp3}
// (where OP in {*, &, 'and'})
//
void CSearch::parse2(void) {
   char oper[8];

   parse3();

   while (true) {
      switch (op) {
         case AND:
            strcpy(oper, Pchar("&"));
            break;
         default:
            return;
         }

      getToken();
      parse3();
      outputPostfix(oper);
      }
   } // parse2

//---------------------------------------------------------------------
// Parse ( exp1 ) | "str" | str
//
void CSearch::parse3(void) {
   if      (op == LPAREN) {
      getToken();
      parse1();

      if (op != RPAREN)
         err = QRY_RPARENEXPECTED;
      }
   else if (op == RPAREN)
      return;
   else if (op == QUOTEDSTR)
      outputPostfix(stoken);
   else if (op == STRING) {
      char work[sizeString];
      ChrCpy(work, '"');
      Strcat(work, stoken, sizeString);
      ChrCat(work, '"',    sizeString);
      outputPostfix(work);
      }
   else
      err = QRY_BADSYMBOL;

   getToken();
   } // parse3

//---------------------------------------------------------------------
// Append the string preceded by a blank to the postfix buffer.
//
void CSearch::outputPostfix(Pchar _symbol) {
   if (!StrEmpty(postFixExp))
      ChrCat(postFixExp, spacer, sizeString);
   Strcat(postFixExp, _symbol, sizeString);
   } // outputPostfix

//=====================================================================
// Tokenizer
//=====================================================================

//---------------------------------------------------------------------
// Initialize the tokenizer
//
void CSearch::openTokenizer(Pchar _inFixExp) {
   pInFixExp = _inFixExp;
   delete[] stoken;
   stoken    = new char[sizeString];
   memset(stoken, 0, sizeString);
   atEnd     = false;
   getCh();
   } // openTokenizer

//---------------------------------------------------------------------
// Get a token
//
void CSearch::getToken(void) {
   // clear token string
   op = NOTKNOWN;
   memset(stoken, 0, sizeString);

   skipSpaces();

   if (atEnd)
      return;

   if      (getNamedOp())
     return;
   else if (ch == strQuote1)                                // "string"
      getQuotedString(strQuote1);
   else if (ch == strQuote2)                                // 'string'
      getQuotedString(strQuote2);
   else if (ch == strQuote3)                                // `string`
      getQuotedString(strQuote3);
   else if (ch != '&' &&
            ch != '*' &&
            ch != '|' &&
            ch != '+' &&
            ch != '-' &&
            ch != '(' &&
            ch != ')')
      getString();
   else
      getOp();
   } // getToken

//---------------------------------------------------------------------
// Skip white space
//
void CSearch::skipSpaces(void) {
   while (ch <= spacer && getCh());
   } // skipSpaces

//---------------------------------------------------------------------
// Get a non-blank string
//
void CSearch::getString(void) {
   op = STRING;
   ChrCat(stoken, ch, sizeString);
   short size = 1;
   while (getCh() && (ch != '&' &&
                      ch != '*' &&
                      ch != '|' &&
                      ch != '+' &&
                      ch != '-' &&
                      ch != '(' &&
                      ch != ')')) {
      size++;
      Assert(size < sizeString);
      ChrCat(stoken, ch, sizeString);
      }

   // If the token is number, convert to base 10 if needed
   if (IsNumber(stoken)) {
      ulong nr = StrToLong(stoken);
      sprintf(stoken, Pchar("%ld"), nr);
      }
   } // getString

//---------------------------------------------------------------------
// Get a quoted string
// (note: does not change underlines)
//
void CSearch::getQuotedString(char _quote) {
   op = QUOTEDSTR;
   ChrCat(stoken, ch, sizeString);
   short size = 1;
   while (getCh() && ch != _quote) {
      size++;
      Assert(size < sizeString);
      ChrCat(stoken, ch, sizeString);
      }
   if (ch != '\0')
      ChrCat(stoken, ch, sizeString);
   getCh();
   } // getQuotedString

//---------------------------------------------------------------------
// Get an operator string.
// (looking for 'and', 'or', or 'butnot')
//
bool CSearch::getNamedOp(void) {
   char work[sizeString];
   ChrCpy(work, ch);
   Strcat(work, pInFixExp, sizeString);
   CapStr(work);

   if      (StrPos(Pchar("AND "),    work) == 0) {
      op = AND;
      strcpy(stoken, Pchar("and"));
      pInFixExp += 3;
      getCh();
      return true;
      }
   else if (StrPos(Pchar("OR "),     work) == 0) {
      op = OR;
      strcpy(stoken, Pchar("or"));
      pInFixExp += 2;
      getCh();
      return true;
      }
   else if (StrPos(Pchar("BUTNOT "), work) == 0) {
      op = BUTNOT;
      strcpy(stoken, Pchar("butnot"));
      pInFixExp += 6;
      getCh();
      return true;
      }

   return false;
   } // getNamedOp

//---------------------------------------------------------------------
// Get an operator string.
// (should be a one-char or two-char operator)
//
void CSearch::getOp(void) {
   ChrCat(stoken, ch, sizeString);

   switch (ch) {
      case '&':
      case '*':
         op = AND;
         getCh();
         break;
      case '|':
      case '+':
         op = OR;
         getCh();
         break;
      case '-':
         op = BUTNOT;
         getCh();
         break;
      case '(':
         op = LPAREN;
         getCh();
         break;
      case ')':
         op = RPAREN;
         getCh();
         break;
      default:
         err = QRY_INVOPERATOR;
      }
   } // getOp

//---------------------------------------------------------------------
// Get the next character
//
bool CSearch::getCh(void) {
   ch = '\0';

   if (atEnd)
      return false;

   if (*pInFixExp == 0) {
      atEnd = true;
      return false;
      }

   ch = *pInFixExp++;
   return true;
   } // getCh

