
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


//*********************************************************************
// search.h
//*********************************************************************
//
// The CSearch class is patterned after the CEvaluate class. It parses
// an embedded infix query expression string into a postfix expression
// string, then evaluates the resulting expression.
//
// Evaluation comprises a set of searches against the specified farBook
// according to the specified expression, each search resulting in a
// list of node names, and logical operations between result lists to
// reduce in a single list.
//
// The recognized operator set includes the following:
//
// Precedence
//
//    3: ( )
//    2: *, &, "and"
//    1: +, |, "or",  -, "butnot"
//
//*********************************************************************

#ifndef SEARCH_H
#define SEARCH_H

#ifndef FARBOOK_H
#include "prefixb.h"
#include "book.h"
#include "findlist.h"
#ifndef FV_NOSTEM
#include "stem.h"
#endif
#endif

//---------------------------------------------------------------------
// Parser errors
//
typedef ulong ulong;
const ulong QRY_SVCERROR1 = 0;

const ulong QRY_INVCONDITIONAL    = QRY_SVCERROR1 + 29;
const ulong QRY_INVOPERATOR       = QRY_SVCERROR1 + 30;
const ulong QRY_LPARENEXPECTED    = QRY_SVCERROR1 + 31;
const ulong QRY_RPARENEXPECTED    = QRY_SVCERROR1 + 32;
const ulong QRY_COMMAEXPECTED     = QRY_SVCERROR1 + 33;
const ulong QRY_PARAMETERCOUNT    = QRY_SVCERROR1 + 34;
const ulong QRY_ENDNOTEXPECTED    = QRY_SVCERROR1 + 35;
const ulong QRY_CANTWRAPSTRING    = QRY_SVCERROR1 + 36;
const ulong QRY_BADSYMBOL         = QRY_SVCERROR1 + 37;

//---------------------------------------------------------------------
//
const ushort iSearchItem = 1000;

_FBDEF(CSearchItem)
class CSearchItem: public CItem {
   public:
      CSearchItem(Pchar _key);
      virtual ~CSearchItem();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      Pchar      key;
      PCFindList list;
   }; // CSearchItem
//---------------------------------------------------------------------
//

_FBDEF(CSearch)
class CSearch: public CAtom {
   public:
      enum fix {inFix, postFix};

      CSearch(PCBook _book);
      virtual ~CSearch();

      virtual FB_USE_(short) Ima(void);
      virtual FB_USE_(Pchar) IAm(void);
      virtual FB_USE_(bool)  Isa(const short _type);

      PCFindList Search     (Pchar _domain,
                             Pchar _expression,
                             bool  _fulText);

      FB_USE_(ulong) Error  (void) { return err; }
      FB_USE_(void)  Error  (ulong _err) { err = _err; }

   protected:
      bool     parse        (Pchar  _inFixExp,
                             Pchar  _postFixExp,
                             ushort _size);
      void     openTokenizer(Pchar  _inFixExp);
      void     getToken     (void);

      bool     replace      (Pchar _inp, Pchar _out, ushort _size);

      //---------------------------------------------------------------
      // Postfix evaluator
      //
      bool queryAnd       (PCList _stack);
      bool queryOr        (PCList _stack);
      bool queryButNot    (PCList _stack);

      bool query          (PCSearchItem _query);
      bool getString      (PCRecord  _exp, Pchar _token, ushort _size);

      //---------------------------------------------------------------
      // Infix expression parser
      //
      void parse1         (void);
      void parse2         (void);
      void parse3         (void);

      void outputPostfix  (Pchar _symbol);

      //---------------------------------------------------------------
      // Tokenizer methods
      void skipSpaces     (void);
      void getString      (void);
      void getQuotedString(char _quote);
      bool getNamedOp     (void);
      void getOp          (void);
      bool getCh          (void);

      //---------------------------------------------------------------
      // Parser operator ids
      //
      enum OpCode {NOTKNOWN,
                   AND,
                   OR,
                   BUTNOT,
                   LPAREN,
                   RPAREN,
                   STRING,
                   QUOTEDSTR};

      bool     usingIndex;
#ifndef FV_NOSTEM
      PCStem   stem;
#endif
      PCBook   pbook;
      PCBook   book;
      Pchar    domain;
      bool     atEnd;
      OpCode   op;
      Pchar    inFixExp;
      Pchar    pInFixExp;             // retained pointer into inFixExp
      Pchar    postFixExp;
      Pchar    stoken;
      ulong    err;
      char     ch;
   }; // CSearch

#endif // SEARCH_H
   
