
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// strng.cpp
//*********************************************************************

#include "prefixb.h"
#include "block.h"
#include "file.h"
#include "strng.h"

//---------------------------------------------------------------------
// Default constructor
//

CString::CString(void):
   CAtom(),
   chars(0),
   size (0) {
   } // CString

//---------------------------------------------------------------------
// Filled constructor
//
CString::CString(char _fill, int _size):
   CAtom(),
   chars(0),
   size (0) {
   if (_size > 0) {
      chars = new char[_size+1];
      memset(chars, _fill, _size);
      chars[_size] = 0;
      }
   } // CString

//---------------------------------------------------------------------
// Conversion constructor
//
CString::CString(const Pchar _chars):
   CAtom(),
   chars(0),
   size (0) {
   Assert(!StrEmpty(_chars));
   size  = short(strlen(_chars));
   chars = new char[size+1];
   strcpy(chars, _chars);
   } // CString

//---------------------------------------------------------------------
// Conversion constructor
//
CString::CString(const long _number, int _radix):
   CAtom(),
   chars(0),
   size (0) {
   char work[32];
//?ltoa(_number, work, _radix);
   sprintf(work, Pchar("%ld"), _number);
   size  = short(strlen(work));
   chars = new char[size+1];
   strcpy(chars, work);
   } // CString

//---------------------------------------------------------------------
// Copy constructor
//
CString::CString(const PCString _string):
   CAtom(),
   chars(0),
   size (0) {
   size  = _string->size;
   chars = new char[size+1];
   strcpy(chars, _string->chars);
   } // CString

//---------------------------------------------------------------------
// Destructor
//
CString::~CString() {
   delete[] chars;
   chars = 0;
   } // ~CString

//---------------------------------------------------------------------
//
RTTI_CPP(String, Atom)

//---------------------------------------------------------------------
// Replace the current value with the specified string
//
void CString::NewString(const Pchar _chars) {
   delete[] chars;
   chars = 0;
   if (!_chars)
      return;

   STRINIT(chars, _chars);
   } // NewString

//---------------------------------------------------------------------
// Assignment
//
const RCString CString::operator=(const RCString _string) {
   if (&_string != this)
      STRINIT(chars, _string.chars);
   return *this;
   } // operator=

//---------------------------------------------------------------------
// Concatenation
//
CString CString::operator+=(const RCString _string) {
   Pchar temp = chars;
   size      += _string.size;
   chars      = new char[size+1];
   strcpy(chars, temp);
   strcat(chars, _string.chars);
   delete[] temp;
   return *this;
   } // operator+=

//---------------------------------------------------------------------
// Concatenation
//
CString operator+(const RCString _strA,
                  const RCString _strB) {
   CString string(_strA);
   string += _strB;
   return string;
   } // operator+

//---------------------------------------------------------------------
// Test for empty
//
bool CString::operator!(void) const {
   return (StrEmpty(chars));
   } // operator!

//---------------------------------------------------------------------
// Test == _string
//
bool CString::operator==(const RCString _string) {
   return strcmp(chars, _string.chars) == 0;
   } // operator==

//---------------------------------------------------------------------
// Test != _string
//
bool CString::operator!=(const RCString _string) {
   return strcmp(chars, _string.chars) != 0;
   } // operator!=

//---------------------------------------------------------------------
// Test < _string
//
bool CString::operator<(const RCString _string) {
   return strcmp(chars, _string.chars) < 0;
   } // operator<

//---------------------------------------------------------------------
// Test > _string
//
bool CString::operator>(const RCString _string) {
   return strcmp(chars, _string.chars) > 0;
   } // operator>

//---------------------------------------------------------------------
// Test <= _string
//
bool CString::operator<=(const RCString _string) {
   return strcmp(chars, _string.chars) <= 0;
   } // operator<=

//---------------------------------------------------------------------
// Test >= _string
//
bool CString::operator>=(const RCString _string) {
   return strcmp(chars, _string.chars) >= 0;
   } // operator>=

//---------------------------------------------------------------------
// Return char reference
//
Rchar CString::operator[](short _index) {
   Assert(_index >= 0 && _index < size);

   return chars[_index];
   } // operator[]

//---------------------------------------------------------------------
// Return substring
//
// if _begin > 0, then offset beginning as specified.
// if _end   > 0, then offset ending as specified.
// if _begin < 0, then add size to specified beginning.
// if _end   < 0, then add size to specified ending.
// if _begin = 0 and _end = 0, then copy entire string.
// if _begin or _end are out of range, an empty substring is returned.
//
CString CString::operator()(short _begin, short _end) {
   CString sub;                                         // empty string

   // Bound the substring to copy
   short begin = 0;
   short end   = size;
   if (_begin == 0 && _end == 0);
   else {
      if      (_begin < 0)
         begin = _begin + size;
      else if (_begin > 0)
         begin = _begin;
      if      (_end   < 0)
         end   = _end   + size;
      else if (_end   > 0)
         end   = _end;
      }
   if (begin < size && end > 0) {
      begin     = Bound(begin, short(0), short(size-1));
      end       = Bound(end,   short(0), short(size-1));
      sub.size  = 0 < (end - begin + 1) ? end - begin + 1 : 0;
      if (sub.size > 0) {
         // Allocate the substring memory
//?      delete[] sub.chars;
         sub.chars = new char[sub.size+1];
         Assert(sub.chars);

         // Copy the substring
         memset(sub.chars, 0, sub.size+1);
         strncpy(sub.chars, &chars[begin], sub.size);
         }
      }

   return sub;
   } // operator()

//---------------------------------------------------------------------
//

bool CString::IsEqual(PCString _strB) {
   return (strcmp(chars, _strB->chars) == 0);
   } // IsEqual

//---------------------------------------------------------------------
//
PCString CString::Insert(const PCString _string, short _where) {
   PCString insert = new CString();

   ushort size  = strlen(chars) + strlen(_string->chars);
   Pchar work   = new char[size + 1];
   strcpy(work, chars);
   StrNsrt(work, _string->chars, _where, size);
   delete[] insert->chars;
   insert->chars = work;
   insert->size  = short(strlen(work));

   return insert;
   } // Insert

//---------------------------------------------------------------------
//
PCString CString::Delete(short _where, short _size) {
   short where = _where;
   Assert(where >= 0 && where <= size);

   short sz = _size;
   Assert(sz > 0 && sz <= size);
   Assert(where + sz <= size);

   PCString del = new CString();
   Pchar work = new char[strlen(chars) + 1];
   strcpy(work, chars);
   strcpy(work + where, work + where + sz);
   del->size  = strlen(work);
   del->chars = work;

   return del;
   } // Delete

//---------------------------------------------------------------------
// Return substring before pattern
//
PCString CString::Before(PCString _string) {
   PCString before = new CString();                     // empty string

   char work[sizeString];
   ::Before(_string->chars, chars, work, sizeString);

   before->size = short(strlen(work));
   delete[] before->chars;
   before->chars = new char[before->size+1];
   strcpy(before->chars, work);

   return before;
   } // Before

//---------------------------------------------------------------------
// Return substring after pattern
//
PCString CString::After(const PCString _string) {
   PCString after = new CString();                      // empty string

   char work[sizeString];
   ::After(_string->chars, chars, work, sizeString);

   after->size = short(strlen(work));
   delete[] after->chars;
   after->chars = new char[after->size+1];
   strcpy(after->chars, work);

   return after;
   } // After

//---------------------------------------------------------------------
// Return length
//
short CString::Size(void) const {
   return size;
   } // Size

//---------------------------------------------------------------------
// Return the position of the pattern
//
short CString::Position(const PCString _string) {
   return StrPos(_string->chars, chars);
   } // Position

//---------------------------------------------------------------------
//
void CString::Format(Pchar _format, ...) {
   if (!StrEmpty(_format)) {
      va_list args;
      va_start(args, _format);
      char line[sizeString];
      vsprintf(line, _format, args);
      va_end(args);
      Assert(strlen(line) < sizeString);
      NewString(line);
      }
   } // Format

//---------------------------------------------------------------------
// Capitalize the string
//
void CString::Cap(void) {
   ::CapStr(chars);
   } // Cap

//---------------------------------------------------------------------
// Remove duplicate sequnces of the specified
// character. Return size of resulting string.
//
short CString::Squeeze(char _ch) {
   Pchar work = new char[size+1];
   strcpy(work, chars);
   ::Squeeze(work, _ch);

   delete[] chars;
   size  = short(strlen(work));
   chars = new char[size+1];
   strcpy(chars, work);

   delete[] work;
   return size;
   } // Squeeze

//---------------------------------------------------------------------
// Remove all spaces. Return size of resulting string.
//
short CString::Pack(void) {
   Pchar work = new char[size+1];
   strcpy(work, chars);
   ::Pack(work);

   delete[] chars;
   size  = short(strlen(work));
   chars = new char[size+1];
   strcpy(chars, work);

   delete[] work;
   return size;
   } // Pack

//---------------------------------------------------------------------
// Remove spaces from the specifed ends. Return size
// of resulting string.
//
short CString::Trim(whichEnd _end) {
   Pchar work = new char[size+1];
   strcpy(work, chars);
   ::Trim(work, _end);

   delete[] chars;
   size  = short(strlen(work));
   chars = new char[size+1];
   strcpy(chars, work);

   delete[] work;
   return size;
   } // Trim

//---------------------------------------------------------------------
// Pad with spaces to the specified size. Return
// size of resulting string.
//
short CString::Pad(short _sz) {
   if (size >= _sz)
      return size;

   Pchar work = new char[_sz+1];
   strcpy(work, chars);
   ::PadStr(work, _sz, _sz);

   delete[] chars;
   size  = short(strlen(work));
   chars = new char[size+1];
   strcpy(chars, work);

   delete[] work;
   return size;
   } // Pad

//---------------------------------------------------------------------
// return true if a number
//
bool CString::IsNumber(void) {
   return ::IsNumber(chars);
   } // IsNumber

//---------------------------------------------------------------------
// return long value
//
long CString::Long(void) {
   return StrToLong(chars);
   } // Long

//---------------------------------------------------------------------
// return short value
//
short CString::Short(void) {
   return StrToShort(chars);
   } // Short

//---------------------------------------------------------------------
// Copy the first string in the block into the string.
//
bool CString::BlockTo(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   // Check for actual data
   Pchar p = block->Lock();
   Assert(p);

   // Check for proper termination
   ulong sz = strlen(p);
   if (sz >= block->GetSize())
      return false;

   // Copy the string
   NewString(p);

   return true;
   } // BlockTo

//---------------------------------------------------------------------
// Copy the string into the block.
//
bool CString::ToBlock(PCAtom _block) {
   // Check it is what it says it is
   if (!(_block->Isa(iBlock)))
      return false;

   PCBlock block = TYPECAST(_block, Block);

   block->NewBlock(chars, size);

   return true;
   } // ToBlock

//---------------------------------------------------------------------
//
ulong CString::BookSize(void) {
   return strlen(chars) + 1;
   } // BookSize
   
//---------------------------------------------------------------------
// Read string from specified file. Return result
// of read.
//
bool CString::ReadFile(PCAtom _file, bool _isBook) {
   char line[sizeString];

   if (!(TYPECAST(_file, File)->ReadString(line, sizeString)))
      return false;

   STRINIT(chars, line)
   size  = strlen(line);

   return true;
   } // ReadFile

//---------------------------------------------------------------------
// Write string to specified file. Return result
// of write.
//
bool CString::WriteFile(PCAtom _file, bool _isBook) {
   return TYPECAST(_file, File)->WriteString(chars);
   } // WriteFile

//---------------------------------------------------------------------
// Read a string from the specified file
//
bool CString::Load(Pchar _fileName) {
   PCFile input = new CFile(_fileName);

   if (!input->OpenInput())
      return false;

   Pchar s   = input->ReadString();
   delete[] chars;
   chars     = s;

   input->Close();
   return true;
   } // Load

//---------------------------------------------------------------------
// Write the contents of the string to the specified file.
//
bool CString::Save(Pchar _fileName) {
   PCFile output = new CFile(_fileName);

   if (!output->OpenOutput())
      return false;

   output->WriteString(chars);
   output->Close();

   return true;
   } // Save


