
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Paul Medlock.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Paul Medlock, <paul@medlock.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*********************************************************************
// CSysLog.cpp
//*********************************************************************
//
// Notes:
//
// Do not replace farview version without resolving differences.
//
// This class provides a logging facility that minimizes its use of
// other farBook classes. It is, however, written using Windows APIs.
//
// After initializing a CSysLog object, You use the Log() static method
// to log a message. The format of the call is as in the following
// example:
//
//   CSysLog::Log('D', "CSession.ackError: Sent %s", msg);
//
// Log() uses command-line flags to determine its behaviour. When the
// 'X' flag is in the command line, Log() outputs the specified log
// message regardless of the specified flag parameter. When 'X' is not
// present, but the specified flag is in the command line, Log()
// outputs the specified log message. Otherwise the log message is
// ignored.
//
// Normally, output is directed to the program's log file and to any
// Windows program watching for the ::OutputDebugString() API. However,
// if the Y-flag is in the command line, the log output Is not directed
// to the log file, but is only sent to the debug watch program.
//
// If you are debugging and you expect that the program under test is
// likely to fail while testing, you can include the 'Z' flag on the
// command line. This flag causes Log() to append and flush log
// messages to the log file immediately. Under normal circumstances,
// however, don't use the 'Z' flag, causing the Log() method to buffer
// log messages to a CStrList, which it dumps to the log file
// periodically.
//
// Note: Log() Asserts when messages exceed sizeString-1 bytes.
//
//*********************************************************************

#include "prefixb.h"
#include "args.h"
//#include "console.h"
#include "misc.h"
#include "strlist.h"
#include "syslog.h"

const char cr = '\r';
const char lf = '\n';

char SysLog::eolDelim[] = {0x0d, 0x0a, 0};
pSysLog SysLog::logger = 0;

//=====================================================================
// loCS class implementation
//
SysLog::loCS::loCS(void) {
#if   defined(FV_MSWIN32)
   ::InitializeCriticalSection(&m_cs);
#endif
   } // loCS

//---------------------------------------------------------------------

//
SysLog::loCS::~loCS() {
#if   defined(FV_MSWIN32)
   ::DeleteCriticalSection(&m_cs);
#endif
   } // ~loCS

//---------------------------------------------------------------------
//
void SysLog::loCS::Enter(void) {
#if   defined(FV_MSWIN32)
   ::EnterCriticalSection(&m_cs);
#endif
   } // Enter

//---------------------------------------------------------------------
//
void SysLog::loCS::Leave(void) {
#if   defined(FV_MSWIN32)
   ::LeaveCriticalSection(&m_cs);
#endif
   } // Leave

//---------------------------------------------------------------------
// If you get back false, don't write to the log file.
//
bool SysLog::Enter(void) {
   if (!logger)
      return false;

   logger->cs.Enter();
   return true;
   } // Enter

//---------------------------------------------------------------------
//
void SysLog::Leave(void) {
   Assert(logger);
   logger->cs.Leave();
   } // Leave

//---------------------------------------------------------------------
//
void SysLog::Log(char _flag, Pchar _format, ...) {
     bool xOn = CArgs::TheArgs->FArg('X');
     bool fOn = CArgs::TheArgs->FArg(_flag);

     if (!xOn && !fOn)
       return;

   if (Enter()) {
      char line[4*sizeString];
      va_list args;
      va_start(args, _format);
      vsprintf(line, _format, args);
      va_end(args);
      Assert(strlen(line) < 4*sizeString);
      char date[sizeLine];
      GetGMT(date, sizeLine);
      char work[4*sizeString];
      sprintf(work, "%c %s %s", _flag, date, line);
      Assert(strlen(work) < 4*sizeString);

      ::OutputDebugString(work);

      if (!logger) {
         Leave();
         return;
         }

//  if (CArgs::TheArgs->FArg('Y'))
//     return;

    if (!CArgs::TheArgs->FArg('Z')) {         // don't change this flag
         if (!StrEmpty(_format)) {
            // Save formatted message to list
            logger->list->StrInsert(work);
            }
         if (logger->list->Number() >= 400) {
            // Dump list of formatted messages
            logger->OpenAppend();
            if (logger->list->GotoHead()) do {
               logger->WriteString(logger->list->StrItem(false));
               } while (logger->list->GotoNext());
            logger->list->Flush();
            logger->Close();
            }
         }
      else {
         logger->OpenAppend();
         if (!StrEmpty(_format)) {
            // Write formatted message to file
            logger->WriteString(work);
            }
         logger->Close();
         }
      Leave();
      }
   } // Log

//---------------------------------------------------------------------
// Constructor
//
SysLog::SysLog(void):
   name        (0),
   fileExists  (false),
   fileIsOpen  (false),
   fileCanWrite(false),
   file        (0),
   whereAt     (0),
   sysErr      (FVE_OK),
   error       (FVE_FILENOTOPEN) {

   Assert(!logger);
   logger = this;

   list   = new CStrList();

   memset(delim, 0, sizeName);
   Strcpy(delim, eolDelim, sizeName);
   } // SysLog

//---------------------------------------------------------------------
// Constructor
//
SysLog::SysLog(const Pchar _name, const Pchar _path):
   name        (0),
   fileExists  (false),
   fileIsOpen  (false),
   fileCanWrite(false),
   file        (0),
   whereAt     (0),
   sysErr      (FVE_OK),
   error       (FVE_FILENOTOPEN) {

   Assert(!logger);

   list   = new CStrList();
   
   Assert(!StrEmpty(_name));
   name   = new char[strlen(_name)+1];
   strcpy(name, _name);

   memset(delim, 0, sizeName);
   Strcpy(delim, eolDelim, sizeName);

   // process the file set parameter if requested
   if (StrEmpty(_path)) {
      // the filename includes the path,
      // else the file is in the cwd
      // Test for file existence
      ulong err;
      if (FileExists(name, err)) {
         // found the file
         fileExists = true;

         // Qualify the file name
         char curDir[sizeString];
         GetCurrentDir(curDir, sizeString);
         char work[sizeString];
         Strcpy(work, name, sizeString);
         CatName(curDir, work, 0, sizeString);
         delete[] name;
         name = new char[strlen(work)+1];
         strcpy(name, work);
         }
      }
   else {
      // prepare the path string
      PCRecord path; path = new CRecord(_path, Pchar(";"));
      char pathName[sizeString];
      char work[sizeString];

      if (path->Number() == 1) {
         // path is specified
         path->GetField(1, pathName, sizeString);
         Strcpy(work, _name, sizeString);
         CatName(pathName, work, 0, sizeString);
         delete[] name;
         name = new char[strlen(work)+1];
         strcpy(name, work);
         if (name[0] == '"')
            strcpy(name, name+1);
         ulong err;
         if (FileExists(name, err))
            // found the file
            fileExists = true;
         }
      // traverse the path list (record)
      else for (short i = 1; i <= path->Number(); i++) {
         path->GetField(i, work, sizeString);
         Strcat(work, pathDelim, sizeString);// pathDelim in [":", "\"]
         Strcat(work, _name, sizeString); // form of dir is di:fileName
                                                      // or di\fileName
         ulong err;
         if (FileExists(work, err)) {
            // found the file
            fileExists = true;
            delete[] name;
            name = new char[strlen(work)+1];
            strcpy(name, work);
            break;
            }
         }
      }

   logger = this;
   } // SysLog

//---------------------------------------------------------------------
// Destructor
//
SysLog::~SysLog() {
//?Log('D', Pchar("SysLog deleted"));
   logger = 0;

   if (!(list->Empty())) {
      // Dump list of formatted messages
      OpenAppend();
      if (list->GotoHead()) do {
         WriteString(list->StrItem(false));
         } while (list->GotoNext());
      Close();
      }

   delete[] name;
   name = 0;
   
   if (fileIsOpen)
      Close();

   file = 0;
   } // ~SysLog

//---------------------------------------------------------------------
// Return the error state of the file
//
ulong SysLog::Error(void) {
   return error;
   } // Error

//---------------------------------------------------------------------
// Set the error state of the file
//
void SysLog::Error(ulong _error) {
   error = _error;
   } // Error

//---------------------------------------------------------------------
// Create a closed file
//
bool SysLog::Create(void) {
   Assert(!fileIsOpen);

   sysErr = FVE_OK;
   error  = FVE_OK;

   if (OpenOutput())
      Close();
   else
      error = FVE_FILEISOPEN;

   return (error == FVE_OK);
   } // Create

//---------------------------------------------------------------------
// Open the file as output, destroying any existing file
//
bool SysLog::OpenOutput(void) {
   sysErr = FVE_OK;
   error  = FVE_OK;

   Assert(!fileIsOpen);

   fileCanWrite = true;

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(CreateFile(name,
                                GENERIC_WRITE,
                                0,
                                &security,
                                CREATE_ALWAYS,
                                FILE_ATTRIBUTE_ARCHIVE |
                                FILE_FLAG_RANDOM_ACCESS,
                                0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      error = FVE_BADOPEN;
      file  = 0;
      return false;
      }

#elif defined(FV_UNIX)
   file = FB_HANDLE(::creat(name, 0770));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      error  = FVE_BADOPEN;
      return false;
      }


#endif // FV_UNIX
   fileIsOpen = true;
   return true;
   } // OpenOutput

//---------------------------------------------------------------------
// Open the file with the file pointer at the end of the file
//
bool SysLog::OpenAppend(void) {
   sysErr = FVE_OK;
   error  = FVE_OK;

   Assert(!fileIsOpen);

   fileCanWrite = true;

#if   defined(FV_MSWIN32)
   SECURITY_ATTRIBUTES security;
   security.nLength              = sizeof(SECURITY_ATTRIBUTES);
   security.lpSecurityDescriptor = 0;
   security.bInheritHandle       = false;

   file = FB_HANDLE(CreateFile(name,
                                GENERIC_WRITE,
                                0,
                                &security,
                                OPEN_ALWAYS,
                                FILE_ATTRIBUTE_ARCHIVE |
                                FILE_FLAG_RANDOM_ACCESS,
                                0));

   if (HANDLE(file) == INVALID_HANDLE_VALUE) {
      error = FVE_BADOPEN;
      file  = 0;
      return false;
      }

#elif defined(FV_UNIX)
   file = FB_HANDLE(::open(name, 1));
   if (file < 0) {
      sysErr = FVE_BADOPEN;
      error  = FVE_BADOPEN;
      return false;
      }

#endif // FV_UNIX

   fileIsOpen = true;
   Seek();
   return true;
   } // OpenAppend

//---------------------------------------------------------------------
// Close the file
//
bool SysLog::Close(void) {
   sysErr = FVE_OK;
   Error(FVE_OK);

if (!fileIsOpen)
   Assert(fileIsOpen);

#if   defined(FV_MSWIN32)
   if (!CloseHandle(HANDLE(file))) {
      sysErr     = GetLastError();
      error      = FVE_BADCLOSE;
      }
   else
      fileIsOpen = false;

#elif defined(FV_UNIX)
   ::close(int(file));

#endif // FV_UNIX
   file = 0;
   return (error == FVE_OK);
   } // Close

//---------------------------------------------------------------------
// Seek to the end of the file
//
bool SysLog::Seek(void) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   error  = FVE_OK;

#if   defined(FV_MSWIN32)
   whereAt = ::SetFilePointer(HANDLE(file), 0, 0, FILE_END);
   if (whereAt < 0)

      sysErr = ::GetLastError();

#elif defined(FV_UNIX)
   whereAt = ::lseek(int(file), 0, 2);
   if (whereAt < 0)
      sysErr = errno;

#endif // FV_UNIX

   if (whereAt < 0)
      return false;
   else
      return true;
   } // Seek

//---------------------------------------------------------------------
// Write a formatted string from _p
//
bool SysLog::Printf(Pchar _format, ...) {
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   error  = FVE_OK;

   if (!StrEmpty(_format)) {
      // Print formatted message
      va_list args;
      va_start(args, _format);
      char line[sizeString];
      vsprintf(line, _format, args);
      va_end(args);
      Assert(strlen(line) < sizeString);
      return WriteString(line);
      }
   return false;
   } // Printf

//---------------------------------------------------------------------
// Write a string from _p
//
bool SysLog::WriteString(Pchar _p) {
if (!fileIsOpen)
   Assert(fileIsOpen);

   sysErr = FVE_OK;
   error  = FVE_OK;

   if (strlen(delim)) {
      if (write(_p, strlen(_p)))
         return write(delim, strlen(delim));
      else
         return false;
      }
   else
      // Include the trailing zero in the output
      return write(_p, strlen(_p)+1);
   } // WriteString

//---------------------------------------------------------------------
// Write a block to the file
//---------------------------------------------------------------------
bool SysLog::write(Pvoid _buffer, ulong _size) {
   if (!fileCanWrite) {
      error = FVE_FILENOTOUTPUT;
      return false;
      }

#if   defined(FV_MSWIN32)
   DWORD size;
   if (!::WriteFile(HANDLE(file), _buffer, _size, &size, 0)) {
      sysErr = GetLastError();
      error  = FVE_BADWRITE;
      }

#elif defined(FV_UNIX)
   ulong size = ::write(int(file), _buffer, size_t(_size));
   if (size != _size) {
      sysErr = FVE_BADWRITE;
      error  = FVE_BADWRITE;
      }

#endif // FV_UNIX
   return (error == FVE_OK);
   } // write

